/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "browserwidget.h"
#include "../tools/difftool.h"
#include "../networkengine/filerepository.h"
#include "../networkengine/connection.h"
#include "../soundengine/soundengine.h"
#include "../global.h"

#include <QVBoxLayout>
#include <QTextBrowser>
#include <QUrl>
#include <QDir>
#include <QDate>
#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <QDesktopServices>
#include <QRegExp>

//creation of the browser object
BrowserWidget::BrowserWidget(QWidget* parent) : QWidget(parent) {
	//-------------text display-----------------------------------
	internalBrowser = new QTextBrowser(this);
	connect(internalBrowser, SIGNAL(anchorClicked(const QUrl&)), this, SLOT(openLink(const QUrl&)) );
	connect(internalBrowser, SIGNAL(highlighted(const QUrl&)), this, SLOT(highlightLink(const QUrl&)) );
	
	internalEdit = new QTextEdit(this);
	internalEdit->setFontPointSize(11);
	internalEdit->setVisible(false);
	connect(internalEdit, SIGNAL(textChanged()), this, SIGNAL(textChanged()) );
	
	//text size (it implies the windows size)
	internalBrowser->setMinimumWidth(Global::BrowserMinimumWidth);
	internalBrowser->setMinimumHeight(Global::BrowserMinimumHeight);
	internalEdit->setMinimumWidth(Global::EditMinimumWidth);
	internalEdit->setMinimumHeight(Global::EditMinimumHeight);
	
	//text parameters
	internalBrowser->setReadOnly(true);
	internalBrowser->setOpenLinks(false);
	internalEdit->setReadOnly(false);
	
	//------------------vertical layout --------------------------
	QVBoxLayout *vboxPopup = new QVBoxLayout( this );
	vboxPopup->setSpacing(5);
	vboxPopup->setMargin(5);
	vboxPopup->addWidget(internalBrowser);
	vboxPopup->addWidget(internalEdit);
	
	//------------------------------------------------------------
	//initialize the list of special strings and their translations
	specialStrings += QString("#AUTHOR");
	specialStringsTranslations += tr("Author");
	specialStrings += QString("#LICENSE");
	specialStringsTranslations += tr("License");
	specialStrings += QString("#SOURCELINK");
	specialStringsTranslations += tr("Source link");
}

//display the content correspondind to a given url
//note : the file has been downloaded when this function is called
//paramter htmlActive : display the string in the browser (normal display) or in the editor (edit mode)
void BrowserWidget::setContent(const QUrl& url, bool htmlActive) {
	QString urlFileName = url.path();
	QString localCacheName = FileRepository::getAbsoluteFileName(urlFileName);
	
	//url without the possible old resource (to identify the file extension correctly)
	QString nameWithoutOld = urlFileName;
	if(FileRepository::isOldFile(urlFileName)) {
		nameWithoutOld = getTargetFileName(urlFileName);
	}
	
	//Read the file
	//Create the string that will be displayed
	QString displayedHtml;
	if(urlFileName.endsWith(FileRepository::logResource)) {
		//case of a history file
		displayedHtml = createHistoryHtml(localCacheName, url.path());
	} else
	if(!url.queryItemValue("diffWith").isEmpty()) {
		//case of a difference between two files
		QString oldFileName = FileRepository::getAbsoluteFileName( getDiffFileName(url.path(), url.queryItemValue("diffWith").toInt()) );
		displayedHtml = createDiffHtml(oldFileName, localCacheName);
	} else
	if(nameWithoutOld.endsWith(".png") || nameWithoutOld.endsWith(".jpg")) {
		//in case of an image file, we display it in a page as a link
		QString currentPath = url.path();
		displayedHtml = QString("[[Image:") + currentPath + "]]\n";
		if(htmlActive){
			formatToStandardHtml(displayedHtml);
			createLeftMenu(displayedHtml);
		}
	} else
	if(nameWithoutOld.endsWith(".wav") || nameWithoutOld.endsWith(".ogg")) {
		//case of a wav file, we play it with the soundengine
		QStringList wavFileNames;
		wavFileNames.append(localCacheName);
		soundEngine->loadWavFiles(wavFileNames);
		soundEngine->enableSound(true);	//activate the sounds on the audio device
		soundEngine->playSound(0, 1); //play the file one time only
		
		displayedHtml = tr("Wav sound file")+ "\n";
		if(htmlActive){
			formatToStandardHtml(displayedHtml);
			createLeftMenu(displayedHtml);
		}
	} else
	if(url.queryItemValue("action")=="credits") {
		//case of the credits for a game file
		displayedHtml = createGameCredits();
		if(htmlActive){
			formatToStandardHtml(displayedHtml);
			createLeftMenu(displayedHtml);
		}
	} else {
		//default case
		QFile htmlFile(localCacheName);
		htmlFile.open( QIODevice::ReadOnly );
		QTextStream text_stream( &htmlFile );
		text_stream.setCodec("UTF-8");
		displayedHtml = text_stream.readAll();
		htmlFile.close();
		if(htmlActive){
			formatToStandardHtml(displayedHtml);
			createLeftMenu(displayedHtml);
		}
	}
	
	//display the resulting string
	if(htmlActive) {
		//case of the normal display mode
		internalEdit->setVisible(false);
		internalBrowser->setVisible(true);
		
		//set the resulting string as the html content of the browser
		QStringList search;
		search += FileRepository::getCurrentHostCacheFolder();
		internalBrowser->setSearchPaths(search);
		internalBrowser->setHtml(displayedHtml);
	} else {
		//case of the edit mode
		internalBrowser->setVisible(false);
		internalEdit->setVisible(true);
		//we need to set current font and black colors (they are modified for example after a click on a link)
		//internalEdit->setCurrentFont(QApplication::font());	//use the system default font
		//internalEdit->setTextColor(Qt::black);				//use black text color
		internalEdit->setPlainText(displayedHtml);
		internalEdit->setFocus();
	}
}

void BrowserWidget::setContentPreview(const QUrl& /*url*/) {
	QString displayedHtml;
	displayedHtml=internalEdit->toPlainText();
	formatToStandardHtml(displayedHtml);
	createLeftMenu(displayedHtml);
	internalEdit->setVisible(true);
	internalBrowser->setVisible(true);
	internalBrowser->setHtml(displayedHtml);
}

//called when an anchor link is clicked in the browser
void BrowserWidget::openLink(const QUrl& link) {
	QString urlCleanString = FileRepository::nameWithoutSpaces(link.toString());
	QUrl cleanLink(urlCleanString);
	
	QString scheme = cleanLink.scheme();
	if(scheme=="http") {
		//http protocol, call the default browser
		QDesktopServices::openUrl(cleanLink);
	} else
	if(scheme=="fkz") {
		//flukz protocol, absolute path
		emit askPage(cleanLink);
	} else
	if(scheme.isEmpty()) {
		//this is a relative path
		QUrl askedUrl = deduceAbsoluteUrl(currentUrl, cleanLink);
		emit askPage(askedUrl);
	}
}

//called when an anchor link is highlighted in the browser (but not clicked)
void BrowserWidget::highlightLink(const QUrl& link) {
	//false highlight : when the user leaves a highlighted link, an empty url higlight signal is received
	if(link.toString().isEmpty()) {
		emit statusMessage(link.toString());
		return;
	}
	
	QString urlCleanString = FileRepository::nameWithoutSpaces(link.toString());
	QUrl cleanLink(urlCleanString);
	//real highlight of a normal link
	if(cleanLink.scheme().isEmpty()) {
		//this is a relative path
		QUrl askedUrl = deduceAbsoluteUrl(currentUrl, cleanLink);
		emit statusMessage(askedUrl.toString());
	} else {
		emit statusMessage(cleanLink.toString());
	}
}

//return the absolute Url obtained from a base url with a relative url
//Note : this function avoids QUrl::resolved(), which doesn't seem to work properly
QUrl BrowserWidget::deduceAbsoluteUrl(const QUrl& baseUrl, const QUrl& relativeUrl) {
	QString relativePath = relativeUrl.path();
	QString resultPath;
	
	if(relativePath.startsWith("/")) {
		//the new url is an absolute one
		resultPath = relativePath;
	} else {
		//the new url is relative to the current one
		QString basePath = baseUrl.path();			//base path
		
		//in case of trash, we need to remove first the folder from the address
		//because relative adress were designed for a file that was originaly outside of this folder
		if(basePath.startsWith(FileRepository::trashFolderName)) {
			basePath.remove(0, FileRepository::trashFolderName.size()-1);
		} else
		if(basePath.startsWith(QString("/")+FileRepository::trashFolderName)) {
			basePath.remove(0, FileRepository::trashFolderName.size());
		}
		
		//remove the name of the current file
		int last = basePath.lastIndexOf('/');
		if(last!=-1) basePath.truncate(last+1);		//removes everything after the last "/"
		
		resultPath = QDir::cleanPath(basePath + relativePath);
	}
	
	QUrl resultUrl = baseUrl;
	resultUrl.setPath(resultPath);
	resultUrl.setQueryItems(relativeUrl.queryItems());
	
	return resultUrl;
}

//used to communicate to the browser the list of files needed for game credits 
void BrowserWidget::setGameCreditsList(QStringList creditsList) {
	gameCreditsList = creditsList;
}

//called in FlukzWidget when a given url and all its dependency files have been downloaded successfuly
void BrowserWidget::downloadEnd(const QUrl& url) {
	currentUrl = url;
	isExistingPage = true;
	setContent(url, /*htmlActive*/ true);
}

//called when the file corresponding to a url was not found
void BrowserWidget::connectionError(Message::ErrorCode errorCode, const QUrl& url, bool isEditable) {
	currentUrl = url;
	isExistingPage = false;
	
	QString errorString;
	if(errorCode==Message::UnknownError) {
		errorString = tr("Unknown network error.");
	} else
	if(errorCode==Message::HostNotFoundError) {
		errorString = tr("The host %1 was not found.").arg(url.host());
	} else
	if(errorCode==Message::ConnectionRefusedError) {
		errorString = tr("%1 refused the connection.").arg(url.host());
	} else
	if(errorCode==Message::SocketTimeoutError) {
		errorString = tr("The connection timed out.");
	} else
	if(errorCode==Message::FileNotFoundError) {
		errorString = tr("The page %1 does not exist yet on %2.\n").arg(url.path(), url.host());
		if(isEditable) {
			errorString += tr("You can create it by starting the edition.");
		}
	} else
	if(errorCode==Message::InvalidProtocolError) {
		errorString = tr("Sorry. Because of some important changes, you need to update to a new version of Flukz:\n[[http://flukz.org/]]");
	}
	
	//display the error in the browser (with the left links)
	QString noLanguageLinks("");
	formatToStandardHtml(errorString);
	addLeftMenu(errorString, noLanguageLinks, /*isErrorPage*/ true);
	internalEdit->setVisible(false);
	internalBrowser->setVisible(true);
	internalBrowser->setHtml(errorString);
}

//start the edit mode
//we display the content of the file in the editor
void BrowserWidget::startEditing() {
	if(isExistingPage) {
		setContent(currentUrl, /*htmlActive*/ false);
	} else {
		//this is a new page, we start with an empty page
		internalBrowser->setVisible(false);
		internalEdit->setVisible(true);
		internalEdit->setPlainText("");
		internalEdit->setFocus();
	}
}

//stop the edit mode
//we display the content of the file in the browser
void BrowserWidget::stopEditing() {
	if(isExistingPage) {
		setContent(currentUrl, /*htmlActive*/ true);
	} else {
		//case of a new page (edition was stopped before the page was saved)
		bool pageIsEditable = true;
		connectionError(Message::FileNotFoundError, currentUrl, pageIsEditable);
	}
}

void BrowserWidget::startPreview(){
	setContentPreview(currentUrl);
}

//save the content of the internal browser in a local temporary file
void BrowserWidget::saveTemporaryFile() {
	QString localCacheName = FileRepository::temporaryFileName();
	QFile htmlFile(localCacheName);
	htmlFile.open( QIODevice::WriteOnly );
	QTextStream text_stream( &htmlFile );
	text_stream.setCodec("UTF-8");
	text_stream << internalEdit->toPlainText();	//write file
	text_stream.flush();
	htmlFile.close();
}

//return the adress for the homePage link
QString BrowserWidget::homePage() {
	return tr("/en/Homepage");
}

//return the name of the target file after removing the last resource (log, talk, old, meta)
QString BrowserWidget::getTargetFileName(const QString& shortFileName) {
	QString targetFileName=shortFileName;
	
	//remove the initial "/" if any
	if(targetFileName.startsWith("/")) {
		targetFileName.remove(0, 1);
	}
	
	//find the last resource separator character
	int dotIndex = targetFileName.lastIndexOf(FileRepository::resourceSeparator);
	if(dotIndex!=-1) {
		QString extension = targetFileName.mid(dotIndex);
		//remove log, old, and talk resources
		if(extension==FileRepository::logResource || extension==FileRepository::talkResource ||
		   extension==FileRepository::metaResource || extension.startsWith(FileRepository::oldResource)
		   ) {
			targetFileName = targetFileName.left(dotIndex);
		}
	}
	
	return targetFileName;
}

//return the name of a file revision, based on the normal name and an oldIndex
//note : an initial "/" is included in the returned file name
QString BrowserWidget::getOldFileName(QString basePath, int oldIndex) {
	//address of the old file
	if(basePath.startsWith("/")) basePath.remove(0, 1);
	QString oldFileName = QString("/") + FileRepository::getOldFileName(basePath, oldIndex);
	return oldFileName;
}

//return the name of an old file of index oldIndex, relatively to a file basePath
//basePath can be either an old file, in which case we change only the oldIndex
//or it can be a normal file, in which case we need to add the old resource 
QString BrowserWidget::getDiffFileName(QString basePath, int oldIndex) {
	//address of the old file
	QString oldFileName;
	if(basePath.startsWith("/")) basePath.remove(0, 1);
	
	//TODO : seems redundant with FileRepository::getOldFileName()
	int oldTagPos = basePath.indexOf(FileRepository::oldResource);
	if(oldTagPos!=-1) {
		//old extension already there, simply change the index
		oldFileName += basePath.left(oldTagPos+FileRepository::oldResource.size());
	} else {
		//add the current path and the old extension
		oldFileName += basePath + FileRepository::oldResource;
	}
	//add the index of the needed old file
	oldFileName += QString("%1").arg(oldIndex);
	
	return oldFileName;
}

//return the user name from the name of the page given in argument
QString BrowserWidget::getPageUserName(const QString& shortFileName) {
	QString cleanedName = shortFileName;
	if(cleanedName.startsWith("/")) cleanedName.remove(0, 1);
	
	if(!cleanedName.startsWith(FileRepository::userFolderName)) return QString("");
	
	//remove the userFolderName at the start
	cleanedName.remove(0, FileRepository::userFolderName.length());
	
	//remove everything after the first '/' (if there is any)
	int pos = cleanedName.indexOf("/");
	if(pos != -1) cleanedName = cleanedName.mid(pos+1);
	
	//remove contrib, log and old resources
	pos = cleanedName.lastIndexOf(FileRepository::userContribExtension);
	if(pos != -1) cleanedName = cleanedName.left(pos);
	pos = cleanedName.lastIndexOf(FileRepository::logResource);
	if(pos != -1) cleanedName = cleanedName.left(pos);
	pos = cleanedName.lastIndexOf(FileRepository::oldResource);
	if(pos != -1) cleanedName = cleanedName.left(pos);
	
	return cleanedName;
}

//return the list of files that are necessary to a given file
QStringList BrowserWidget::filesDependencyList(const QUrl& url) {
	QStringList neededfiles;
	
	QString askedPath = url.path();
	if(askedPath.startsWith("/")) askedPath.remove(0, 1);
	if(askedPath.endsWith(FileRepository::logResource)) {
		//nothing to do
		return neededfiles;
	}
	
	//in case of a diff, we need to download the old file with which we want to compare the text
	QString diffIndex = url.queryItemValue("diffWith");
	if(!diffIndex.isEmpty()) {
		bool ok=false;
		int oldIndex = diffIndex.toInt(&ok, 10);
		if(ok) {
			//address of the old file
			QString oldFileName = getDiffFileName(askedPath, oldIndex);
			neededfiles += QDir::cleanPath(oldFileName);
		}
	}
	
	//in case of a diff or an old file, add the log resource to the list of dependencies
	if(!diffIndex.isEmpty() || FileRepository::isOldFile(askedPath)) {
		QString nameWithoutOld;
		if(FileRepository::isOldFile(askedPath)) {
			nameWithoutOld = getTargetFileName(askedPath);
		} else {
			nameWithoutOld = askedPath;
		}
		
		QString logFileName = FileRepository::getLogFileName(nameWithoutOld);
		neededfiles += logFileName;
	}
	
	//read the text file
	QString localCacheName = FileRepository::getAbsoluteFileName(url.path());
	QFile htmlFile(localCacheName);
	htmlFile.open( QIODevice::ReadOnly );
	QTextStream text_stream( &htmlFile );
	text_stream.setCodec("UTF-8");
	QString wikiString = text_stream.readAll();
	htmlFile.close();
	
	//look for the [[Image:link]] tag
	//we don't need [[Image::link]] because :link is an address to an image bundled in the software
	QRegExp rxLang("\\[\\[Image:(?!:)([^]]*)\\]\\]");
	int pos = rxLang.indexIn(wikiString, 0);
	while (pos != -1) {
		int optionsIndex = rxLang.cap(1).indexOf('|');
		if(optionsIndex != -1) {
			//there are options, they are not a part of the file name
			neededfiles += rxLang.cap(1).left(optionsIndex);
		} else {
			neededfiles += rxLang.cap(1);
		}
		pos += rxLang.cap(0).size();
		pos = rxLang.indexIn(wikiString, pos);
	}
	
	return neededfiles;
}

//transform a string in the wiki format into a html standard format
//Transformations :
//add some <br/> line breaks
//[[a|b]] -> href tag
void BrowserWidget::formatToStandardHtml(QString & wikiString) {
	//special case of a * at the beginning of the first line
	if(wikiString.startsWith("*")) {
		wikiString.remove(0, 1);
		wikiString.insert(0, "<li>");
	}
	
	//An item of a list is created with the symbol * at the beginning of a line
	wikiString.replace(QRegExp("\\n\\*"), "<li>");
	
	//A list ends at the first endline that is not followed by a *
	wikiString.replace(QRegExp("(<li>[^\\n]*)"), "<ul>\\1</ul>");
	
/*
	//indent data when newline is followed by one or more ":"
	QRegExp rx("\\n\\:+[^\\n]*(?=\\n)");
	//QString str = "Offsets: 12 14 99 231 7";
	//QStringList newString;
	int pos = rx.indexIn(wikiString, 0);
	while (pos != -1) {
		std::cout << pos << std::endl;
		QString matched = rx.matchedLength();
		QString completeString;
		QString afterString;
		int i;
		for(i=1; i < rx.captureCount(); i++) {
			if(rx.cap(i)==QString(":")) {
				completeString += "<dl><dd>";
				afterString += "</dd></dl>";
			} else {
				completeString += rx.cap(i);
			}
		}
		completeString += afterString;
		
		wikiString.replace(pos, , completeString);
		
		pos++;
		pos = rx.indexIn(wikiString, pos);
	}
*/	
	//wikiString.replace(QRegExp("\\n\\:([^\\n]*)(?=\\n)"), "<dl><dd>\\1</dd></dl>");
	
	// === title === <h3> title </h3>
	// == title == -> <h2> title </h2>
	wikiString.replace(QRegExp("[\\s\\n]*===([^=]*)===[\\s\\n]*"), "<h3>\\1</h3>");
	wikiString.replace(QRegExp("[\\s\\n]*==([^=]*)==[\\s\\n]*"), "<h2>\\1</h2>");
	
	//replace two newlines by <br/><br/>, and one newline by <br/>,
	//except after a html tag detected by a > followed by some spaces or before a html tag detected by some spaces followed by a <
	// ([^>\\n]) means anything different from > and \n
	// \s* means zero or any number of white spaces
	wikiString.replace(QRegExp("([^>\\n])\\s*\\n\\n\\s*([^<\\n])"), "\\1<br /><br />\\2");
	wikiString.replace(QRegExp("([^>\\n])\\s*\\n\\s*([^<\\n])"), "\\1<br />\\2");
	
	//this is here that we detect images [[Image:link]]
	QRegExp rxLang("\\[\\[([^]:]*):(?!//)([^]]*)\\]\\]");
	int pos = rxLang.indexIn(wikiString, 0);
	while (pos != -1) {
		QString linkName = rxLang.cap(1);
		if(linkName=="Image") {
			//separate the options (with the | character)
			QStringList optionsList = rxLang.cap(2).split('|');
			QString targetImageName, targetImageFile;
			if(!optionsList.isEmpty()) targetImageName=optionsList.at(0);
			bool noneOption = false;
			if(optionsList.size()>=2) {
				if(optionsList.at(1)=="none") noneOption=true;
			}
			
			if(targetImageName.startsWith('/')) {
				//this is an absolute path
				targetImageFile = FileRepository::getAbsoluteFileName(targetImageName);
			} else
			if(targetImageName.startsWith(':')) {
				//case of an internal resource
				targetImageFile = targetImageName;
			} else {
				//this is a relative path
				//TODO
			}
			
			QString imageLink = QString("<img src=\"%1\">").arg(targetImageFile);
			//except for internal resource images, we add a link to the image file
			if(!targetImageName.startsWith(':')) {
				imageLink = linkString(targetImageName, imageLink);
			}
			
			if(noneOption) imageLink = QString("<p>%1</p>").arg(imageLink);
			wikiString.replace(pos, rxLang.cap(0).size(), imageLink);
		}
		pos++;
		pos = rxLang.indexIn(wikiString, pos);
	}
	
	//[[address]] -> <a href="address">address</a>
	// ([^]|:]*) means any characters except ] or | or :
	//we detect strings beginning by "http:" or that contains no ":"
	//we don't detect here langages like en: or fr:
	wikiString.replace(QRegExp("\\[\\[(http:|)([^]|:]*)\\]\\]"), "<a href=\"\\1\\2\">\\1\\2</a>");
	
	//[[address|description]] -> <a href="address">description</a>
	wikiString.replace(QRegExp("\\[\\[(http:|)([^]|:]*)\\|([^]|]*)\\]\\]"), "<a href=\"\\1\\2\">\\3</a>");
	
	//replace some particular strings by their translations
	for(int i=0; i<wikiString.size(); i++) {
		if(wikiString[i]=='#') {
			for(int j=0; j<specialStrings.size(); j++) {
				if(wikiString.mid(i, specialStrings[j].size()) == specialStrings[j]) {
					//we have found a special string, replace it by its translation
					wikiString.replace(i, specialStrings[j].size(), specialStringsTranslations[j]);
				}
			}
		}
	}
	
	//can be used to display the resulting html
	//std::cout << wikiString.toStdString() << std::endl;
}

void BrowserWidget::createLeftMenu(QString & wikiString){
	//find the langage tags [[lang:link]] and cut them
	//they will be put in the left cell of the table
	QString languageLinks;
	QRegExp rxLang("\\[\\[([^]:]*):(?!//)([^]]*)\\]\\]");
	int pos = rxLang.indexIn(wikiString, 0);
	while (pos != -1) {
		if(rxLang.cap(1)!="Image") {
			//lang tag
			QString langName = nativeNameOf(rxLang.cap(1));
			languageLinks+=QString("<a href=\"/%1/%2\">%3</a><br/>").arg(rxLang.cap(1)).arg(rxLang.cap(2)).arg(langName);
			wikiString.replace(pos, rxLang.cap(0).size(), QString(""));		//delete the link from the main text
		}
		pos++;
		pos = rxLang.indexIn(wikiString, pos);
	}
	addLeftMenu(wikiString, languageLinks);
}

//add the left menu to the string
//wikistring is modified by this function
void BrowserWidget::addLeftMenu(QString & wikiString, QString languageLinks, bool isErrorPage) {
	//-------------------------------------------------------------
	//main content inside a table cell on the right
	//language links in another table cell on the left
	QString horizontalLinks;
	QString usualLinks;
	QString currentPath = currentUrl.path();

	QUrl relativeUrl = currentUrl;
	QList<QPair<QString, QString> > noQueryItems;
	relativeUrl.setQueryItems(noQueryItems);
	QString linkTitle;
	QString pageTitle;
	bool licenceIsNeeded = true;
	if(isErrorPage) licenceIsNeeded = false;
	
	//remove the log, old or talk resources from the file name
	QString targetFileName = getTargetFileName(currentPath);
	
	//further remove the talk resource (special case of talk>log file)
	QString baseFileName = getTargetFileName(targetFileName);
			
	QString spaces("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
	
	if(!FileRepository::isVirtualPage(baseFileName) &&
	   !currentPath.endsWith(FileRepository::userContribExtension))
	{
		//link to the article page (except for the article page itself)
		linkTitle = tr("Article");
		if(currentPath != baseFileName && currentPath != QString("/") + baseFileName) {
			//link to the main page
			relativeUrl.setPath(QString("/") + baseFileName);
			horizontalLinks += linkString(relativeUrl.toString(), linkTitle);
		} else {
			horizontalLinks += linkTitle;
			int lastSlash = baseFileName.lastIndexOf('/');
			pageTitle = baseFileName.mid(lastSlash+1);
			pageTitle.replace('_', ' ');
		}
		horizontalLinks +=spaces;
		
		//link to the talk page (except for the talk page itself)
		linkTitle = tr("Talk");
		if(!currentPath.endsWith(FileRepository::talkResource)) {
			//link to the talk page
			relativeUrl.setPath(QString("/") + baseFileName + FileRepository::talkResource);
			horizontalLinks += linkString(relativeUrl.toString(), linkTitle);
		} else {
			horizontalLinks += linkTitle;
		}
		horizontalLinks +=spaces;
		
		//link to the history page (except for the history page itself)
		linkTitle = tr("Article history");
		if(currentPath.endsWith(FileRepository::talkResource+FileRepository::logResource) ||
			currentPath.endsWith(FileRepository::metaResource+FileRepository::logResource) ||
			!currentPath.endsWith(FileRepository::logResource)
		) {
			QUrl historyUrl = currentUrl;
			historyUrl.setQueryItems(noQueryItems);
			historyUrl.setPath( QDir::cleanPath(FileRepository::getLogFileName(baseFileName)) );
			horizontalLinks += linkString(historyUrl.toString(), linkTitle);
		} else {
			horizontalLinks += linkTitle;
		}
		horizontalLinks +=spaces;
		
		//link to the talk history only for the talk page and its old versions
		if(currentPath.endsWith(FileRepository::talkResource) ||
		   targetFileName.endsWith(FileRepository::talkResource)
		   ) {
			//title of the talk page
			if(currentPath.endsWith(FileRepository::talkResource)) {
				pageTitle = tr("Talk about %1").arg(baseFileName);
			}
			
			//link for the tak history, except for the talk history page itself
			linkTitle = tr("Talk history");
			if(!currentPath.endsWith(FileRepository::logResource)) {
				if(currentPath.startsWith("/")) currentPath.remove(0, 1);
				QUrl historyUrl = currentUrl;
				historyUrl.setQueryItems(noQueryItems);
				QString talkPageName = currentPath;
				if(!currentPath.endsWith(FileRepository::talkResource)) {
					talkPageName = targetFileName;
				}
				historyUrl.setPath( QDir::cleanPath(FileRepository::getLogFileName(talkPageName)) );
				horizontalLinks += linkString(historyUrl.toString(), linkTitle);
				pageTitle = tr("Talk about %1").arg(baseFileName);
			} else {
				horizontalLinks += linkTitle;
			}
			horizontalLinks +=spaces;
		}
		
		//link to the meta page (only for files with an extension, i.e. a '.' followed by up to 4 characters)
		linkTitle = tr("Meta");
		if(!currentPath.endsWith(FileRepository::metaResource) &&
				baseFileName.contains('.') &&
				baseFileName.lastIndexOf('.') > baseFileName.size()-6
		) {
			relativeUrl.setPath(QString("/") + baseFileName + FileRepository::metaResource);
			horizontalLinks += linkString(relativeUrl.toString(), linkTitle);
			horizontalLinks +=spaces;
		} else
		if(currentPath.endsWith(FileRepository::metaResource)) {
			horizontalLinks += linkTitle;
			horizontalLinks +=spaces;
		}
		
		//link to the meta history only for the meta page and its old versions
		//TODO : code very similar to the talk history page
		if(currentPath.endsWith(FileRepository::metaResource) ||
		   targetFileName.endsWith(FileRepository::metaResource)
		   ) {
			//title of the meta page
			if(currentPath.endsWith(FileRepository::metaResource)) {
				pageTitle = tr("Meta information of file %1").arg(baseFileName);
			}
			
			//link for the meta history, except for the meta history page itself
			linkTitle = tr("Meta history");
			if(!currentPath.endsWith(FileRepository::logResource)) {
				if(currentPath.startsWith("/")) currentPath.remove(0, 1);
				QUrl historyUrl = currentUrl;
				historyUrl.setQueryItems(noQueryItems);
				QString metaPageName = currentPath;
				if(!currentPath.endsWith(FileRepository::metaResource)) {
					metaPageName = targetFileName;
				}
				historyUrl.setPath( QDir::cleanPath(FileRepository::getLogFileName(metaPageName)) );
				horizontalLinks += linkString(historyUrl.toString(), linkTitle);
			} else {
				horizontalLinks += linkTitle;
			}
			horizontalLinks +=spaces;
		}
		
		//no licence for log pages
		if(currentPath.endsWith(FileRepository::logResource)) {
			licenceIsNeeded = false;
		}
	} else {
		licenceIsNeeded = false;
	}
	
	//title of old files
	int dotIndex = currentPath.lastIndexOf(FileRepository::resourceSeparator);
	if(dotIndex!=-1) {
		QString extension = currentPath.mid(dotIndex);
		if(extension.startsWith(FileRepository::oldResource)) {
			pageTitle = tr("Old revision of %1").arg(targetFileName);
		}
	}
	
	//flukz home page
	usualLinks += linkString(homePage(), tr("Home page")) + "<br/>";
	
	//recent changes url
	QUrl recentChangesUrl = currentUrl;
	recentChangesUrl.setQueryItems(noQueryItems);
	recentChangesUrl.setPath( QDir::cleanPath(FileRepository::recentChangesPageName) );
	usualLinks += linkString(recentChangesUrl.toString(), tr("Recent changes")) + "<br/>";
	
	//server status
	QUrl serverStatusUrl = currentUrl;
	serverStatusUrl.setQueryItems(noQueryItems);
	serverStatusUrl.setPath( FileRepository::statusPageName);
	usualLinks += linkString(serverStatusUrl.toString(), tr("Server status")) + "<br/>";
	
	//all pages special page
	QUrl allPagesUrl = currentUrl;
	allPagesUrl.setQueryItems(noQueryItems);
	allPagesUrl.setPath( FileRepository::listFilesPageName);
	usualLinks += linkString(allPagesUrl.toString(), tr("All pages list")) + "<br/>";
	
	//for user pages, add a link to the user contributions
	QString userName = getPageUserName(currentPath);
	if(!userName.isEmpty()) {
		QString userPage = QString("/") + FileRepository::userFolderName + userName + FileRepository::userContribExtension;
		usualLinks += "<br/>" + linkString(userPage, tr("User contributions")) + "<br/>";
	}
	
	//in case of debug, we also add a link to localhost and to the main server
	#ifdef DEBUG
		usualLinks += "<br/>" + linkString("fkz://localhost/en/Homepage", tr("Localhost")) + "<br/>";
		usualLinks += linkString( QString("fkz://%1/en/Homepage").arg(Connection::serverDefaultHostName), tr("Flukz server")) + "<br/>";
	#endif
	
	QString pageLicence = tr("Text is available under CC BY-SA Licence.");
	
	//add the usual links and the language links in a cell before the main content (so that it will be displayed on the left)
	if(!languageLinks.isEmpty()) languageLinks.insert(0, QString("<hr width=\"90%\">%1<br/>").arg(tr("Languages")));
	if(horizontalLinks.isEmpty()) {
		wikiString.insert(0, QString("<table><tr><td width=\"140\" bgcolor=\"#F3F8F4\" rowspan=\"3\">%1%2</td> <td>").arg(usualLinks).arg(languageLinks));
	} else
	if(pageTitle.isEmpty()) {
		wikiString.insert(0, QString("<table><tr><td width=\"140\" bgcolor=\"#F3F8F4\" rowspan=\"4\">%1%2</td> <td>%3<br/>&nbsp;</td></tr> <tr><td>").arg(usualLinks).arg(languageLinks).arg(horizontalLinks));
	} else {
		wikiString.insert(0, QString("<table><tr><td width=\"140\" bgcolor=\"#F3F8F4\" rowspan=\"5\">%1%2</td> <td>%3<br/>&nbsp;</td></tr> <tr><td><h1>%4</h1></td></tr> <tr><td><br/>").arg(usualLinks).arg(languageLinks).arg(horizontalLinks).arg(pageTitle));
	}
	
	wikiString.append("</td></tr>");
	if(licenceIsNeeded) {
		wikiString.append(QString("<tr><td>&nbsp;<br/>&nbsp;</td></tr><tr><td style=\"font-size:small; vertical-align:bottom\">%1</td></tr>").arg(pageLicence));
	}
	wikiString.append("</table>");
	
	//add the html header and table tag
	wikiString.insert(0, "<html><head></head><body><font size=4>");
	wikiString.append("</font></body></html>");
}

//create the html corresponding to a log file
//used to display a log file in a readable format
QString BrowserWidget::createHistoryHtml(const QString& localFileName, const QString& shortFileName) {
	QString targetFileName=getTargetFileName(shortFileName);
	
	//read the history file (text format) and create a corresponding html string for display
	QFile file(localFileName);
	file.open( QIODevice::ReadOnly );
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	QString resultingHtml;
	
	bool isRecent = false;
	if(shortFileName.endsWith(FileRepository::recentChangesPageName)) {
		isRecent = true;
		QString revisionTitle=tr("Recent changes");
		resultingHtml.append(QString("<h2>") + revisionTitle + QString("</h2>"));
	} else
	if(shortFileName.endsWith(FileRepository::userContribExtension)) {
		isRecent = true;
		QString revisionTitle=tr("Contributions of user");
		QString userName = getPageUserName(shortFileName);
		resultingHtml.append(QString("<h2>") + revisionTitle + QString(" %1</h2>").arg(userName));
	} else {
		QString revisionTitle=tr("Revision history of");
		resultingHtml.append(QString("<h2>") + revisionTitle + QString(" %1</h2>").arg(targetFileName));
	}
	
	//read header
	if (!text.atEnd() ) {
		QString headerName("[Flukz Revision File]");
		if (text.readLine()!=headerName) {
			file.close();
			return resultingHtml;
		}
	}
	
	//read string format version
	int stringFormatVersion=-1;
	if (!text.atEnd() ) {
		QString line = text.readLine();
		stringFormatVersion = line.toInt();
	}
	
	if(stringFormatVersion!=0) {
		//unsupported version
		file.close();
		return resultingHtml;
	}
	
	//read the plain name of the target file (not used)
	if (!text.atEnd() ) {
		QString targetPlainName = text.readLine();
	}
	
	//read the descriptions of the old versions
	resultingHtml.append("<p><ul>");
	int insertPosition = resultingHtml.size();
	int index;
	QString changedFileName, timeT, time, date, userName, comment;
	while ( !text.atEnd() ) {
		if(isRecent) {
			//in case of the recent changes file, the changed file is the first field
			text >> changedFileName;
			if(changedFileName.startsWith("/")) changedFileName.remove(0, 1);
		}
		text >> index;
		text >> timeT;
		text >> userName;
		comment = text.readLine();
		
		QDateTime dateTime = QDateTime::fromTime_t(timeT.toUInt());
		time = dateTime.toString("hh:mm");
		date = dateTime.toString("dd MMMM yyyy");
		
		QString oneChangeHtml;
		
		if(isRecent) {
			oneChangeHtml.append("<li>");
			if(index>0) {
				oneChangeHtml.append(QString("<a href=\"/%1%2%3?diffWith=%4\">%5</a>&nbsp;&nbsp;&nbsp;").arg(changedFileName).arg(FileRepository::oldResource).arg(index).arg(index-1).arg(tr("diff")));
			} else {
				oneChangeHtml.append("new&nbsp;");
			}
			oneChangeHtml.append(QString("<a href=\"/%1\">%1</a>&nbsp;&nbsp;%2&nbsp;&nbsp;&nbsp;").arg(changedFileName).arg(date + ", " + time));
		} else {
			oneChangeHtml.append("<li>");
			if(text.atEnd()) {
				//current version
				//link for diff
				if(index>0) {
					oneChangeHtml.append(QString("<a href=\"/%1?diffWith=%2\">%3</a>&nbsp;&nbsp;&nbsp;").arg(targetFileName).arg(index-1).arg(tr("diff")));
				}
				
				//link to the current file
				oneChangeHtml.append(QString("<a href=\"/%1\">%2</a>&nbsp;&nbsp;&nbsp;").arg(targetFileName).arg(date + ", " + time));
			} else {
				//old version
				//construct the link for viewing the diff
				QString oldName = getOldFileName(targetFileName, index);
				if(index>0) {
					oneChangeHtml.append(QString("<a href=\"%1?diffWith=%2\">%3</a>&nbsp;&nbsp;&nbsp;").arg(oldName).arg(index-1).arg(tr("diff")));
				}
				
				//construct the link to the old file in the attic folder
				oneChangeHtml.append(QString("<a href=\"%1\">%2</a>&nbsp;&nbsp;&nbsp;").arg(oldName).arg(date + ", " + time));
			}
		}
		oneChangeHtml.append(QString(" <b><a href=\"/%1%2\">%2</a></b> ").arg(FileRepository::userFolderName).arg(userName));
		oneChangeHtml.append(comment);
		
		//insert the change at the correct position ( in order to obtain recent changes at the top)
		resultingHtml.insert(insertPosition, oneChangeHtml);
	}
	
	file.close();
	resultingHtml.append("</ul></p>");
	
	//add the left menu
	QString languageLinks;	//no language links
	addLeftMenu(resultingHtml, languageLinks);
	
	return resultingHtml;
}

//create the html corresponding to the difference between two files
QString BrowserWidget::createDiffHtml(const QString& oldFileName, const QString& newFileName) {
	QString result;
	
	//name of the more recent file without the old extension (if any)
	QString shortFileName = currentUrl.path();	//TODO : currentUrl should not be used
	QString targetFileName;
	if(FileRepository::isOldFile(shortFileName)) {
		targetFileName = getTargetFileName(shortFileName);
	} else {
		targetFileName = shortFileName;
		if(targetFileName.startsWith("/")) {
			targetFileName.remove(0, 1);
		}
	}
	
	//name of the older file
	int oldRevision = currentUrl.queryItemValue("diffWith").toInt();
	QString oldFileLink = getOldFileName(targetFileName, oldRevision);
	
	//html content for the old and the new file
	QString newFileContent, oldFileContent;
	if(targetFileName.endsWith(".png") || targetFileName.endsWith(".jpg")) {
		//case of two images, just display the images
		newFileContent = QString("[[Image:") + shortFileName + "]]\n";
		formatToStandardHtml(newFileContent);
		oldFileContent = QString("[[Image:") + oldFileLink + "]]\n";
		formatToStandardHtml(oldFileContent);
	} else {
		//case of usual wiki content, find the differences
		
		//read the content of the two files
		QFile newFile(newFileName);
		newFile.open( QIODevice::ReadOnly );
		QTextStream textStream( &newFile );
		textStream.setCodec("UTF-8");
		newFileContent = textStream.readAll();
		newFile.close();
		
		QFile oldFile(oldFileName);
		oldFile.open( QIODevice::ReadOnly );
		QTextStream textStreamB( &oldFile );
		textStreamB.setCodec("UTF-8");
		oldFileContent = textStreamB.readAll();
		oldFile.close();
		
		//find the differences between the two strings
		//and modifies them in order to display the diff parts with colors
		DiffTool::diffHtml(oldFileContent, newFileContent);
	}
	
	//display the content in a table
	QString titleString = tr("<a href=\"%1\">Revision %2</a> of %3");
	
	QString oldFileTitle= titleString.arg(oldFileLink).arg(oldRevision).arg(targetFileName);
	
	int newRevision = oldFileRevision(shortFileName);
	QString newFileTitle;
	if(newRevision==-1) {
		//case of the latest revision of the file
		newFileTitle = tr("<a href=\"%1\">Latest Revision</a> of %2").arg(QString("/")+targetFileName).arg(targetFileName);
	} else {
		QString newFileLink = getOldFileName(targetFileName, newRevision);
		newFileTitle = titleString.arg(newFileLink).arg(newRevision).arg(targetFileName);
	}
	
	//revert links
	QString revertString = tr("<a href=\"%1?revert=%2\">Revert to this revision</a>");
	QString oldFileRevert = revertString.arg(QString("/")+targetFileName).arg(oldRevision);
	
	result = QString("<table><tr>");
	result += QString("<td width=\"50%\"><b>%1</b><br /></td>").arg(oldFileTitle);
	result += QString("<td width=\"50%\"><b>%1</b><br /></td>").arg(newFileTitle);
	result += QString("</tr><tr>");
	result += QString("<td width=\"50%\">%1<br /></td>").arg(oldFileRevert);
	result += QString("<td width=\"50%\"> <br /></td>");
	result += QString("</tr><tr>");
	result += QString("<td width=\"50%\">%1</td>").arg(oldFileContent);
	result += QString("<td width=\"50%\">%1</td>").arg(newFileContent);
	result += QString("</tr></table>");
	
	//add the left menu
	QString languageLinks;	//no language links
	addLeftMenu(result, languageLinks);
	
	return result;
}

int BrowserWidget::oldFileRevision(const QString& shortFileName) {
	QString targetFileName=shortFileName;
	
	//remove the initial "/" if any
	if(targetFileName.startsWith("/")) targetFileName.remove(0, 1);
	
	//find the last old resource
	int dotIndex = targetFileName.lastIndexOf(FileRepository::oldResource);
	
	//there is no old resource, this is the latest revision
	if(dotIndex==-1) return dotIndex;
	
	return targetFileName.mid(dotIndex+FileRepository::oldResource.size()).toInt();
}

//=============== Game credits ====================

QString BrowserWidget::createGameCredits() {
	//we create a list of the files used in the game, with the associated meta information
	//the list is created in the wiki syntax (it will be transformed later in html)
	QString result, soundsCredits, graphicsCredits;
	result += QString("Credits of the file used in this level");
	for(int i=0; i<gameCreditsList.size(); i++) {
		QString shortFileName = gameCreditsList.at(i);
		QString fileCredits;
		if(!shortFileName.endsWith(FileRepository::metaResource)) {
			//file title
			fileCredits += "===[[/" + shortFileName + "|" + shortFileName + "]]===\n";
			
			//display the meta information of the file
			QString metaFileName = FileRepository::getAbsoluteFileName(shortFileName + FileRepository::metaResource);
			QFile file(metaFileName);
			file.open( QIODevice::ReadOnly );
			QTextStream text_stream( &file );
			text_stream.setCodec("UTF-8");
			fileCredits += text_stream.readAll();
			file.close();
			fileCredits += "\n";
			
			//display the image in case of image files
			//TODO : button to play the sound for sound files in the future
			if(shortFileName.endsWith(".png")) {
				fileCredits += "[[Image:/" + shortFileName + "]]\n";
			}
			
			if(shortFileName.endsWith(".wav") || shortFileName.endsWith(".ogg")) {
				soundsCredits += fileCredits;
			} else {
				graphicsCredits += fileCredits;
			}
		}
	}
	
	if(!soundsCredits.isEmpty()) {
		result += "==Music and sounds==\n" + soundsCredits;
	}
	if(!graphicsCredits.isEmpty()) {
		result += "==Images==\n" + graphicsCredits;
	}
	return result;
}

//=============== Language names ===================

//Read the file containing the language names
void BrowserWidget::readLanguageFile(QString const& shortFileName) {
	//open file
	QString filename = FileRepository::getAbsoluteFileName(shortFileName);
	QFile file( filename );
	if ( !file.open( QIODevice::ReadOnly ) ) return;
	
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	//read header
	if (!text.atEnd() ) {
		QString headerName("[Flukz Language File]");
		if (text.readLine()!=headerName) {
			//incorrect format
			file.close();
			return;
		}
	}
	
	//read string format version
	int stringFormatVersion=-1;
	if (!text.atEnd() ) {
		QString line = text.readLine();
		stringFormatVersion = line.toInt();
	}
	
	if(stringFormatVersion!=0) {
		//unsupported version
		file.close();
		return;
	}
	
	//read the language codes and associated full names
	QString languageCode;
	QString languageName;
	while(!text.atEnd() ) {
		text >> languageCode;
		languageName = text.readLine();
		
		if(!languageCode.isEmpty() && !languageName.isEmpty()) {
			languageMap.insert(languageCode, languageName);
		}
	}
	
	//close file
	file.close();
}

//return the native name of a language code if found in the internal map
//otherwise, return the language code
QString BrowserWidget::nativeNameOf(const QString& languageCode) {
	QMap<QString, QString>::iterator languageEntry = languageMap.find(languageCode);
	if(languageEntry != languageMap.end()) {
		return languageEntry.value();
	}
	return languageCode;
}
