/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// test
#include "gamewidget.h"
#include "../soundengine/soundengine.h"
#include "../networkengine/filerepository.h"
#include "../networkengine/highscorefile.h"
#include "../tools/quicktimer.h"
#include "../tools/parameterwidget.h"
#include "../tools/simpledialogwidget.h"

#include <QPixmap>
#include <QLabel>
#include <QPushButton>
#include <QGraphicsProxyWidget>
#include <QToolBar>
#include <QAction>
#include <QMessageBox>
#include <QSignalMapper>
#include <QTimer>
#include <QtCore/qmath.h>

#ifdef USE_OPENGL_RENDERING
	#include <QGLWidget>
#endif

#if QT_VERSION >= 0x040700
	#include <QElapsedTimer>
#endif

//maximum number of displayed items in the sprite bar
int GameWidget::spriteBarMaxDisplayed = 17;

GameWidget::GameWidget( QWidget *parent) : QGraphicsView( parent ) {
	//dimension of the scene
	Global::gameViewX = -300;
	Global::gameViewY = -300;
	Global::gameViewW = 600;
	Global::gameViewH = 600;
	
	level = new Level;		//level object to store the data of a given level
	
	//create the scenes for the game
	createGameScene();		//displayed when the user plays
	createEditScene();		//displayed when editing new levels
	createMenuScene();		//menu scene at the start of the program
	isEditMode = false;
	isTestMode = false;
	
	//settings of the graphicsview
	setScene(menuScene);
	setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);
	setRenderHint(QPainter::Antialiasing);
	setBackgroundBrush(Qt::black);
	
	#ifdef USE_OPENGL_RENDERING
		setViewport(new QGLWidget(QGLFormat(QGL::SampleBuffers | QGL::DirectRendering)));
		setViewportUpdateMode(QGraphicsView::FullViewportUpdate);
	#else
		setViewportUpdateMode(QGraphicsView::BoundingRectViewportUpdate);
	#endif
	
	setCacheMode(QGraphicsView::CacheBackground);
	
	setDragMode(QGraphicsView::ScrollHandDrag);
	resize(610, 610);
	
	//timer to animate the game scene
	gameTimer = new QuickTimer(this);
	connect(gameTimer, SIGNAL(timeout()), this, SLOT(levelAdvance()));
	
	//timer to animate the high score
	highScoreTimer = new QuickTimer(this);
	connect(highScoreTimer, SIGNAL(timeout()), this, SLOT(updateHighScoreText()));
	
	#if QT_VERSION >= 0x040700
	//measure the elapsed time more precisely than the timers
	elapsedTimer = new QElapsedTimer;
	#endif
	
	//initialize the status for playing the sounds or not
	currentSoundStatus = AllSounds;
	
	//init the default key map
	keyMap.insert(Qt::Key_Up,     Qt::Key_Up );
	keyMap.insert(Qt::Key_Down,   Qt::Key_Down );
	keyMap.insert(Qt::Key_Left,   Qt::Key_Left );
	keyMap.insert(Qt::Key_Right,  Qt::Key_Right );
	keyMap.insert(Qt::Key_Space,  Qt::Key_Space );
	keyMap.insert(Qt::Key_Z,      Qt::Key_Space);	//Z key can be used to shoot (default in many modern shmups)
	keyMap.insert(Qt::Key_Delete, Qt::Key_Delete);
	keyMap.insert(Qt::Key_Escape, Qt::Key_Escape);
	keyMap.insert(Qt::Key_Return, Qt::Key_Return);
	keyMap.insert(Qt::Key_M, Qt::Key_M);			//used to stop/start the music
}

void GameWidget::setLevel(const QString& levelFileName) {
	currentLevelName = levelFileName;
	QString absoluteFileName = FileRepository::getAbsoluteFileName(levelFileName);
	level->readFile(absoluteFileName);
	
	//itialize the startLevelName
	//this step is done here instead of startGame() because startLevelName must be initialized before calling filesDependencyList()
	if(currentLevelName != expectedNextLevelName) {
		//remember the name of this start level (all scores wil be saved in the node of the start level)
		startLevelName = currentLevelName;
	}
	
	//used to know in Sprite class which pixmaps corresponds to which sprite
	Sprite::currentLevel = level;
}

QStringList GameWidget::filesDependencyList(bool isCreditsDependencies) {
	QStringList result = level->filesDependencyList(isCreditsDependencies);
	
	//add the file for highscores if we play a game (not needed for credits)
	if(!isCreditsDependencies) {
		result += FileRepository::getHighScoreFileName(startLevelName);
	}
	
	return result;
}

//key events are transfered to the sprite of the player
void GameWidget::keyPressEvent(QKeyEvent *event) {
	if( event->isAutoRepeat() || !keyMap.contains(event->key()) ) {
		event->ignore();
		return;
	}
	
	int pressedKey = keyMap[event->key()];
	
	if(playerSprite!=0) {
		if( playerSprite->treatKey(pressedKey, true) ) {
			event->accept();
			return;
		}
	}
	
	if(!isEditMode && gameStatus!=GameRunning) {
		if(pressedKey==Qt::Key_Down || pressedKey==Qt::Key_Up) {
			if(playButton->hasFocus()) {
				quitButton->setFocus();
				event->accept(); return;
			} else
			if(playButton->isVisible()) {
				//note : if no button has the focus, give it to play button
				 playButton->setFocus();
				event->accept(); return;
			}
		}
		
		if(pressedKey==Qt::Key_Space || pressedKey==Qt::Key_Return) {
			if(playButton->hasFocus()) {
				playButton->click();
				event->accept(); return;
			} else
			if(quitButton->hasFocus()) {
				quitButton->click();
				event->accept(); return;
			}
		}
	}
	
	event->ignore();
}

//key events are transfered to the sprite of the player
//except the "delete" key in edit mode that is used to remove a sprite
void GameWidget::keyReleaseEvent(QKeyEvent *event) {
	if( event->isAutoRepeat() || !keyMap.contains(event->key()) ) {
		event->ignore();
		return;
	}
	
	int pressedKey = keyMap[event->key()];
	if(isEditMode) {
		if(pressedKey==Qt::Key_Delete) {
			//delete the currently selected sprite (in edit mode only)
			removeCurrentSprite();
			event->accept(); return;
		} else
		if(pressedKey==Qt::Key_Escape) {
			//similar to a click on the hand in the sprite bar
			emit editModeActivateHand();
			event->accept(); return;
		}
	}
	
	//case of the M key to stop/start the music
	if(pressedKey==Qt::Key_M && gameStatus==GameRunning) {
		changeSoundStatus();
		event->accept(); return;
	}
	
	if(playerSprite!=0) {
		if( playerSprite->treatKey(pressedKey, false) ) {
			event->accept(); return;
		}
	}
	
	event->ignore();
}

void GameWidget::setBackgroundColor(const std::string& color) {
	QString backColor = QString::fromStdString(color);
	
	//note : QColor::isValidColor() avoided to keep compatibility with Qt 4.6
	QColor groundColor;
	groundColor.setNamedColor(backColor);
	if(groundColor.isValid()) {
		setBackgroundBrush(groundColor);
	} else {
		setBackgroundBrush(Qt::black);
	}
}

void GameWidget::changeSoundStatus() {
	//switch the current status to the next one of the list
	switch(currentSoundStatus) {
		case AllSounds :
			//in this version, we don't try to stop only the music (soundengine not ready for that)
			//currentSoundStatus=NoMusic;
			//TODO : stop the music
			//break;
		case NoMusic:
			currentSoundStatus=NoSound; 
			soundEngine->stopAllSounds();		//stop all the remaining sounds
			//note : we don't stop completely the audio device with enableSound(false)
			//   because this operation is very slow
			break;
		case NoSound : default :
			currentSoundStatus=AllSounds;
			soundEngine->enableSound(true);
			playSound(0, -1);					//start to play the music, with infinite loop
	}
}

//================================ Game mode ====================================================

void GameWidget::createGameScene() {
	gameScene = new QGraphicsScene;
	gameScene->setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);
	gameScene->setItemIndexMethod(QGraphicsScene::NoIndex);
	playerSprite = 0;
	gameStatus = GameOver;
	createScoreSprite();
	createEnergyBarItem();
}

//start the game, called when a level has been selected in the browser interface
void GameWidget::startGame() {
	deleteAllSprites();		//to be sure that there is no remaining sprite of the previous level
	stopAllTimers();		//especially to stop the timer of the highscore
	setMouseShape("");		//reinitialize the mouse cursor to default
	
	//itialize the score
	if(currentLevelName != expectedNextLevelName) {
		//this is a new level, not related to the previous one
		//we reinitialize the expected next level name to avoid any possibility of "cheating"
		expectedNextLevelName.clear();
		setScore(0);
		previousLevelScore = 0;
	} else {
		//this is the follow-up of a previous level
		//the starting score is the score at the end of the previous level
		setScore(previousLevelScore);
	}
	
	level->init();											//initialize some variables inside the level
	soundEngine->loadWavFiles(level->getSoundFileList());	//load the wav files
	Sprite::initKeyboardState();							//ensure all keys are in "released" state
	
	playerSprite = 0;
	gameStatus = GameRunning;
	isEditMode = false;
	setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);	//size of the viewport
	setBackgroundColor(level->levelParam.backgroundColor);
	setScene(gameScene);			//display the game Scene
	setFocus();						//give the focus to the game widget (needed for test mode)
	resize(610, 610);				//resize the window to avoid scrollbars
	if(currentSoundStatus!=NoSound) {
		//activate the sounds on the audio device
		soundEngine->enableSound(true);
	}
	playSound(0, -1);				//start to play the music, with infinite loop
	
	#if QT_VERSION >= 0x040700
	elapsedTimer->start();			//start the precise measure of the elapsed time (note : no stop needed)
	#endif
	synchronizedElapsedTime = 0;
	
	totalUpdateSteps = 0;			//count the number of steps to sychronize some positions to integer values
	
	gameTimer->start(10); 			//start the timer to animate the game Scene (100 frames/s)
}

//start or stop the game test during edition
void GameWidget::trigGameTest() {
	if(isTestMode) {
		//if we were already testing, stop the test mode and go back to the edit mode
		stopGame();
		startEdit();
	} else {
		//otherwise start the test mode
		//stop the add sprite mode without destroying the sprites stored in the edit scene
		addSpriteMode=false;					//stop the addSpriteMode
		if(spriteBar!=0) spriteBar->hide();		//hide the sprite bar to select a sprite
		parameterWidget->setEnabled(false);		//no change of the parameters during the test
		
		//find the coordinates of the currently displayed part of the scene
		//(because we want to start the test at the displayed scene)
		QPoint viewOrigin(0, 0);
		QPointF displayedSceneTopLeft = mapToScene(viewOrigin);
		qreal partialLevelStartY = displayedSceneTopLeft.y();
		
		//read the level from the curent edit scene
		QList<QGraphicsItem*> itemList = editScene->items();
		level->createPartialLevelFromScene(itemList, partialLevelStartY);
		Sprite::currentLevel = level;
		isTestMode = true;
		startGame();
	}
}

//stop the game
//note : this function does nothing if game is not running
//gameStatus must be changed only after stopGame() is executed 
void GameWidget::stopGame() {
	if(gameStatus!=GameRunning) return;
	
	gameTimer->stop();					//stop the timer that animates the game Scene
	soundEngine->stopAllSounds();		//stop all the remaining sounds
	soundEngine->enableSound(false);	//stop completely the audio device (no remaining sound on the speakers)
	playerSprite = 0;
	
	//delete all the sprites of the game scene
	//note : it is possible that some sprites are created after this call
	//       so this function should be called again at the start of a new level
	deleteAllSprites();
}

//convenient function to ensure that all timers are stopped
//must be called before making the widget invisible
void GameWidget::stopAllTimers() {
	gameTimer->stop();
	highScoreTimer->stop();
}

void GameWidget::levelAdvance() {
	#if QT_VERSION >= 0x040700
		//number of steps (of 10ms) than should be executed to keep the synchronization with the absolute time
		Sprite::stepsToExecute = (elapsedTimer->elapsed() - synchronizedElapsedTime)/10;
		
		//keep a maximum 1% time late
		if(Sprite::stepsToExecute > totalUpdateSteps/100) {
			Sprite::stepsToExecute = 2;
		} else {
			Sprite::stepsToExecute = 1;
		}
	#else
		//before Qt 4.7 version, we cannot use the elapsedTimer synchronization
		//so we execute just one step after each timer tick
		Sprite::stepsToExecute = 1;
	#endif
	
	if(Sprite::stepsToExecute>0) {
		readLevel();			//create the next sprites of the level
		gameScene->advance();	//update all the sprites position
		totalUpdateSteps+=Sprite::stepsToExecute;
	}
	
	//update the time until which we have synchronized the sprites
	//note: if the number of steps is negative, time will be sustracted
	synchronizedElapsedTime += Sprite::stepsToExecute * 10;
}

//display the results after a level is finished
void GameWidget::displayResult() {
	if(isTestMode) {
		//go back to the edit mode
		emit gameTestFinished();
		startEdit();
	} else
	if(gameStatus == LevelComplete && level->hasNextLevel()) {
		//launch automatically the next level
		expectedNextLevelName = level->nextLevelName();
		currentLevelName = expectedNextLevelName;	//minor hack to display the correct message
		displayLoadingMessage();
		previousLevelScore = levelScore;
		emit askPage(level->nextLevelName());
	} else {
		//usual game mode, display the results
		updateResultText();
		updateScoreText();
		labelHighScore->setVisible(false);	//not visible until the updated high scores are downloaded
		setBackgroundBrush(Qt::black);
		setScene(menuScene);	//display the menu Scene
		
		//check if it is a new high record
		QString scoreFileName = FileRepository::getHighScoreFileName(startLevelName);
		QString absoluteFileName = FileRepository::getAbsoluteFileName(scoreFileName, FileRepository::FileReadOnly);
		HighScoreFile scoreFile(absoluteFileName);
		if(scoreFile.isNewHighScore(levelScore)) {
			//send a message to the server indicating the current score
			emit askScoreCommit(startLevelName, levelScore);
		} else {
			//directly ask the current high scores
			emit askPage(scoreFileName);
		}
	}
}

//-------------- score item -------------------------
//create the sprite with the score 
void GameWidget::createScoreSprite() {
	scoreItem = new QGraphicsSimpleTextItem;
	QFont scoreFont("Times", 16, QFont::Bold);
	scoreItem->setFont(scoreFont);
	scoreItem->setZValue(100);			//score is above all other sprites
	scoreItem->setBrush(Qt::yellow);
	setScore(0);						//initialize the score value
	gameScene->addItem(scoreItem);
}

//directly set the value of the score (mainly used for initialization)
void GameWidget::setScore(int score) {
	levelScore = score;
	updateScoreLabel();
}

//change the value of the score with some given variation
void GameWidget::changeScore(int scoreVariation) {
	levelScore += scoreVariation;
	updateScoreLabel();
}

//update the display of the score on the screen
void GameWidget::updateScoreLabel() {
	scoreItem->setText(QString("%1").arg(levelScore));
	
	//update the position of the label
	qreal width = scoreItem->boundingRect().width();
	scoreItem->setPos(Global::gameViewX+Global::gameViewW - width - 10, Global::gameViewY + 13);
}

//-------------- energy bar item -------------------------
void GameWidget::createEnergyBarItem() {
	//yellow part
	energyBarItem = new QGraphicsRectItem;
	energyBarItem->setZValue(100);		//above all other sprites
	energyBarItem->setBrush(QBrush(Qt::yellow));
	gameScene->addItem(energyBarItem);
	
	//gray part
	energyBarGrayItem = new QGraphicsRectItem;
	energyBarGrayItem->setZValue(100);		//above all other sprites
	energyBarGrayItem->setBrush(QBrush(Qt::gray));
	gameScene->addItem(energyBarGrayItem);
}

void GameWidget::setEnergyBarFull(int fullValue) {
	energyBarFullPoints = fullValue;
	setEnergyBarCurrent(fullValue);
	//energyBarItem->setRect(Global::gameViewX+Global::gameViewW - 150 - 2, Global::gameViewY + 2, 150, 8);
	//energyBarGrayItem->setRect(Global::gameViewX+Global::gameViewW - 150 - 2, Global::gameViewY + 2, 0, 8);
}

void GameWidget::setEnergyBarCurrent(int currentValue) {
	energyBarCurrentPoints = currentValue;
	if(energyBarCurrentPoints == energyBarFullPoints) {
		//full energy bar in blue
		energyBarItem->setBrush(QBrush(Qt::cyan));
	} else
	if(energyBarCurrentPoints == 1) {
		//almost dead bar in red
		energyBarItem->setBrush(QBrush(Qt::red));
	} else {
		//default case in yellow
		energyBarItem->setBrush(QBrush(Qt::yellow));
	}
	
	qreal yellowpart = 150.0 * (qreal) energyBarCurrentPoints / (qreal) energyBarFullPoints;
	qreal graypart = 150.0 - yellowpart;
	energyBarItem->setRect(Global::gameViewX+Global::gameViewW - 150 - 2 + graypart, Global::gameViewY + 2, yellowpart, 8);
	energyBarGrayItem->setRect(Global::gameViewX+Global::gameViewW - 150 - 2, Global::gameViewY + 2, graypart, 8);
}

//------------------- sprite items ----------------------
//delete all the sprites of the scene
//except the static items like the score and the energy bar
void GameWidget::deleteAllSprites() {
	QList<QGraphicsItem*> list = gameScene->items();
	int i;
	for (i = 0; i < list.size(); i++) {
		if(list.at(i) != (QGraphicsItem *) scoreItem &&
		   list.at(i) != (QGraphicsItem *) energyBarItem &&
		   list.at(i) != (QGraphicsItem *) energyBarGrayItem)
		{
			//the sprite will be removed from the scene and deleted from memory
			((Sprite *) list.at(i))->deleteLater();
		}
	}
}

//called by a sprite when he wants to launch fire
void GameWidget::createFireSprite(SpriteParameter paramFire) {
	Sprite* fireSprite = new Sprite;
	
	//apply the level scrolling speed to fire only if it is in the same direction as the scrolling
	bool fireSign = (paramFire.y_speed > 0);
	bool scrollingSign = (level->levelParam.scrollingSpeed > 0);
	if(fireSign == scrollingSign) {
		paramFire.y_speed += level->levelParam.scrollingSpeed;
	}
	fireSprite->setParameters(paramFire);
	connect(fireSprite, SIGNAL(destroy()), this, SLOT(destroySprite()));
	gameScene->addItem(fireSprite);
}

//called by a sprite when he wants to be destroyed
//sprites decide themselves if they should be destroyed or not
//(for example when they go out of the screen)
void GameWidget::destroySprite() {
	QObject * senderSprite = sender();
	if(senderSprite == (QObject *) playerSprite) {
		//the player sprite has been destroyed -> game over
		senderSprite->deleteLater();
		playerSprite = 0;
		stopGame();
		gameStatus = GameOver;	//must be set after stopGame
		displayResult();
	} else {
		//delete the sprite from memory
		senderSprite->deleteLater(); //the sprite will be automatically removed from the scene
	}
}

//main function when playing a level
//this function is called at regular intervals and add progressively the sprites defined in the level
void GameWidget::readLevel() {
	if(gameStatus!=GameRunning) return;		//discard remaining calls after the game is stopped
	
	level->y_front = level->y_front - Sprite::stepsToExecute * (level->levelParam.scrollingSpeed);
	bool endLoop = false;
	
	//display the sprite slightly before they are touched by the level front
	//this is mainly to obtain perfect transitions for large background images
	qreal displayMargin = 2.0;
	
	while(level->currentSprite!=level->spriteMap.end() && !endLoop) {
		//compute the sprite height
		int currentSpriteKind = level->currentSprite.value().spriteShape;
		
		if(-level->currentSprite.key().y > level->y_front - displayMargin) {
			//create the sprite pointed by currentSprite
			Sprite * sprite = new Sprite;
			SpriteParameter p = level->currentSprite.value();
			
			#ifdef USE_OPENGL_RENDERING
				sprite->setCacheMode(QGraphicsItem::DeviceCoordinateCache);
			#endif
			
			//we floor the level front to ensure that the position is still an integer at this point
			p.y_pos = Global::gameViewY + p.y_pos - qFloor(level->y_front) - displayMargin;
			
			p.y_speed += level->levelParam.scrollingSpeed;
			
			//correct slightly the position
			//  so that all sprites with the same speed have a position with the same fractional part
			if(currentSpriteKind!=0) {
				qreal currentPositionSameSpeed = totalUpdateSteps * (p.y_speed);
				qreal positionSynchro = currentPositionSameSpeed - qFloor(currentPositionSameSpeed);
				p.y_pos += positionSynchro;
			}
			
			sprite->setParameters(p);
			gameScene->addItem(sprite);
			connect(sprite, SIGNAL(fire(SpriteParameter)), this, SLOT(createFireSprite(SpriteParameter)) );
			connect(sprite, SIGNAL(destroy()), this, SLOT(destroySprite()));
			connect(sprite, SIGNAL(askSound(int)), this, SLOT(playSound(int)));
			connect(sprite, SIGNAL(changeScore(int)), this, SLOT(changeScore(int)));
			
			//if this is the player sprite, remember the address (used by keyboard functions)
			if(currentSpriteKind==0) {
				playerSprite = sprite;
				setEnergyBarFull(sprite->getLifePoints());
				connect(sprite, SIGNAL(lifePointsChanged(int)), this, SLOT(setEnergyBarCurrent(int)));
			}
			//point to the next sprite
			level->currentSprite++;
		} else {
			endLoop = true;
		}
	}
	
	//stop the game when the level has been entirely displayed
	//the length of the level is not the same in real game mode and in test mode
	if( (isTestMode && level->y_front < level->level_execute_end) ||
		(!isTestMode && level->y_front < level->levelParam.y_end)
	  ) {
		stopGame();
		
		//game win only if it is not already over
		//(case of level completed during the destroy animation of the player ship)
		if(gameStatus!=GameOver) {
			//note we don't win if the game is already over
			gameStatus = LevelComplete;	//must be set after stopGame
		}
		displayResult();
	}
}

//play a sound given by its index in the list of sound files
//connected to Sprite::askSound
void GameWidget::playSound(int index, int repeatTimes /*=0*/) {
	//case where the user has stopped the sounds
	if(currentSoundStatus==NoSound) return;
	
	//index for the music is always 0
	//TODO : a more robust way to distinguish music would be better
	if(index==0 && currentSoundStatus==NoMusic) return;
	
	soundEngine->playSound(index, repeatTimes); //play the file of given index, with no repetition
}

//================================ Editor mode ====================================================

void GameWidget::createEditScene() {
	//---------- main edit scene --------------------------
	//note : we use the default index method instead of QGraphicsScene::NoIndex in gameScene, because of the biggest number of items
	editScene = new QGraphicsScene;
	editScene->setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);
	connect(editScene, SIGNAL(selectionChanged()), this, SLOT(editSelectionChanged()) );
	
	addSpriteMode = false;
	
	parameterWidget = new ParameterWidget;
	parameterWidgetMode = UninitializedMode;
	setParameterMode(SpriteParameterMode);
	
	//user changes in the parameterWidget are written in the currently selected sprite
	connect(parameterWidget, SIGNAL(contentChanged(std::string)), this, SLOT(parameterChanged(std::string)));
	
	parameterWidget->hide();
	
	spriteBar = 0;
	spriteBarMapper = 0;
}

void GameWidget::createSpriteBar() {
	//delete old objects before creating new ones
	if(spriteBar!=0) spriteBar->deleteLater();
	if(spriteBarMapper!=0) spriteBarMapper->deleteLater();
	
	//------------ Toolbar with the different sprites ----------
	spriteBar = new QToolBar("Sprites", this);
	spriteBar->hide();				//hide the sprite bar
	spriteBar->setStyleSheet("QToolBar { background: lightgray; }");
	
	spriteBarGroup = new QActionGroup(spriteBar);	//action group ensures exclusivity selection
	spriteBarMapper=new QSignalMapper(this);
	connect( spriteBarMapper, SIGNAL(mapped(int)), this, SLOT(spriteBarClicked(int)) );
	
	addSpriteAction(":/images/openhand.png", tr("Select and Move"), -1, /*isChecked*/ true);
	spriteBarItemsNumber=1;
	
	int i;
	for(i=0; i < (int) level->levelParam.spriteKinds.size(); i++) {
		QString pixmapName = QString::fromStdString(level->levelParam.spriteKinds[i].pixmapFileName);
		QString description;
		if(i==0) {
			description = QString(tr("Player sprite"));
		} else {
			description = QString(tr("Add: %1")).arg(pixmapName);	//TODO : there should be a description in spritekindparameter
		}
		int spriteIndex = level->levelParam.spriteKinds[i].spriteKind;
		addSpriteAction(pixmapName, description, spriteIndex, false);
		spriteBarItemsNumber++;
		if(spriteBarItemsNumber==spriteBarMaxDisplayed) spriteBar->resize(spriteBar->sizeHint());
	}
	
	addSpriteAction(":/images/add.png", tr("Add a new kind of sprite..."), -2, false);
	spriteBarItemsNumber++;
	
	if(spriteBarItemsNumber<spriteBarMaxDisplayed) spriteBar->resize(spriteBar->sizeHint());
}

//add an action to the spriteBar, with the given icon and description
void GameWidget::addSpriteAction(QString iconName, QString description, int spriteIndex, bool isChecked) {
	QString iconAbsoluteName;
	if(iconName.startsWith(":")) {
		//this is an internal resource file
		iconAbsoluteName = iconName;
	} else {
		iconAbsoluteName = FileRepository::getAbsoluteFileName(iconName);
	}
	
	QPixmap pixmap(iconAbsoluteName);
	if(pixmap.isNull()) {
		//file does not exist or format is invalid
		//use the default embedded image
		pixmap.load(":/images/file_broken.png");
	}
	
	QAction *action = new QAction(QIcon(pixmap), description, spriteBar);
	spriteBarGroup->addAction(action);
	spriteBar->addAction(action);
	
	action->setData(spriteIndex);				//data stored in the action, used to remember the associated sprite
	action->setCheckable(true);					//on/off behaviour when clicked
	action->setChecked(isChecked);				//"select mode" (open hand) by default
	
	spriteBarMapper->setMapping(action, spriteIndex);	//map this action to the value spriteIndex
	connect(action, SIGNAL(triggered()), spriteBarMapper, SLOT(map()) );
	
	if(spriteIndex==-1) {
		//this is the hand action, it can be called directly (when the user presses escape)
		connect(this, SIGNAL(editModeActivateHand()), action, SLOT(trigger()));
	}
	
	if(isChecked) {
		//note : setChecked() doesn't trigger the action, we need to do it manualy here
		action->trigger();	//triggered will be emitted, which will call spriteBarClicked
	}
}

//change the icon associated to a given sprite
void GameWidget::changeActionIcon(QString iconName, int spriteIndex) {
	QList<QAction *> actionList = spriteBarGroup->actions();
	int i;
	for(i=0; i<actionList.size(); i++) {
		if(actionList.at(i)->data() == spriteIndex) {
			//update the action
			QString iconAbsoluteName = FileRepository::getAbsoluteFileName(iconName);
			QPixmap pixmap(iconAbsoluteName);
			if(pixmap.isNull()) {
				//file does not exist or format is invalid
				//use the default embedded image
				pixmap.load(":/images/file_broken.png");
			}
			
			actionList.at(i)->setIcon(QIcon(pixmap));
		}
	}
}

void GameWidget::startEdit() {
	isEditMode = true;
	createSpriteBar();
	
	setEditSceneLimits();	//size of the scene so that the entire level is displayed
	setMinimumWidth(620);
	setMinimumHeight(610);
	resize(620, 610);		//resize the window to avoid scrollbars (for an unknown reason, the size is slightly different from game mode)
	
	stopAllTimers();
	
	//display all the sprites of the level
	if(isTestMode) {
		//this is only the end of the test mode, don't add the sprites again
		isTestMode = false;
		parameterWidget->setEnabled(true);
	} else {
		for(level->currentSprite=level->spriteMap.begin(); level->currentSprite!=level->spriteMap.end(); level->currentSprite++) {
			//create the sprite pointed by currentSprite
			SpriteParameter p = level->currentSprite.value();
			addSpriteToEditScene(p);
		}
	}
	
	unselectAllItems();
	
	//set the background color
	setBackgroundColor(level->levelParam.backgroundColor);
	
	setScene(editScene);		//display the game Scene
	spriteBar->show();			//show the sprite bar to select a sprite
	parameterWidget->show();	//show the dialog with the parameters
}

//set the edit scene rectangle so that the entire level is displayed
void GameWidget::setEditSceneLimits() {
	Global::sceneXStart = Global::gameViewX;
	Global::sceneYStart = level->levelParam.y_end;
	Global::sceneXEnd = Global::gameViewX + Global::gameViewW;
	Global::sceneYEnd = Global::gameViewY + Global::gameViewH;
	
	editScene->setSceneRect(Global::sceneXStart, Global::sceneYStart, Global::sceneXEnd - Global::sceneXStart, Global::sceneYEnd - Global::sceneYStart);
	setSceneRect(Global::sceneXStart, Global::sceneYStart, Global::sceneXEnd - Global::sceneXStart, Global::sceneYEnd - Global::sceneYStart);
}

void GameWidget::setParameterMode(ParametersMode mode) {
	if(parameterWidgetMode == mode) return;
	parameterWidgetMode = mode;
	if(mode == SpriteParameterMode) {
		parameterWidget->setParameterList(tr("Sprite Parameters"), SpriteParameter::getParameters(), SpriteParameter::getSeparator());
	} else
	if(mode == LevelParameterMode) {
		parameterWidget->setParameterList(tr("Level global Parameters"), LevelParameter::getParameters(), LevelParameter::getSeparator());
	} else
	if(mode == SpriteKindParameterMode) {
		parameterWidget->setParameterList(tr("Sprite Kind Default Parameters"), SpriteKindParameter::getParameters(), SpriteKindParameter::getSeparator());
	}
}

//add a sprite to the edit scene
void GameWidget::addSpriteToEditScene(SpriteParameter p) {
	Sprite * sprite = new Sprite;
	sprite->setSelectable(true);
	sprite->setMovable(true);
	sprite->setMovableOutScene(false);
	sprite->setParameters(p);
	connect(sprite, SIGNAL(spriteSelected(SpriteParameter)), this, SLOT(spriteSelected(SpriteParameter)));
	connect(sprite, SIGNAL(spriteMoved(SpriteParameter)), this, SLOT(spriteMoved(SpriteParameter)));
	editScene->addItem(sprite);
}

void GameWidget::spriteBarClicked(int spriteIndex) {
	barSelectedSprite = spriteIndex;
	unselectAllItems();
	if(spriteIndex == -1) {
		//this is the select mode
		addSpriteMode = false;
		setParameterMode(LevelParameterMode);
		parameterWidget->setParametersValue(level->levelParam.toUserString());
		cursorPixmapName.clear();
	} else
	if(spriteIndex == -2) {
		//add a new kind of sprite
		//ask confirmation + the id of the sprite kind to be copied
		SimpleDialogWidget simpleDialogWidget(this);
		simpleDialogWidget.setParameters(tr("Add a new kind of sprite") , tr("A copy of an already sprite kind will be added.\nId of an already existing sprite kind :"), tr("Add"));
		int currentMaxId = level->levelParam.spriteKinds.size() - 1;
		simpleDialogWidget.setDefaultAnswer(QString("%1").arg(currentMaxId));
		int accept=simpleDialogWidget.exec();
		if(!accept) return;
		
		bool ok=false;
		int askedId = simpleDialogWidget.getUserAnswer().toInt(&ok);
		if(!ok || askedId<0 || askedId > currentMaxId) return;
		
		//note : we simply add a copy of the asked sprite
		int newSpriteIndex = level->levelParam.addSpriteKind(askedId);
		
		//add a new action for this sprite
		QString pixmapName = QString::fromStdString(level->levelParam.spriteKinds[newSpriteIndex].pixmapFileName);
		QString description = QString("Add: %1").arg(pixmapName);	//TODO : there should be a description in spritekindparameter
		addSpriteAction(pixmapName, description, newSpriteIndex, /*isChecked*/ true);
		spriteBarItemsNumber++;
		
		//resize the spriteBar
		if(spriteBarItemsNumber<=spriteBarMaxDisplayed) spriteBar->resize(spriteBar->sizeHint());
		
		//display a message for the user
		QMessageBox msgBox(parameterWidget);
		msgBox.setText(tr("We have added a copy of the asked sprite kind.\nPlease edit the parameters of this new sprite kind."));
		msgBox.exec();
		
		cursorPixmapName = pixmapName;
	} else {
		//this sprite kind will be used to add new sprites; refresh the parameter widget
		addSpriteMode = true;
		setParameterMode(SpriteKindParameterMode);
		SpriteKindParameter skp = level->getSpriteKindParameter(barSelectedSprite);
		cursorPixmapName = QString::fromStdString(skp.pixmapFileName);
		parameterWidget->setParametersValue(skp.toLinearString());
	}
	
	setMouseShape(cursorPixmapName);
}

//set the mouse cursor from the shortname of a pixmap
void GameWidget::setMouseShape(QString shortPixmapName) {
	//particular case of an empty file name
	//we simply use the default cursor (open hand)
	if(shortPixmapName.isEmpty()) {
		viewport()->setCursor(Qt::OpenHandCursor);
		return;
	}
	
	QPixmap pixmap;		//create a null pixmap
	
	//result is a null pixmap if the file could not be loaded correctly
	pixmap.load(FileRepository::getAbsoluteFileName(shortPixmapName));
	
	if(pixmap.isNull()) {
		//file does not exist or invalid format
		//we use the default pixmap embedded in the binary
		pixmap.load(":/images/file_broken.png");
	}
	
	QCursor cursor(pixmap);
	viewport()->setCursor(cursor);
}

//called when a sprite is selected
void GameWidget::spriteSelected(SpriteParameter p) {
	currentSelectedSprite = (Sprite *) sender();
	setParameterMode(SpriteParameterMode);
	parameterWidget->setParametersValue(p.toString());
}

//called when a sprite is moved (similar to a selection, but we indicate that scene changes occured)
void GameWidget::spriteMoved(SpriteParameter p) {
	currentSelectedSprite = (Sprite *) sender();
	setParameterMode(SpriteParameterMode);
	parameterWidget->setParametersValue(p.toString());
	emit sceneChanged();
}

//called when a change is made in the parameter dialog
void GameWidget::parameterChanged(std::string s) {
	emit sceneChanged();	//signals that the scene changed
	
	if(parameterWidgetMode == SpriteParameterMode) {
		//update the values of the currently selected sprite
		SpriteParameter p;
		p.fromString(s);
		p.limitValuesRange();
		
		//limits of the scene for x and y
		SpriteKindParameter skp = level->getSpriteKindParameter(p.spriteShape);
		p.enforceEditSceneLimits(skp.spriteWidth, skp.spriteHeight);
		
		if(currentSelectedSprite) {
			currentSelectedSprite->setParameters(p);
		} else {
			std::cout << "Error : GameWidget::parameterChanged called without any selected sprite" << std::endl;
		}
		
		//update the parameter widget (some values out of range may have changed)
		parameterWidget->setParametersValue(p.toString());
	} else
	if(parameterWidgetMode == LevelParameterMode) {
		level->levelParam.fromUserString(s);
		level->levelParam.limitValuesRange();
		
		//in case the background color changed
		setBackgroundColor(level->levelParam.backgroundColor);
		
		//update the parameter widget (some values out of range may have changed)
		parameterWidget->setParametersValue(level->levelParam.toUserString());
		
		if(Global::sceneYStart != level->levelParam.y_end) {
			//the length of the level has changed, reset the scene limits
			setEditSceneLimits();
		}
	} else
	if(parameterWidgetMode == SpriteKindParameterMode) {
		SpriteKindParameter skp;
		skp.fromLinearString(s);
		skp.detectSpriteDimensions();
		QString oldPixmapName = level->getPixmapName(skp.spriteKind);
		level->setSpriteKindParameter(barSelectedSprite, skp);
		skp.limitValuesRange();
		
		//update the parameter widget (some values out of range may have changed)
		parameterWidget->setParametersValue(skp.toLinearString());
		
		QString newPixmapName = QString::fromStdString(skp.pixmapFileName);
		if(oldPixmapName != newPixmapName) {
			//the pixmap changed : we need to change all the displayed pixmaps of sprites of this kind
			QList <QGraphicsItem *> itemsList = editScene->items();
			int i;
			for(i=0; i<itemsList.size(); i++) {
				Sprite * sprite = (Sprite *) itemsList.at(i);
				if(sprite->getSpriteShape()==skp.spriteKind) {
					//we need to change the pixmap, and maybe also the coordinates to ensure visibility
					SpriteParameter p = sprite->getParameters();					//get the current parameters
					p.enforceEditSceneLimits(skp.spriteWidth, skp.spriteHeight);	//ensure visibility in the scene
					sprite->setParameters(p);										//update the sprite with the new parameters
				}
			}
			
			//we also change the icon of the action associated to this sprite
			changeActionIcon(newPixmapName, skp.spriteKind);
			
			//change the mouse cursor
			cursorPixmapName = newPixmapName;
			setMouseShape(cursorPixmapName);
		}
	}
}

//reimplement mouse-click for edit mode
void GameWidget::mousePressEvent(QMouseEvent *mouseEvent) {
	if(addSpriteMode) {
		if(barSelectedSprite==0) return; //we don't accept multiple player sprites
		
		//add a sprite at the cursor position with the default parameters
		SpriteKindParameter skp = level->getSpriteKindParameter(barSelectedSprite);
		SpriteParameter p = skp.defaultParameters;
		QPointF scenePos = mapToScene(mouseEvent->pos());		//position of the click in the scene
		p.x_pos = scenePos.x() - skp.spriteWidth/2;
		p.y_pos = scenePos.y() - skp.spriteHeight/2;
		p.enforceEditSceneLimits(skp.spriteWidth, skp.spriteHeight);
		addSpriteToEditScene(p);	//add a new sprite in the edit scene
		emit sceneChanged();		//signals that the scene changed
	} else {
		QGraphicsView::mousePressEvent(mouseEvent); //default implementation
	}
}

void GameWidget::mouseReleaseEvent(QMouseEvent *mouseEvent) {
	if(addSpriteMode) {
		//if a cursor pixmap is set, we must reset it when the mouse is released
		//(default implementation of Qt set the mouse cursor back to default)
		if(!cursorPixmapName.isEmpty()) setMouseShape(cursorPixmapName);
	} else {
		QGraphicsView::mouseReleaseEvent(mouseEvent); //default implementation
	}
}

void GameWidget::editSelectionChanged() {
	if(addSpriteMode==false && isEditMode && editScene->selectedItems().isEmpty()) {
		setParameterMode(LevelParameterMode);
		parameterWidget->setParametersValue(level->levelParam.toUserString());
	}
}

//remove the currently selected sprite (used in edit mode when we press the delete key)
void GameWidget::removeCurrentSprite() {
	//check that some valid sprite is currently selected
	if(currentSelectedSprite) {
		//check that the selected sprite is not the player sprite (deletion is forbidden)
		if(currentSelectedSprite->getSpriteShape() != 0) {
			currentSelectedSprite->deleteLater();
			currentSelectedSprite = 0;
			emit sceneChanged();
		}
	}
}

//remove the focus from the current item, to ensure that no item is selected
void GameWidget::unselectAllItems() {
	currentSelectedSprite = 0;
	editScene->clearSelection();		//remove any selection
}

void GameWidget::saveTemporaryFile() {
	//update the level from the scene objects
	QList<QGraphicsItem*> itemList = editScene->items();
	level->createLevelFromScene(itemList);
	
	//write the level in a temporary file
	level->writeFile(FileRepository::temporaryFileName());
}

void GameWidget::closeEditMode() {
	//delete all the sprites of the scene
	QList<QGraphicsItem*> list = editScene->items();
	int i;
	for (i = 0; i < list.size(); i++) {
		((Sprite *) list.at(i))->deleteLater(); //the sprite will be removed from the scene and deleted from memory
	}
	
	addSpriteMode=false;					//stop the addSpriteMode
	if(spriteBar!=0) spriteBar->hide();		//hide the sprite bar to select a sprite
	parameterWidget->hide();				//hide the dialog with the parameters
	resize(610, 610);						//resize the window to avoid scrollbars (for an unknown reason, the size is slightly different from game mod)
	setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);	//size of the viewport
	isEditMode = false;
}

//================================ Menu ===========================================================

void GameWidget::createMenuScene() {
	menuScene = new QGraphicsScene;
	menuScene->setSceneRect(Global::gameViewX, Global::gameViewY, Global::gameViewW, Global::gameViewH);
	menuScene->setItemIndexMethod(QGraphicsScene::NoIndex);
	
	QFont bigFont("Impact", 32, QFont::Bold);
	QString labelStyle("QLabel {background-color : black; color : #009900; }");
	QString pushbuttonStyle("QPushButton {background-color : black; color : #009900; border-style: solid; border-width : 5px; border-color : black; padding: 6px;} QPushButton:focus {border-color : #009900; } QPushButton:hover{ background-color: midnightblue; }" );
	
	//label centered in the screen for the loading screen
	labelLoading = new QLabel;
	labelLoading->setText("AAAAAAAAAAAAAAAA");
	labelLoading->setFont(bigFont);
	labelLoading->setStyleSheet(labelStyle);
	labelLoading->setFixedSize(labelLoading->sizeHint());
	QGraphicsProxyWidget *proxyWidget = menuScene->addWidget( labelLoading );
	proxyWidget->setPos(-130, -30);
	
	labelResult = new QLabel;
	labelResult->setText("AAAAAAAAAAAAAAAA");
	labelResult->setFont(bigFont);
	labelResult->setStyleSheet(labelStyle);
	labelResult->setFixedSize(labelResult->sizeHint());
	proxyWidget = menuScene->addWidget( labelResult );
	proxyWidget->setPos(-220, -250);
	
	labelScore = new QLabel;
	labelScore->setText("AAAAAAAAAAAAAAAA");
	labelScore->setFont(bigFont);
	labelScore->setStyleSheet(labelStyle);
	labelScore->setFixedSize(labelScore->sizeHint());
	proxyWidget = menuScene->addWidget( labelScore );
	proxyWidget->setPos(-220, -200);
	
	//a big black label to mask the top of the high score when scrolling up
	QLabel* labelMask = new QLabel;
	labelMask->setStyleSheet(labelStyle);
	labelMask->setFixedSize(QSize(Global::gameViewW+4, 74 - Global::gameViewY));	//size such that the down is at y=-10
	proxyWidget = menuScene->addWidget( labelMask );
	proxyWidget->setPos(Global::gameViewX-2, Global::gameViewY - 6);		//placed at the top
	proxyWidget->setZValue(-1);	//under other items except labelHighScore
	
	QFont highScoreFont("Impact", 24, QFont::Bold);
	labelHighScore = new QLabel;
	labelHighScore->setFont(highScoreFont);
	labelHighScore->setStyleSheet(labelStyle);
	proxyHighScore = menuScene->addWidget( labelHighScore );
	proxyHighScore->setPos(-220, -200);
	proxyHighScore->setZValue(-2);	//under labelMask
	
	playButton = new QPushButton;
	playButton->setFont(bigFont);
	playButton->setStyleSheet(pushbuttonStyle);
	proxyWidget = menuScene->addWidget( playButton );
	proxyWidget->setPos(-150, -100);
	connect(playButton, SIGNAL(clicked()), this, SLOT(playLevel()));
	
	quitButton = new QPushButton;
	quitButton->setFont(bigFont);
	quitButton->setStyleSheet(pushbuttonStyle);
	proxyWidget = menuScene->addWidget( quitButton );
	proxyWidget->setPos(-150, -20);
	connect(quitButton, SIGNAL(clicked()), this, SLOT(quitGameWidget()));
	
	updateResultText();
}

void GameWidget::updateResultText() {
	labelLoading->setVisible(false);
	labelResult->setVisible(true);
	
	//initialization of the quit button done here to have the translations
	quitButton->setVisible(true);
	quitButton->setText(tr("Quit"));
	quitButton->setFixedSize(quitButton->sizeHint());
	
	if(gameStatus == GameOver) {
		labelResult->setText(tr("  Game over"));
		
		if(previousLevelScore!=0) {
			//the previous level score is decreased by 10%
			//this is a handicap if the player want to restart from the current level
			previousLevelScore = (90 * previousLevelScore) / 100;
			if(previousLevelScore==0) previousLevelScore = 1;
			QString retryString = tr("Continue with %1").arg(previousLevelScore);
			playButton->setText(retryString);
		} else {
			playButton->setText(tr("Restart"));
		}
		
		playButton->setVisible(true);
		playButton->setFocus();
	} else
	if(gameStatus == LevelComplete) {
		labelResult->setText(QString("<font color='red'>") + tr("  Game completed") + QString("</font>"));
		playButton->setVisible(false);
		quitButton->setFocus();
	}
	
	//resize the play button according to the text
	playButton->setFixedSize(playButton->sizeHint());
}

void GameWidget::updateScoreText() {
	labelScore->setVisible(true);
	labelScore->setText(tr("  Score: %1").arg(levelScore));
}

void GameWidget::updateHighScoreText() {
	qreal newYPos = 230-(qreal)(highScoreAnimationIndex);
	
	//when the bottom of the high score label is above y=0, restart the animation loop
	if(newYPos + labelHighScoreHeight < 0) highScoreAnimationIndex = 0;
	
	proxyHighScore->setPos(-220, 230-(qreal)(highScoreAnimationIndex) );
	highScoreAnimationIndex++;
}

void GameWidget::startHighScoreAnimation() {
	//start the animation of the high scores
	highScoreAnimationIndex = 0;
	QString scoreFileName = FileRepository::getHighScoreFileName(startLevelName);
	QString absoluteFileName = FileRepository::getAbsoluteFileName(scoreFileName, FileRepository::FileReadOnly);
	HighScoreFile scoreFile(absoluteFileName);
	labelHighScore->setText(scoreFile.getHtmlScoreList());
	labelHighScore->setFixedSize(QSize(10,10));		//to be sure that size is reduced
	QSize labelSize = labelHighScore->sizeHint();
	labelHighScoreHeight = labelSize.height();	//save the height, used to loop the animation
	labelHighScore->setFixedSize(labelSize);
	
	updateHighScoreText();	//called once directly before making the high score visible
	highScoreTimer->start(50);
	
	labelHighScore->setVisible(true);
}

//called by the flukzwidget when the updated high score is received
void GameWidget::refreshHighScore() {
	if(gameStatus!=GameRunning) {
		startHighScoreAnimation();
	}
}

//called after a click on the play Button
void GameWidget::playLevel() {
	if(gameStatus == GameOver) {
		//retry the current level
		startGame();
	} else
	if(gameStatus == LevelComplete) {
		//play the next level
		expectedNextLevelName = level->nextLevelName();
		previousLevelScore = levelScore;
		emit askPage(level->nextLevelName());
	}
}

void GameWidget::displayLoadingMessage() {
	if(currentLevelName == expectedNextLevelName) {
		labelLoading->setText(tr("Next level loading..."));
	} else {
		labelLoading->setText(tr("Loading..."));
	}
	labelLoading->setVisible(true);
	labelResult->setVisible(false);
	labelScore->setVisible(false);
	labelHighScore->setVisible(false);
	playButton->setVisible(false);
	quitButton->setVisible(false);
	setBackgroundBrush(Qt::black);
	setScene(menuScene);
}

//called after a click on the quit Button
void GameWidget::quitGameWidget() {
	//a void string means that we simply ask to quit
	emit askPage(QString(""));
}
