/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "level.h"
#include "sprite.h"
#include "../networkengine/filerepository.h"

#include<QFile>
#include<QTextStream>

Level::Level() {

}

void Level::init() {
	//initialize the variables
	currentSprite = spriteMap.begin();
	y_front = -300;
}

//overwrite the level in the current file
//used to save the level after editing
void Level::overWrite() {
	writeFile(currentFileName);
}

//Write the level data in a file
void Level::writeFile(const QString &filename) {
	//open file
	QFile file( filename );
	if ( !file.open( QIODevice::WriteOnly ) ) return;
	
	//write database
	QTextStream text( &file );
	
	//header
	QString headerName("[Flukz File]");
	text << headerName << "\n";
	
	//current string format version
	int stringFormatVersion = 0;
	text << stringFormatVersion << "\n";
	
	//save the level global parameters
	text << QString::fromStdString(levelParam.toString()) << "\n";
	
	//save the sprite parameters
	QMultiMap<SpriteCoordinate, SpriteParameter>::iterator iter;
	
	for(iter=spriteMap.begin(); iter!=spriteMap.end(); iter++) {
		text << QString::fromStdString(iter.value().toString()) << "\n";
	}
	
	//close file
	file.close();
}

//Read the level data from a file
void Level::readFile(const QString &filename) {
	currentFileName = filename;
	
	//open file
	QFile file( filename );
	if ( !file.open( QIODevice::ReadOnly ) ) return;
	
	QTextStream text( &file );
	
	//read header
	if (!text.atEnd() ) {
		QString headerName("[Flukz File]");
		if (text.readLine()!=headerName) {
			file.close();
			return;
		}
	}
	
	//read string format version
	int stringFormatVersion=-1;
	if (!text.atEnd() ) {
		QString line = text.readLine();
		stringFormatVersion = line.toInt();
	}
	
	if(stringFormatVersion!=0) {
		//unsupported version
		file.close();
		return;
	}
	
	//read the level global parameters
	std::string str;
	if (!text.atEnd() ) {
		str = text.readLine().toStdString();
		levelParam.fromString(str, stringFormatVersion);
	}
	
	//clear previous sprite parameters
	spriteMap.clear();
	
	//read the sprite parameters
	levelParam.itemCount=0;									//initialization of the ids
	SpriteParameter spriteParam;
	while ( !text.atEnd() ) {
		str = text.readLine().toStdString();
		spriteParam.fromString(str, stringFormatVersion);
		addSpriteToMap(spriteParam);
	}
	
	//close file
	file.close();
}

//Create the level data from a list of items (the list is given as QGraphicsItem*, which are all Sprite*
//used in the edit mode to save the changes
void Level::createLevelFromScene(QList<QGraphicsItem*> itemList) {
	spriteMap.clear();			//clear previous sprite parameters
	levelParam.itemCount=0;		//initialization of the ids
	int i;
	for (i = 0; i < itemList.size(); i++) {
		SpriteParameter spriteParam = ((Sprite *) itemList.at(i))->getParameters();
		addSpriteToMap(spriteParam);
	}
}

//create a partial level data from a list of items, starting at some y value
//this is used to test only a part of the level in the edit mode
void Level::createPartialLevelFromScene(QList<QGraphicsItem*> itemList, qreal y_start) {
	spriteMap.clear();			//clear previous sprite parameters
	levelParam.itemCount=0;		//initialization of the ids
	int i;
	qreal translation = Global::gameViewY - y_start;
	qreal margin = 10;
	for (i = 0; i < itemList.size(); i++) {
		SpriteParameter spriteParam = ((Sprite *) itemList.at(i))->getParameters();
		if(spriteParam.spriteShape==0 || translation <= margin) {
			//player ship or no translation, nothing special to do
			//note : we consider there is no translation if less than 10 (margin for not so clear reason)
			addSpriteToMap(spriteParam);
		} else
			if(spriteParam.y_pos < y_start || (spriteParam.collisionPolicy==SpriteParameter::NoCollision) ) {
			//case of translation (partial level)
			//we keep only items that are above the start position or background items
			//and we translate them so that y_start is the top of the beginning screen
			spriteParam.y_pos += translation;
			addSpriteToMap(spriteParam);
		}
	}
	
	//shorten the level so that the partial level ends at the same position as the real complete level
	if(translation > margin) {
		level_execute_end = levelParam.y_end + translation;
	} else {
		level_execute_end = levelParam.y_end;
	}
}

//give a unique id to the sprite parameters, and add it to the map
void Level::addSpriteToMap(SpriteParameter p) {
	p.id = levelParam.itemCount;
	levelParam.itemCount++;
	
	//sprite coordinates in the map (used only to sort the sprites)
	//sprites are sorted with the y coordinate of the bottom part of the pixmap (not the upper part)
	//so we need to apply a correction to take into account the height of the sprite
	SpriteCoordinate spc;
	spc.x = p.x_pos;
	spc.y = -p.y_pos - (qreal) (getPixmapHeight(p.spriteShape));
	
	//add the sprite to the game Map
	spriteMap.insert(spc, p);
}

SpriteKindParameter Level::getSpriteKindParameter(int spriteIndex) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		return levelParam.spriteKinds[spriteIndex];
	} else {
		//error case
		std::cout << "Level::getSpriteKindParameter: " << spriteIndex << " is not an existing sprite index" << std::endl;
		SpriteKindParameter p;
		return p;
	}
}

void Level::setSpriteKindParameter(int spriteIndex, SpriteKindParameter skp) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		levelParam.spriteKinds[spriteIndex] = skp;
	} else {
		//error case
		std::cout << "Level::getSpriteKindParameter: " << spriteIndex << " is not an existing sprite index" << std::endl;
	}
}

SpriteParameter Level::getSpriteDefaultParameters(int spriteIndex) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		return levelParam.spriteKinds[spriteIndex].defaultParameters;
	} else {
		//error case
		std::cout << "Level::getSpriteDefaultParameters: " << spriteIndex << " is not an existing sprite index" << std::endl;
		SpriteParameter p;
		return p;
	}
}

//return the pixmap file name associated to a given sprite index
QString Level::getPixmapName(int spriteIndex) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		return QString::fromStdString(levelParam.spriteKinds[spriteIndex].pixmapFileName);
	} else {
		//error case
		std::cout << "Level::getSpriteKindParameter: " << spriteIndex << " is not an existing sprite index" << std::endl;
		QString s;
		return s;
	}
}

//return the width of the pixmap associated to a given sprite index
int Level::getPixmapWidth(int spriteIndex) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		return levelParam.spriteKinds[spriteIndex].spriteWidth;
	} else {
		//error case
		std::cout << "Level::getSpriteKindParameter: " << spriteIndex << " is not an existing sprite index" << std::endl;
		return 0;
	}
}

//return the height of the pixmap associated to a given sprite index
int Level::getPixmapHeight(int spriteIndex) {
	if(spriteIndex>=0 && spriteIndex < (int) levelParam.spriteKinds.size()) {
		return levelParam.spriteKinds[spriteIndex].spriteHeight;
	} else {
		//error case
		std::cout << "Level::getSpriteKindParameter: " << spriteIndex << " is not an existing sprite index" << std::endl;
		return 0;
	}
}

//return the list of sounds of the level
//we return a list of absolute file names (used to load the files in the sound engine)
QStringList Level::getSoundFileList() {
	QStringList result;
	unsigned int i;
	for(i=0; i<levelParam.soundFileNames.size(); i++) {
		QString shortFileName = QString::fromStdString(levelParam.soundFileNames[i]);
		result.append(FileRepository::getAbsoluteFileName(shortFileName));
	}
	return result;
}

QStringList Level::filesDependencyList(bool isCreditsDependencies) {
	//image and sound dependencies
	return levelParam.filesDependencyList(isCreditsDependencies);
}

//indicates if a next level is defined or not
bool Level::hasNextLevel() {
	if(levelParam.nextLevelFileName.empty()) {
		return false;
	} else {
		return true;
	}
}

//return the file name for the next level
QString Level::nextLevelName() {
	return QString::fromStdString(levelParam.nextLevelFileName);
}
