/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "levelparameter.h"
#include "../tools/stringconverter.h"
#include "../networkengine/filerepository.h"

#include <QColor>

//default values of the parameters
LevelParameter::LevelParameter() {
	y_end = -1000;
	scrollingSpeed = 0.2;
}

//export the level parameters as a string
std::string LevelParameter::toString() const {
	StringConverter conv;
	//"UT " = activate separator between elements
	//"E]" = use ] as the separator, and } as STL separator
	conv << CvCd("UT") << CvCd("E]") << CvCd("S}");
	conv << y_end << scrollingSpeed << backgroundColor << soundFileNames << nextLevelFileName << spriteKinds;
	return conv.getString();
}

//import the level parameters from a string
//the version number is for compatibility to read files from older versions of Flukz
void LevelParameter::fromString(const std::string& stringA, int version) {
	soundFileNames.clear();
	spriteKinds.clear();
	
	StringConverter conv(stringA);
	if(version==0) {
		conv << CvCd("UT") << CvCd("E]") << CvCd("S}");
		conv >> y_end >> scrollingSpeed >> backgroundColor >> soundFileNames >> nextLevelFileName >> spriteKinds;
	}
}

//export the parameters as a string for the parameter widget
//Note : the only difference with toString() is that we don't add the sprite kinds parameters
std::string LevelParameter::toUserString() const {
	StringConverter conv;
	//"UT " = activate separator between elements
	//"E;" = use ; as the separator
	conv << CvCd("UT") << CvCd("E;");
	conv << y_end;
	conv << scrollingSpeed;
	conv << backgroundColor;
	for(int i=0; i<3; i++) {
		if(i < (int) soundFileNames.size())
			conv << soundFileNames[i];
		else {
			conv << std::string(" ");
		}
	}
	
	if(nextLevelFileName.empty()) {
		conv << std::string(" ");
	} else {
		conv << nextLevelFileName;
	}
	return conv.getString();
}

//import the parameters from a string of the parameter widget
//the version number is for compatibility to read files from older versions of Flukz
//Note : the only difference with fromString() is that we don't have the sprite kinds parameters
void LevelParameter::fromUserString(const std::string& stringA) {
	StringConverter conv(stringA);
	//if(version==0) {
		conv << CvCd("UT") << CvCd("E;");
		conv >> y_end;
		conv >> scrollingSpeed;
		conv >> backgroundColor;
		while(soundFileNames.size()<3) soundFileNames.push_back(std::string(" "));
		conv >> soundFileNames[0] >> soundFileNames[1] >> soundFileNames[2];
		conv >> nextLevelFileName;
		if(nextLevelFileName == std::string(" ")) nextLevelFileName.clear();
	//}
}

//limit the values to the correct range
//it can be used after the user changed the values
//return true if some values was out of range
bool LevelParameter::limitValuesRange() {
	bool outOfRange = false;
	if(scrollingSpeed<0.05) {
		scrollingSpeed=0.05;
		outOfRange = true;
	}
	if(scrollingSpeed>10) {
		scrollingSpeed=10;	
		outOfRange = true;
	}
	if(y_end>-500) {
		y_end = -500;
		outOfRange = true;
	}
	if(y_end<-20000) {
		y_end = -20000;
		outOfRange = true;
	}
	
	//note : QColor::isValidColor() avoided to keep compatibility with Qt 4.6
	QColor color;
	color.setNamedColor(QString::fromStdString(backgroundColor));
	if(!color.isValid()) {
		backgroundColor = "lightgray";
		outOfRange = true;
	}
	
	return outOfRange;
}

//return a list of int values that describe the type of the parameters of a sprite
QList<Parameter> LevelParameter::getParameters() {
	QList<Parameter> result;
	
	//y_end
	result += Parameter(Parameter::QrealType,
						QObject::tr("Level length"),
						QObject::tr("length of the level (in pixels)"));
	
	//scrolling speed
	result += Parameter(Parameter::QrealType,
						QObject::tr("Scrolling speed"),
						QObject::tr("speed at which the level scrolls"));
	
	//background color of the level
	result += Parameter(Parameter::StringType,
						QObject::tr("Background color"),
						QObject::tr("background color of the level (examples: gray, #808080)"));
	
	//music of the level
	result += Parameter(Parameter::StringType,
						QObject::tr("Music file name"),
						QObject::tr("name of the file that contains the music (wave format)"));
	
	//sound for the explosion
	result += Parameter(Parameter::StringType,
						QObject::tr("Explosion file name"),
						QObject::tr("name of the file that contains the explosion sound (wave format)"));
	
	//sound for fire collision without explosion
	result += Parameter(Parameter::StringType,
						QObject::tr("Fire collision file name"),
						QObject::tr("name of the file that contains the sound for collision with fire without explosion (wave format)"));
	
	//next level file name
	result += Parameter(Parameter::StringType,
						QObject::tr("Next level file name"),
						QObject::tr("name of the file that will be launched when the current level is completed"));
	return result;
}

//return the string command to initialize stringConverter correctly
std::string LevelParameter::getSeparator() {
	return "E;";
}

//return the list of files refered in the parameters of the level
QStringList LevelParameter::filesDependencyList(bool isCreditsDependencies) {
	QStringList result;
	
	//add the list of files used for the images
	unsigned int i;
	int j;
	for(i=0; i<spriteKinds.size(); i++) {
		result.append(QString::fromStdString(spriteKinds[i].pixmapFileName));
		
		//for credits, we also add the meta page in the dependencies
		if(isCreditsDependencies) {
			result.append(QString::fromStdString(spriteKinds[i].pixmapFileName) + FileRepository::metaResource);
		}
		
		//list of animation files for this image
		for(j=0; j<spriteKinds[i].animationLength; j++) {
			QString fileName = animationFileName(i, j+1);
			if(!result.contains(fileName, Qt::CaseSensitive))	{
				result.append(fileName);
				
				//for credits, we also add the meta page in the dependencies
				if(isCreditsDependencies) {
					result.append(fileName + FileRepository::metaResource);
				}
			}
		}
	}
	
	//add the list of sounds
	for(i=0; i<soundFileNames.size(); i++) {
		QString fileName = QString::fromStdString(soundFileNames[i]);
		if(!result.contains(fileName, Qt::CaseSensitive)) {
			result.append(fileName);
			
			//for credits, we also add the meta page in the dependencies
			if(isCreditsDependencies) {
				result.append(fileName + FileRepository::metaResource);
			}
		}
	}
	
	return result;
}

//return the file name of an animation file
//for the sprite of index i and the file number j in the animation
QString LevelParameter::animationFileName(int spriteIndex, int animationIndex) {
	QString result;
	if(spriteIndex<0 || spriteIndex>=(int)spriteKinds.size()) return result; //error case
	result = QString::fromStdString(spriteKinds[spriteIndex].animationFileName);
	int p = result.indexOf(".");
	if(p!=-1) {
		result.insert(p, QString("%1").arg(animationIndex));
	}
	return result;
}

//add a new kind of sprite with a copy of an existing one
//return the index of the new kind of sprite
//if the askedId is invalid, we automatically use the last id
int LevelParameter::addSpriteKind(int askedId) {
	if(spriteKinds.empty()) {
		SpriteKindParameter skp;
		skp.spriteKind = 0;
		skp.defaultParameters.spriteShape = 0;
		spriteKinds.push_back(skp);
		return skp.spriteKind;
	} else {
		if(askedId<0 || askedId > (int) spriteKinds.size()) {
			askedId = spriteKinds.size() - 1;
		}
		SpriteKindParameter skp = spriteKinds[askedId];
		skp.spriteKind = spriteKinds.size();
		skp.defaultParameters.spriteShape = skp.spriteKind;
		spriteKinds.push_back(skp);
		return skp.spriteKind;
	}
}
