/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

 
#include "sprite.h"
#include "level.h"
#include "../networkengine/filerepository.h"

#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>

bool Sprite::leftKeyPressed;
bool Sprite::rightKeyPressed;
bool Sprite::upKeyPressed;
bool Sprite::downKeyPressed;

int Sprite::playerFireSpeed;	//speed of player fire
int Sprite::playerXSpeed;		//speed of player along x (when a key is pressed)
int Sprite::playerYSpeed;		//speed of player along y (when a key is pressed)

QPointF Sprite::mouseInitialCoordinates;	//initial mouse coordinates during a drag-drop
qreal Sprite::spriteInitialX;				//initial sprite X coordinate during a drag-drop
qreal Sprite::spriteInitialY;				//initial sprite Y coordinate during a drag-drop

Level* Sprite::currentLevel;

int Sprite::stepsToExecute;

Sprite::Sprite() {
	movableOutScene = true;
}

//QGraphicsItem flag
void Sprite::setSelectable(bool isSelectable) {
	setFlag(ItemIsSelectable, isSelectable);
}

//QGraphicsItem flag
void Sprite::setMovable(bool isMovable) {
	setFlag(ItemIsMovable, isMovable);
}

//specific to Sprite
void Sprite::setMovableOutScene(bool isMovable) {
	movableOutScene = isMovable;
}

void Sprite::setSpritePixmap(const QString& shortFileName) {
	QPixmap pixmap;		//create a null pixmap
	
	if(!shortFileName.isEmpty()) {
		//result is a null pixmap if the file could not be loaded correctly
		pixmap.load(FileRepository::getAbsoluteFileName(shortFileName));
	}
	
	if(pixmap.isNull()) {
		//file does not exist or invalid format
		//we use the default pixmap embedded in the binary
		pixmap.load(":/images/file_broken.png");
	}
	
	setPixmap(pixmap);	//function of QGraphicsPixmapItem
	
	//initialize the width and height of the sprite
	QRectF rect = boundingRect();
	spriteWidth = rect.width();
	spriteHeight = rect.height();
}

void Sprite::setParameters(SpriteParameter p) {
	//init the pixmap
	QString shortFileName = currentLevel->getPixmapName(p.spriteShape);
	setSpritePixmap(shortFileName);
	
	//set the parameters
	param = p;
	setPos(param.x_pos, param.y_pos);
	setZValue(param.z_pos);
	
	//special case of the player sprite
	if(param.spriteShape==0) {
		//save the values of x, y and fire speed
		playerFireSpeed = param.fireSpeed;
		playerXSpeed = param.x_speed;
		playerYSpeed = param.y_speed;
		
		//initial speeds are 0 (move is triggered by keyboard)
		param.fireSpeed = 0;
		param.x_speed = 0;
		param.y_speed = 0;
	}
	
	//initialize also the other variables
	if(param.fireSpeed!=0) {
		//case of automated fire : set fire_cycle so that we are near the end of the cycle
		fire_cycle = 3*param.fireSpeed/4;
	} else {
		fire_cycle = 0;
	}
	isDestroyed = false;
	explodingCycle = 0;
	
	//currently, the value of a sprite for the score is always the same as the number of lifePoints
	//in the future, it will become a parameter
	scoreValue = param.lifePoints;
}

SpriteParameter Sprite::getParameters() {
	if(param.spriteShape==0) {
		//special case of the player sprite
		//the parameters that can be controlled by the players are stored in different variables
		SpriteParameter playerParameters = param;
		playerParameters.fireSpeed = playerFireSpeed;
		playerParameters.x_speed = playerXSpeed;
		playerParameters.y_speed = playerYSpeed;
		return playerParameters;
	} else {
		//normal case, return the current parameters
		return param;
	}
}

int Sprite::getSpriteShape() {
	return param.spriteShape;
}

int Sprite::getLifePoints() {
	return param.lifePoints;
}

//called when scene.advance() is called (at regular intervals)
//called one time with step=0 and one time with step!=0
//all the necessary computations for collisions are done when step=0
//in step=1, we emit the destroy signals for the sprites that were destroyed
//TODO : this function should be split in more elementary functions
void Sprite::advance(int step) {
	if(step==0) {
		//variation of the position depending on the speed
		qreal x_new = param.x_pos + stepsToExecute * param.x_speed;
		qreal y_new = param.y_pos + stepsToExecute * param.y_speed;
		
		//apply the limit policy if the sprite goes out of the scene
		applyLimitPolicy(x_new, y_new);
		
		param.x_pos = x_new;
		param.y_pos = y_new;
		setPos(param.x_pos, param.y_pos);
		
		if(param.fireSpeed !=0) {
			for(int s=0; s<stepsToExecute; s++) {
				if(fire_cycle == param.fireSpeed) {
					SpriteParameter paramFire = currentLevel->getSpriteDefaultParameters(param.fireIndex);
					qreal margin = 3;	//small margin to avoid collision of the firing sprite with its own fire
					
					//initialize the kind of sprite that is emitting this fire
					paramFire.emittingSpriteShape = param.spriteShape;
					
					//fire at the center of the firing sprite
					//also take into account the width of the fire
					int fireWidth = currentLevel->getPixmapWidth(param.fireIndex);
					paramFire.x_pos = param.x_pos + 0.5*spriteWidth - 0.5*fireWidth;
					
					if(paramFire.y_speed <= 0) {
						//fire is going up : start from the up of firing sprite
						//fire initial position needs to be moved up from the height of fire
						int fireHeight = currentLevel->getPixmapHeight(param.fireIndex);
						paramFire.y_pos = param.y_pos - margin - fireHeight;
					} else {
						//fire is going down : start from the down of firing sprite
						paramFire.y_pos = param.y_pos + spriteHeight + margin;
					}
					emit fire(paramFire);
					fire_cycle = 0;
				} else {
					fire_cycle++;
				}
			}
		}
		
		//we check only the collisions of items that have a collision policy set to Destroyed
		//These items will set directly items with DestroyedIndirectly policy to destroyed
		//collisions are currently NOT detected if exploding cycle is already launched
		//note : it means that everything can go through an explosing item
		//maybe a parameter for this would be good in the future
		if(param.collisionPolicy==SpriteParameter::Destroyed && explodingCycle==0) {
			QList<QGraphicsItem *> collidingItemsList = scene()->collidingItems(this);
			
			//look for something different than background items
			int i;
			int n = collidingItemsList.size();
			for(i=0; i<n; i++) {
				//sprite is killed if it hits anything at the same "depth" (z_pos) as itself
				Sprite * collidingSprite = (Sprite *) (collidingItemsList.at(i));
				if(collidingSprite->param.z_pos == param.z_pos && 
				   //no collision with already destroyed or exploding sprites
				   !collidingSprite->isDestroyed &&
				   collidingSprite->explodingCycle==0 &&
				   //the player ship cannot collide with its own fire
				   !(param.spriteShape == 0 && collidingSprite->param.emittingSpriteShape == 0)
				) {
					int remainingLifePoints = param.lifePoints;
					//start the exploding cycle only if it was not started already
					if(explodingCycle==0) {
						//decrease the life points from the number of life points of the colliding sprite
						//in case of fire, life points describe the destruction power
						param.lifePoints -= collidingSprite->param.lifePoints;
						
						//in the case of the player ship, signal the new life points
						if (param.spriteShape ==0) {
							emit lifePointsChanged(param.lifePoints);
						}
						
						//sprite starts to explode when it has no more life points
						if(param.lifePoints<=0) {
							//negative values cause a bug when the collision is detected from the colliding sprite
							param.lifePoints = 0;
							explosionStart();
						} else {
							emit askSound(2);
						}
					}
					
					if(collidingSprite->param.collisionPolicy==SpriteParameter::DestroyedIndirectly) {
						//destroy immediately the sprite that collided if it is fire
						collidingSprite->isDestroyed=true;
					} else
					if(collidingSprite->param.collisionPolicy==SpriteParameter::Destroyed) {
						//decrease the life points of the colliding sprite if it is a ship
						//negative values not allowed to avoid a bug when collision is detected from the colliding sprite
						collidingSprite->param.lifePoints -= remainingLifePoints;
						if(collidingSprite->param.lifePoints < 0) collidingSprite->param.lifePoints = 0;
					}
				}
			}
		}
		
		//we also check life points here, because the sprite can be destroyed in the colliding loop of another sprite
		if(explodingCycle==0 && param.lifePoints<=0) {
			explosionStart();
		}
		
		if(explodingCycle!=0) {
			for(int s=0; s<stepsToExecute; s++) {
				int animDelay = currentLevel->getSpriteKindParameter(param.spriteShape).animationSpeed;
				if(explodingCycle == animDelay * currentLevel->getSpriteKindParameter(param.spriteShape).animationLength ) {
					//end of the explosing cycle, delete the ship from the scene
					isDestroyed=true;
				} else
				if(explodingCycle%animDelay == 0) {
					//the size of the explosion images is not necessarily the same
					//we center all images on the same point
					qreal oldSpriteWidth = spriteWidth;
					qreal oldSpriteHeight = spriteHeight;
					
					QString animFileName = currentLevel->levelParam.animationFileName(param.spriteShape, (explodingCycle/animDelay)+1);
					setSpritePixmap(animFileName);
					
					//correction of positions to keep the center fixed
					param.x_pos += 0.5*(oldSpriteWidth - spriteWidth);
					param.y_pos += 0.5*(oldSpriteHeight - spriteHeight);
					setPos(param.x_pos, param.y_pos);
				}
				explodingCycle++;
			}
		}
	} else {
		//called after the first step
		if(isDestroyed) {
			emit destroy();
		}
	}
}

//actions done when the sprite starts exploding
void Sprite::explosionStart() {
	emit askSound(1);
	explodingCycle = 1;
	param.fireSpeed = 0;	//stop any fire from the exploding ship
	fire_cycle = 0;
	if(param.spriteShape==0) {
		//stop the movement if it is the player ship
		param.x_speed = 0;
		param.y_speed = 0;
	} else {
		//score reward if an enemy starts exploding
		emit changeScore(scoreValue);
	}
}

//apply the limit policy of the sprite when it goes out of the scene
//note : this function modifies directly the parameters x_new and y_new
void Sprite::applyLimitPolicy(qreal& x_new, qreal& y_new) {
	if(x_new < Global::gameViewX) {
		if(param.limitPolicy==SpriteParameter::BlockInScreen) {
			x_new = Global::gameViewX;
		} else
		if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
			//destroy the widget only when it becomes invisible
			//(with a margin of 5 to avoid problems when creating items)
			if(!isDestroyed && x_new < Global::gameViewX - spriteWidth - 5) {
				isDestroyed = true;
			}
		}
	}
	
	if(x_new > Global::gameViewX + Global::gameViewW - spriteWidth) {
		if(param.limitPolicy==SpriteParameter::BlockInScreen) {
			x_new = Global::gameViewX + Global::gameViewW - spriteWidth;
		} else
		if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
			//destroy the widget only when it becomes invisible
			//(with a margin of 5 to avoid problems when creating items)
			if(!isDestroyed && x_new > Global::gameViewX + Global::gameViewW + 5) {
				isDestroyed = true;
			}
		}
	}
	
	if(y_new < Global::gameViewY) {
		if(param.limitPolicy==SpriteParameter::BlockInScreen) {
			y_new = Global::gameViewY;
		} else
		if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
			//destroy the widget only when it becomes invisible
			//(with a margin of 5 to avoid problems when creating items)
			if(!isDestroyed && y_new < Global::gameViewY - spriteHeight - 5) {
				isDestroyed = true;
			}
		}
	}
	
	if(y_new > Global::gameViewY + Global::gameViewH - spriteHeight) {
		if(param.limitPolicy==SpriteParameter::BlockInScreen) {
			y_new = Global::gameViewY + Global::gameViewH - spriteHeight;
		} else
		if(param.limitPolicy==SpriteParameter::DestroyWhenOut) {
			//destroy the widget only when it becomes invisible
			//(with a margin of 5 to avoid problems when creating items)
			if(!isDestroyed && y_new > Global::gameViewY + Global::gameViewH + 5) {
				isDestroyed = true;
			}
		}
	}
}

//======================  Keyboard and mouse =========================

//initialize the keybaord state to all keys released
//without this initialization, the initial state of the keys is the state of the key when the previous level was finished
void Sprite::initKeyboardState() {
	upKeyPressed = false;
	downKeyPressed = false;
	leftKeyPressed = false;
	rightKeyPressed = false;
}

//return true if the key is accepted by the sprite
//we remember the last pressed key in x and y direction, so that we don't react to the release of the other key
//it gives better control because most players begin to press a key before releasing the previous one
bool Sprite::treatKey(int pressedKey, bool isPressed) {
	//we ignore key events if the player ship has already started exploding
	if(explodingCycle!=0) return false;
	
	switch (pressedKey) {
		case Qt::Key_Up:
			upKeyPressed = isPressed;
			if(isPressed) {
				param.y_speed = -playerYSpeed;
			} else {
				//up key was just released, go down if down key is pressed
				if(downKeyPressed) {
					param.y_speed = playerYSpeed;
				} else{
					param.y_speed = 0;
				}
			}
			break;
		case Qt::Key_Down:
			downKeyPressed = isPressed;
			if(isPressed) {
				param.y_speed = playerYSpeed;
			} else {
				//down key was just released, go up if up key is pressed
				if(upKeyPressed) {
					param.y_speed = -playerYSpeed;
				} else {
					param.y_speed = 0;
				}
			}
			break;
		case Qt::Key_Left:
			leftKeyPressed = isPressed;
			if(isPressed) {
				param.x_speed = -playerXSpeed;
			} else {
				//left key was just released, go right if right key is pressed
				if(rightKeyPressed) {
					param.x_speed = playerXSpeed;
				} else {
					param.x_speed = 0;
				}
			}
			break;
		case Qt::Key_Right:
			rightKeyPressed = isPressed;
			if(isPressed) {
				param.x_speed = playerXSpeed;
			} else {
				//right key was just released, go left if left key is pressed
				if(leftKeyPressed) {
					param.x_speed = -playerXSpeed;
				} else {
					param.x_speed = 0;
				}
			}
			break;
		case Qt::Key_Space:
			if(isPressed) {
				param.fireSpeed = playerFireSpeed;
				fire_cycle = playerFireSpeed;			//same value to shoot immediatly
			} else {
				param.fireSpeed = 0;
			}
			break;
		default:
			return false;
	}
	return true;
}


//We reimplement QGraphicsItem::mousePressEvent to catch the mouse initial position
//This will be used to compute the move coordinates
void Sprite::mousePressEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	mouseInitialCoordinates = mouseEvent->scenePos();
	spriteInitialX = param.x_pos;
	spriteInitialY = param.y_pos;
	emit spriteSelected(getParameters());			//signal the parameters of the sprite (used in edit mode)
	QGraphicsItem::mousePressEvent(mouseEvent);		//call the default implementation
}

//The default function QGraphicsItem::mouseMoveEvent doesn't allow us to know simply the coordinates change
// -> we reimplement the mouseMoveEvent to catch the change
// it also allows us to perform some operations (for example : forbid the movement out of the scene)
//note : the current version only takes into account the last selected item
//the code is directly inspired from QGraphicsItem::mouseMoveEvent
void Sprite::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	if ((mouseEvent->buttons() & Qt::LeftButton) && (flags() & ItemIsMovable)) {
		//compute the new position coordinates from the current mouse position
		QPointF translation = mouseEvent->scenePos() - mouseInitialCoordinates;
		param.x_pos = spriteInitialX + translation.x();
		param.y_pos = spriteInitialY + translation.y();
		
		//take into account the scene limit
		param.enforceEditSceneLimits(spriteWidth, spriteHeight);
		
		//set the new position
		setPos(param.x_pos, param.y_pos);
		emit spriteMoved(param);
	} else {
		mouseEvent->ignore();
	}
}

/*
void Sprite::mouseReleaseEvent(QGraphicsSceneMouseEvent *mouseEvent) {
	QGraphicsItem::mouseReleaseEvent(mouseEvent);		//call the default implementation
}
*/
