/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

 
#include "spriteparameter.h"
#include "../networkengine/filerepository.h"

#include <QPixmap>
#include <QObject>

//=================================== SpriteParameter =========================

int SpriteParameter::latestFormatVersion=0;

//default values of the parameters
SpriteParameter::SpriteParameter() {
	x_pos = 0;
	y_pos = 0;
	z_pos = 0;
	x_speed = 0;
	y_speed = 0;
	limitPolicy = BlockInScreen;
	collisionPolicy = NoCollision;
	fireIndex = 0;
	fireSpeed = 0;
	lifePoints = 1;
	emittingSpriteShape = -1;
}

//correct the x_pos and y_pos to ensure that the sprite is entirely in the scene
void SpriteParameter::enforceEditSceneLimits(int spriteWidth, int spriteHeight) {
	if(x_pos < Global::sceneXStart ) x_pos = Global::sceneXStart;
	if(x_pos > Global::sceneXEnd - spriteWidth) x_pos = Global::sceneXEnd - spriteWidth;
	if(y_pos < Global::sceneYStart ) y_pos = Global::sceneYStart;
	if(y_pos > Global::sceneYEnd - spriteHeight) y_pos = Global::sceneYEnd - spriteHeight;
}

//export the sprite parameters as a string
std::string SpriteParameter::toString() const {
	StringConverter conv;
	//"UT " = activate separator between elements
	//"E;" = use ; as the separator
	conv << CvCd("UT") << CvCd("E;");
	conv << spriteShape << x_pos << y_pos << z_pos << x_speed << y_speed;
	conv << limitPolicy << collisionPolicy << fireIndex << fireSpeed << lifePoints;
	return conv.getString();
}

//import the sprite parameters from a string (formated in the latest format version)
void SpriteParameter::fromString(const std::string& stringA) {
	fromString(stringA, latestFormatVersion);
}

//import the sprite parameters from a string
//the version number is for compatibility to read files from older versions of Flukz
void SpriteParameter::fromString(const std::string& stringA, int version) {
	StringConverter conv(stringA);
	if(version==0) {
		conv << CvCd("UT") << CvCd("E;");
		conv >> spriteShape >> x_pos >> y_pos >> z_pos >> x_speed >> y_speed;
		conv >> limitPolicy >> collisionPolicy >> fireIndex >> fireSpeed >> lifePoints;
	}
}

StringConverter& operator<<(StringConverter& conv, const SpriteParameter& p) {
	conv << p.spriteShape << p.x_pos << p.y_pos << p.z_pos << p.x_speed << p.y_speed;
	conv << p.limitPolicy << p.collisionPolicy << p.fireIndex << p.fireSpeed << p.lifePoints;
	return conv;
}

StringConverter& operator>>(StringConverter& conv, SpriteParameter& p) {
	conv >> p.spriteShape >> p.x_pos >> p.y_pos >> p.z_pos >> p.x_speed >> p.y_speed;
	conv >> p.limitPolicy >> p.collisionPolicy >> p.fireIndex >> p.fireSpeed >> p.lifePoints;
	return conv;
}

//limit the values to the correct range
//it can be used after the user changed the values
//return true if some value was out of range
bool SpriteParameter::limitValuesRange() {
	bool outOfRange = false;
	if(x_speed>10) {
		x_speed=10;
		outOfRange = true;
	}
	if(x_speed<-10) {
		x_speed=-10;
		outOfRange = true;
	}
	if(y_speed<-10) {
		y_speed=-10;
		outOfRange = true;
	}
	if(y_speed>10) {
		y_speed=10;
		outOfRange = true;
	}
	if(z_pos<-20) {
		z_pos = -20;
		outOfRange = true;
	}
	if(z_pos>20) {
		z_pos = 20;
		outOfRange = true;
	}
	
	if(limitPolicy!=BlockInScreen && limitPolicy!=DestroyWhenOut) {
		limitPolicy = DestroyWhenOut;
		outOfRange = true;
	}
	
	if(collisionPolicy!=NoCollision && collisionPolicy!=Destroyed &&
	   collisionPolicy!=DestroyedIndirectly)
	{
		collisionPolicy=NoCollision;
		outOfRange = true;
	}
	
	if(fireSpeed<0) {
		fireSpeed=0;
		outOfRange = true;
	}
	
	if(fireSpeed>2000) {
		fireSpeed=2000;
		outOfRange = true;
	}
	
	if(lifePoints<1) {
		lifePoints=1;
		outOfRange = true;
	}
	
	if(lifePoints>20000) {
		lifePoints=20000;
		outOfRange = true;
	}
	
	return outOfRange;
}

//return a list of int values that describe the type of the parameters of a sprite
//default : isRealSprite = true
//when part of spriteKind, isRealSprite = false and some parameters are read-only
QList<Parameter> SpriteParameter::getParameters(bool isRealSprite) {
	QList<Parameter> result;
	
	//spriteShape
	result += Parameter(Parameter::IntType,
						QObject::tr("Shape index"),
						QObject::tr("index refering to one of the sprite kind available in the toolbar"),
						Parameter::ReadOnly);
	//x_pos
	result += Parameter(Parameter::QrealType,
						QObject::tr("x coordinate"),
						QObject::tr("position of the sprite along the x axis"),
						isRealSprite);
	//y_pos
	result += Parameter(Parameter::QrealType,
						QObject::tr("y coordinate"),
						QObject::tr("position of the sprite along the y axis"),
						isRealSprite);
	//z_pos
	result += Parameter(Parameter::QrealType,
						QObject::tr("z coordinate"),
						QObject::tr("position of the sprite along the z axis"));
	//x_speed
	result += Parameter(Parameter::QrealType,
						QObject::tr("x speed"),
						QObject::tr("speed of the sprite along the x axis"));
	//y_speed
	result += Parameter(Parameter::QrealType,
						QObject::tr("y speed"),
						QObject::tr("speed of the sprite along the y axis"));
	//limitPolicy
	result += Parameter(Parameter::IntType,
						QObject::tr("Limit policy"),
						QObject::tr("behaviour of the sprite when reaching the limits of the window.\n0=blocked to the window limit, 1=go out of the window and destroyed"));
	//collisionPolicy
	result += Parameter(Parameter::IntType,
						QObject::tr("Collision policy"),
						QObject::tr("behaviour of the sprite when colliding with another sprite.\n0=no reaction to collisions, 1=destroyed, 2=destroyed indirectly.\nType 1 is destroyed by type 1 and type 2. Type 2 is destroyed only by type 1."));
	//fireIndex
	result += Parameter(Parameter::IntType,
						QObject::tr("Fire index"),
						QObject::tr("index refering to one of the sprite kind available in the toolbar."));
	//fireSpeed
	result += Parameter(Parameter::IntType,
						QObject::tr("Fire delay"),
						QObject::tr("delay between repeated fire. 0 = no fire."));
	//lifePoints
	result += Parameter(Parameter::IntType,
						QObject::tr("Life Points"),
						QObject::tr("number of times fire is needed to be destroyed"));
	return result;
}

//return the string command to initialize stringConverter correctly
std::string SpriteParameter::getSeparator() {
	return "E;";
}

//=================================== SpriteKindParameter =========================

//default values of the parameters
SpriteKindParameter::SpriteKindParameter() {
	spriteKind = 0;
	spriteWidth = 0;
	spriteHeight = 0;
	animationLength = 0;
	animationSpeed = 0;
}

void SpriteKindParameter::detectSpriteDimensions() {
	QPixmap pixmap(FileRepository::getAbsoluteFileName(QString::fromStdString(pixmapFileName)));
	spriteWidth = pixmap.width();
	spriteHeight = pixmap.height();
}

//export the parameters as a string
//Since SpriteParameter is an element of SpriteKindParameter, it is important to use
//a separator element ")" different from SpriteParameter ";" 
std::string SpriteKindParameter::toString() const {
	StringConverter conv;
	//"UT " = activate separator between elements
	//"E]" = use ] as the separator
	conv << CvCd("UT") << CvCd("E)");
	conv << spriteKind << defaultParameters.toString();
	conv << pixmapFileName << spriteWidth << spriteHeight;
	conv << animationFileName << animationLength << animationSpeed;
	return conv.getString();
}

//import the parameters from a string (formated in the latest format version)
void SpriteKindParameter::fromString(const std::string& stringA) {
	fromString(stringA, SpriteParameter::latestFormatVersion);
}

//import the parameters from a string
//the version number is for compatibility to read files from older versions of Flukz
void SpriteKindParameter::fromString(const std::string& stringA, int version) {
	StringConverter conv(stringA);
	if(version==0) {
		conv << CvCd("UT") << CvCd("E)");
		std::string defaultParamString;
		conv >> spriteKind >> defaultParamString;
		conv >> pixmapFileName >> spriteWidth >> spriteHeight;
		conv >> animationFileName >> animationLength >> animationSpeed;
		defaultParameters.fromString(defaultParamString);
	}
}

//export the parameters as a string with only ; separator
std::string SpriteKindParameter::toLinearString() const {
	StringConverter conv;
	//"UT " = activate separator between elements
	//"E;" = use ; as the separator
	conv << CvCd("UT") << CvCd("E;");
	conv << spriteKind << pixmapFileName << animationFileName << animationLength << animationSpeed;
	conv << defaultParameters;
	return conv.getString();
}

//import the parameters from a string
//the version number is for compatibility to read files from older versions of Flukz
void SpriteKindParameter::fromLinearString(const std::string& stringA) {
	StringConverter conv(stringA);
	//if(version==0) {
		conv << CvCd("UT") << CvCd("E;");
		conv >> spriteKind >> pixmapFileName >> animationFileName >> animationLength >> animationSpeed;
		conv >> defaultParameters;
	//}
}

//limit the values to the correct range
//it can be used after the user changed the values
//return true if some value was out of range
bool SpriteKindParameter::limitValuesRange() {
	bool outOfRange = defaultParameters.limitValuesRange();
	
	if(animationSpeed<0) {
		animationSpeed=0;
		outOfRange = true;
	}
	if(animationSpeed>100) {
		animationSpeed=100;
		outOfRange = true;
	}
	
	return outOfRange;
}

//return a list of int values that describe the type of the parameters of a sprite
QList<Parameter> SpriteKindParameter::getParameters() {
	QList<Parameter> result;
	
	//spriteKind
	result += Parameter(Parameter::IntType,
						QObject::tr("Sprite id"),
						QObject::tr("unique id to identify this kind of sprites"),
						Parameter::ReadOnly);
	//pixmap name
	result += Parameter(Parameter::StringType,
						QObject::tr("Pixmap name"),
						QObject::tr("name of the file that contains the sprite image"));
	//animation name
	result += Parameter(Parameter::StringType,
						QObject::tr("Animation name"),
						QObject::tr("generic name of the file that contains the animation image.\nIf the generic name is shipExplo.png, animation files will be shipExplo1.png, shipExplo2.png, ..."));
	//animation length
	result += Parameter(Parameter::IntType,
						QObject::tr("Animation length"),
						QObject::tr("number of files in the animation"));
	//animation speed
	result += Parameter(Parameter::IntType,
						QObject::tr("Animation delay"),
						QObject::tr("delay between two images of the animation"));
	//defaultParameters
	//argument means that this is not the parameters of a real sprite
	result += SpriteParameter::getParameters(false);
	return result;
}

//return the string command to initialize stringConverter correctly
std::string SpriteKindParameter::getSeparator() {
	return "E;";
}

StringConverter& operator<<(StringConverter& conv, const SpriteKindParameter& p) {
	conv << p.toString();
	return conv;
}

StringConverter& operator>>(StringConverter& conv, SpriteKindParameter& p) {
	std::string s;
	conv >> s;
	p.fromString(s);
	return conv;
}
