/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QFileInfo>
#include <QDateTime>

#include "clientobject.h"
#include "filerepository.h"

ClientObject::ClientObject(QObject *parent) :  QObject(parent) {
	userIsAuthenticated = false;
}

QString ClientObject::cleanString(const QString& s) {
	QString result = s;
	result.remove(' ');
	return result;
}

QStringList ClientObject::cleanString(const QStringList& listA) {
	QStringList result = listA;
	int i;
	for(i=0; i<result.size(); i++) {
		result[i].remove(' ');
	}
	return result;
}

void ClientObject::connectToServer(const QString& serverName) {
	Connection * connection = new Connection(this);
	connect(connection, SIGNAL(disconnected()), connection, SLOT(deleteLater()));
	connect(connection, SIGNAL(messageReceived(const Message&, Connection*)), this, SLOT(processMessage(const Message&, Connection*)) );
	connect(connection, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(processError(QAbstractSocket::SocketError)) );
	connection->connectToHost(serverName, Connection::serverDefaultPort);
}

void ClientObject::processError(QAbstractSocket::SocketError socketError) {
	switch (socketError) {
		case QAbstractSocket::RemoteHostClosedError:
			//the remote server closed the connection before us
			//TODO : there are cases when this should be an error
			break;
		case QAbstractSocket::HostNotFoundError:
			emit connectionError(Message::HostNotFoundError);
			break;
		case QAbstractSocket::ConnectionRefusedError:
			emit connectionError(Message::ConnectionRefusedError);
			break;
		case QAbstractSocket::SocketTimeoutError:
			emit connectionError(Message::SocketTimeoutError);
			break;
		default:
			emit connectionError(Message::UnknownError);
	}
}

void ClientObject::processMessage(const Message& message, Connection* connection) {
	if(message.commandCode == Message::BeginConnection) {
		//the connection with the server begins
		//process files or messages in the waiting list
		if(!waitingMessages.isEmpty()) {
			Message response = waitingMessages.takeFirst();
			if(userIsAuthenticated) response.setUser(userName, password);
			if(response.commandCode==Message::File) {
				//TODO : this is probably not the best place to do the test of existence
				QFile file(response.absoluteFileName);
				if(!file.exists()) return;
				emit permanentStatusMessage(tr("Sending %1 ...").arg(response.absoluteFileName));
			} else {
				emit permanentStatusMessage(response.defaultDisplayString());
			}
			connection->sendData(response);
		} else {
			//download mode
			downloadNextFileWith(connection);
		}
	} else
	if(message.commandCode == Message::FileHash) {
		for(int i=0; i<message.fileNameList.size(); i++) {
			QString shortFileName = message.fileNameList[i];
			bool downloadNeeded = true;
			if(message.fileHashList[i] == FileRepository::fileNotFoundHash) {
				//this file does not exist on the server, no need to try to download it
				downloadNeeded = false;
				failedFiles += shortFileName;
				emit temporaryStatusMessage(tr("%1 not found.").arg(shortFileName));
				
				//we also need to delete the local copy if existing
				FileRepository::fileHardRemove(shortFileName);
			} else {
				QFileInfo localFileInfo( FileRepository::getAbsoluteFileName(shortFileName) );
				if(localFileInfo.exists()) {
					//compare the received hash to the hash of the local file
					//we don't need to download again the server file if the local hash is the same
					QString localHash = FileRepository::fileHash(shortFileName);
					if(i < message.fileHashList.size()) {
						if(localHash == message.fileHashList[i]) {
							downloadNeeded = false;
							emit temporaryStatusMessage(tr("%1 is uptodate.").arg(shortFileName));
						}
					}
				}
			}
			
			if(downloadNeeded) filesNeededDownload.append(shortFileName);
		}
		
		//launch the real download of the files that need it
		downloadNextFileWith(connection);
	} else
	if(message.commandCode == Message::File) {
		//the requested file has been downloaded
		emit temporaryStatusMessage(message.defaultDisplayString());
		
		//download the next one in the list
		downloadNextFileWith(connection);
	} else
	if(message.commandCode == Message::FileNotFound) {
		//the requested file was not found on the server
		emit temporaryStatusMessage(message.defaultDisplayString());
		failedFiles += message.fileName;
		
		//download the next file in the list
		downloadNextFileWith(connection);
	} else
	if(message.commandCode == Message::InvalidProtocolVersion) {
		connection->disconnect();
		emit connectionError(Message::InvalidProtocolError);
	} else
	if(message.statusCode == Message::ActionSuccess ||
	   message.statusCode == Message::ActionFailed ||
	   message.statusCode == Message::UserAlreadyExists ||
	   message.commandCode == Message::FileReceived ||
	   message.commandCode == Message::WritingFailure ||
	   message.commandCode == Message::WritingForbidden
	) {
		connection->disconnect();
		emit temporaryStatusMessage(message.defaultDisplayString());
		emit networkMessage(message);
	}
}

void ClientObject::downloadNextFileWith(Connection* connection) {
	if(filesToDownload.isEmpty() && filesNeededDownload.isEmpty()) {
		//no more files to download, end of the connection
		connection->disconnect();
		
		//send the signal indicating that the download ended
		//we indicate in parameter the list of files not correctly downloaded
		emit downloadEnd(failedFiles);
		return;
	}
	
	if(!filesToDownload.isEmpty()) {
		//check the hash of all the files before download
		QStringList filesNeededHash = filesToDownload;
		filesToDownload.clear();
		askFileHashToServer(filesNeededHash, connection);
		return;
	}
	
	//download the next file in the list of needed download
	QString nextFileName = filesNeededDownload.takeFirst();
	askFileToServer(nextFileName, connection);
}

//send to the server a message requesting a given file
void ClientObject::askFileToServer(const QString& fileName, Connection* connection) {
	Message message;
	message.commandCode=Message::AskFile;
	message.fileName=fileName;
	if(userIsAuthenticated) message.setUser(userName, password);
	emit permanentStatusMessage(message.defaultDisplayString());
	connection->sendData(message);			//Request the file to the server
}

//send to the server a message requesting the hash of a list of files
void ClientObject::askFileHashToServer(const QStringList& fileNames, Connection* connection) {
	Message message;
	message.commandCode=Message::AskFileHash;
	message.fileNameList = fileNames;
	if(userIsAuthenticated) message.setUser(userName, password);
	emit permanentStatusMessage(message.defaultDisplayString());
	connection->sendData(message);
}

//launch a download of a specific file
void ClientObject::download(const QString& fileName, const QString& serverName) {
	failedFiles.clear();
	filesToDownload.clear();
	filesToDownload.append(cleanString(fileName));
	connectToServer(cleanString(serverName));
}

//launch the download of a list of files
void ClientObject::download(const QStringList& fileNames, const QString& serverName) {
	failedFiles.clear();
	filesToDownload.clear();
	filesToDownload.append(cleanString(fileNames));
	connectToServer(cleanString(serverName));
}

//add a new message to the list of waiting messages and start the connection to the given server
void ClientObject::addMessageAndConnect(const Message& message, const QString& serverName) {
	waitingMessages.append(message);
	connectToServer(cleanString(serverName));
}

//set the user and password
//they will be used for all communications until they are unset
void ClientObject::setUser(const QString & userName_init, const QString & password_init) {
	userName = userName_init;
	password = password_init;
	userIsAuthenticated = true;
}

//logout
void ClientObject::userLogout() {
	userName.clear();
	password.clear();
	userIsAuthenticated = false;
}
