/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "connection.h"
#include "filerepository.h"
#include "../global.h"

#include <QFile>
#include <QHostAddress>

//the magic number is used as a header of all communications
//in order to discard incorrect messages or very simple spam
quint32 Connection::magicNumber=1234*56+790;

quint16 Connection::serverDefaultPort=8042;
QString Connection::serverDefaultHostName=QString("game.flukz.org");

//messages are curently limited to around 3Mb
quint32 Connection::messageSizeLimit=3200000;

unsigned long Connection::receivedGigaBytes = 0;
unsigned long Connection::receivedBytes = 0;
unsigned long Connection::sentGigaBytes = 0;
unsigned long Connection::sentBytes = 0;

void Connection::resetBandwidthStat() {
	receivedGigaBytes = 0;
	receivedBytes = 0;
	sentGigaBytes = 0;
	sentBytes = 0;
}

void Connection::addReceivedBytes(quint32 receivedSize) {
	receivedBytes += receivedSize;
}

void Connection::addSentBytes(quint32 sentSize) {
	sentBytes += sentSize;
}

Connection::Connection( QObject *parent, QTcpSocket *socket) : QObject( parent ) {
	//initialize the internal objects to wait data
	initWaitState();
	
	//Tcpsocket object
	if(socket == 0) {
		//not connected socket (usually used in client mode, connectToHost is needed)
		tcpSocket = new QTcpSocket(this);
		isConnected=false;
	} else {
		//already connected socket (usually received from a QTcpServer object)
		tcpSocket = socket;
		isConnected = true;
	}
	
	connect(tcpSocket, SIGNAL(readyRead()), this, SLOT(receiveData()));
	connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(processError(QAbstractSocket::SocketError)));
	connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)), this, SIGNAL(error(QAbstractSocket::SocketError)));
	connect(tcpSocket, SIGNAL(connected()), this, SLOT(socketConnected()) );
	connect(tcpSocket, SIGNAL(disconnected()), this, SLOT(socketDisconnected()) );
}

//connect the tcpsocket to a new host
//any pending data in the current socket is lost
void Connection::connectToHost(const QString& hostName, quint16 port) {
	tcpSocket->abort();		//drop any pending data in the current socket
	tcpSocket->connectToHost(hostName, port);
	initWaitState();
	isConnected = true;
}

//called when tcpsocket is connected
void Connection::socketConnected() {
	isConnected=true;
	
	//indicate that the connection has been established
	Message message;
	message.commandCode=Message::BeginConnection;
	emit messageReceived(message, this);
}

//disconnect from the host
//when tcpsocket is disconnected, the socketDisconnected() slot will be called
void Connection::disconnect() {
	tcpSocket->disconnectFromHost();
}

//called when tcpsocket is disconnected
void Connection::socketDisconnected() {
	tcpSocket->deleteLater();
	isConnected=false;
	emit disconnected();
}

//initialize the object to receive new data
void Connection::initWaitState() {
	blockSize = 0;
	magicNumberReceived = false;
}

void Connection::sendData(const Message& message) {
	if(!isConnected) return;	//nothing to do if we are not connected
	
	QByteArray block;
	QDataStream networkData(&block, QIODevice::WriteOnly);
	networkData.setVersion(QDataStream::Qt_4_0);
	
	//the network data always begins with the magic number and the protocol version
	networkData << magicNumber;
	
	//4 bytes indicating the size of the data
	networkData << (quint32) 0;
	
	//network protocol version
	networkProtocolVersion = 0;		//current protocol version is 0
	networkData << networkProtocolVersion;
	
	//message content
	networkData << message;
	
	networkData.device()->seek(4);
	networkData << (quint32) (block.size() - /*magicNumber*/ sizeof(quint32) - /*bytes indicating the size*/ sizeof(quint32));
	//std::cout << block.size() - sizeof(quint32) - sizeof(quint32) << std::endl;
	addSentBytes(block.size());
	
	tcpSocket->write(block);
}

//TODO : the following code does not take into account the case of mutiple messages
//that would be received in the same block, some kind of while loop is needed
void Connection::receiveData() {
	if(!isConnected) return;	//nothing to do if we are not connected or waiting data
	
	QDataStream networkData(tcpSocket);
	networkData.setVersion(QDataStream::Qt_4_0);
	
	//discard everything that does not begin with the magic number
	if(!magicNumberReceived) {
		if (tcpSocket->bytesAvailable() < (int) sizeof(quint32)) return;
		magicNumberReceived = true;
		quint32 magicHeader;
		networkData >> magicHeader;
		if(magicHeader != magicNumber) {
			//this communication is not correct
			tcpSocket->abort();
			disconnect();
			return;
		}
	}
	
	if(blockSize == 0) {
		if (tcpSocket->bytesAvailable() < (int)sizeof(quint32)) return;
		networkData >> blockSize;
	}
	
	//if a message is too big, the connection is closed (for security)
	if(blockSize>messageSizeLimit) {
		tcpSocket->abort();
		disconnect();
		return;
	}
	
	if (tcpSocket->bytesAvailable() < blockSize) return;
	addReceivedBytes(blockSize + 2*sizeof(quint32));
	
	//network protocol version
	networkData >> networkProtocolVersion;
	if(networkProtocolVersion!=0) {
		//this protocol version is not valid
		Message response;
		response.commandCode = Message::InvalidProtocolVersion;
		response.statusCode = Message::ActionFailed;
		response.msg = QString("");	//could be customized if needed
		sendData(response);
		return;
	}
	
	Message message;
	
	//initialize the user to the IP address (will be overwritten if user is authenticated)
	message.user = tcpSocket->peerAddress().toString();
	
	//read the message content
	networkData >> message;
	
	//if user was rejected (wrong password), we abort the connection
	//TODO : it would be better to reply a message indicating that the specified user was not valid 
	if(message.userRejected) {
		tcpSocket->abort();
		disconnect();
		return;
	}
	
	//indicate that the data was received
	emit messageReceived(message, this);
	
	//wait the next data
	initWaitState();
}

void Connection::processError(QAbstractSocket::SocketError /*socketError*/) {
	disconnect();	//disconnect from the host
}
