/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "filerepository.h"
#include "../global.h"
#include "highscorefile.h"

#include <QFile>
#include <QTextStream>
#include <QDateTime>
#include <QDir>
#include <QCryptographicHash>

bool FileRepository::isClientSide=false;
bool FileRepository::isDefined=false;
QString FileRepository::repositoryPath;

//resources
//resource separator to access to resources (log, old, etc.) inside a node
QChar FileRepository::resourceSeparator=QChar('>');
QString FileRepository::logResource = resourceSeparator + QString("log");
QString FileRepository::oldResource = resourceSeparator + QString("old");
QString FileRepository::talkResource = resourceSeparator + QString("talk");
QString FileRepository::metaResource = resourceSeparator + QString("meta");
QString FileRepository::contribResource = resourceSeparator + QString("contrib");
QString FileRepository::highScoreResource = resourceSeparator + QString("highscore");

QString FileRepository::trashFolderName=QString("trash/");
QString FileRepository::protectedFolderName=QString("protected/");
QString FileRepository::userFolderName=QString("user/");
QString FileRepository::langFolderName=QString("lang/");
QString FileRepository::virtualFolderName=QString("virtual/");
QString FileRepository::valueFolderName=QString("value/");

QString FileRepository::statusPageName=QString("virtual/status");
QString FileRepository::listFilesPageName=QString("virtual/allpages");
QString FileRepository::recentChangesPageName=QString("virtual/recent") + FileRepository::logResource;

QString FileRepository::userContribExtension=FileRepository::contribResource+FileRepository::logResource;

QString FileRepository::hostName;
QString FileRepository::hostHash;

//particular "hash" values (not real hash)
QString FileRepository::notFixedHash=QString("notFixed");
QString FileRepository::fileNotFoundHash=QString("notFound");
QString FileRepository::invalidHashPassword=QString("invalidHash");

//data of the users
QMap<QString, UserData> FileRepository::userMap;
QString FileRepository::passwordSalt;

QMap<QString, int> FileRepository::latestVersionMap;

void FileRepository::initRepo(bool isClientSideIni) {
	isClientSide = isClientSideIni;
	
	//set the default repository (.flukData in the user home directory)
	setDefaultRepo();
	
	//initializations done only on the server side
	if(!isClientSide) {
		//create an empty recent changes file
		//simply getting the name will create the file
		QString recentFileName = getAbsoluteFileName(recentChangesPageName, FileWrite);
		
		//read the user table
		readUserFile();
		
		//set the salt to hash passwords
		passwordSalt = QString("MyPreciousRecursivePasswordSalt");
	}
}

void FileRepository::setDefaultRepo() {
	//choose a directory
	QDir homeDir(QDir::homePath());
	QString defaultName(".flukzData");
	if(!isClientSide) defaultName = QString("flukzServerData");
	
	if(!homeDir.cd(defaultName)) {
		//create .flukData subdirectory in the user home
		bool success = homeDir.mkdir(defaultName);
		if(!success) {
			repositoryPath=QString("");
			isDefined = false;
		}
	}
	
	repositoryPath = QDir::homePath()+ QString("/") + defaultName;
	isDefined = true;
}

//set the current host (client side only)
void FileRepository::setCurrentHostName(QString currentHost) {
	hostName = currentHost;
	if(!currentHost.isEmpty()) {
		hostHash = hexMD5Hash(currentHost);
	} else {
		//case of an empty host, we use also an empty hash
		hostHash = currentHost;
	}
}

//return the absolute path to the folder used for the current host
//this is similar to getAbsoluteFileName but without a specific file name
QString FileRepository::getCurrentHostCacheFolder() {
	if(isDefined) {
		QDir directory(repositoryPath);
		if(!hostHash.isEmpty()) {
			directory.cd(hostHash);
		}
		return directory.absolutePath();
	} else {
		//current repository is not defined
		//this should not be used
		return QString("");
	}
}

//======================== Operations on the file names ===========================

//return the absolute file name corresponding to a file in the repository
//fileOpenMode indicates whether the file will be accessed for writing or not
//If writing is needed: * we create the path to the folder if it doesn't exist yet
//                      * on server side, we create the log file
QString FileRepository::getAbsoluteFileName(const QString& shortFileName, FileOpenMode fileOpenMode) {
	if(!isDefined) {
		//if file repository is not defined, simply use the shortFileName
		//files will be written in the "current" directory of the system
		//this should usually not be used
		return shortFileName;
	}
	
	//remove the first "/" if any, in order to construct a path relative to the directory of the repository
	QString cleanedName = shortFileName;
	if(cleanedName.startsWith("/")) cleanedName.remove(0, 1);
	
	//detect first log and old resources
	int dotIndex = cleanedName.lastIndexOf(resourceSeparator);
	QString extension;
	bool isResource = false;
	if(dotIndex!=-1) {
		extension = cleanedName.mid(dotIndex);
		if(extension==logResource || extension.startsWith(oldResource)) {
			//remove these resources
			cleanedName = cleanedName.left(dotIndex);
			extension.remove(0,1);
			isResource = true;
		} else {
			//other extensions are ignored at this step
			extension.clear();
		}
	}
	
	//detect talk, meta and contrib resources
	//contrib is considered as a resource only in the case of the contrib>log case
	dotIndex = cleanedName.lastIndexOf(resourceSeparator);
	QString extensionB;
	if(dotIndex!=-1) {
		extensionB = cleanedName.mid(dotIndex);
		if(extensionB==talkResource || extensionB==metaResource || extensionB==highScoreResource ||
		   (extensionB==contribResource && extension=="log")
		  ) {
			cleanedName = cleanedName.left(dotIndex);
			extensionB.remove(0,1);
			if(extension.isEmpty()) {
				extension = extensionB;
			} else {
				//don't use the separator character here
				//just in case it is not supported by the underlying file system
				extension = extensionB + "." + extension;
			}
			isResource = true;
		} else {
			//other extensions are ignored
			extensionB.clear();
		}
	}
	
	//hash sha1 of the short file name
	QString pathHashName = hexSHA1Hash(cleanedName) + "/";
	
	//add the hostHash repository if defined (client side to distinguish hosts)
	if(!hostHash.isEmpty()) {
		pathHashName = hostHash + "/" + pathHashName;
	}
	
	//creation of the path if writing is needed
	QDir directory(repositoryPath);
	QString absoluteName = directory.absoluteFilePath(pathHashName);
	if(!absoluteName.endsWith("/")) absoluteName.append("/");
	
	if(fileOpenMode==FileWrite) {
		directory.mkpath(absoluteName);
		
		//create the log file if it does not exist yet (only on server side)
		//some files don't need any log file (highscore files)
		if(!isClientSide && extensionB!=highScoreResource) {
			QString logFileName = absoluteName;
			QString plainName = cleanedName;
			QString logExtension = logResource;
			logExtension.remove(0,1);
			if(extensionB.isEmpty()) {
				logFileName += logExtension;
			} else {
				logFileName += extensionB + "." + logExtension;
				plainName = extensionB;
			}
			
			createLogFile(logFileName, plainName);
		}
	} else
	if(fileOpenMode==HashedFolderName) {
		return absoluteName;
	}
	
	//"main" is used as the name of the main file inside a node
	//the extension is used as the name for all the other files inside the node
	if(isResource) {
		absoluteName += extension;
	} else {
		absoluteName += QString("main");
	}
	
	return absoluteName;
}

//return the name of a file that can be used for temporary operations
QString FileRepository::temporaryFileName() {
	return getAbsoluteFileName(protectedFolderName+"temporaryFile.fkz", FileWrite);
}

//this function returns a "cleaned" string of a file name
//it resolves the ../ folders and remove a possible initial '/'
QString FileRepository::cleanFileName(const QString& shortFileName) {
	QString cleanedName = QDir::cleanPath(shortFileName);
	if(cleanedName.startsWith("/")) cleanedName.remove(0, 1);
	return cleanedName;
}

//return the short name of the log file associated to a given file
//we suppose that the given short file name is "clean" (no initial /)
QString FileRepository::getLogFileName(const QString& shortFileName) {
	return shortFileName + logResource;
}

//return the short name of a high score file
//we suppose that the given short file name is "clean" (no initial /)
QString FileRepository::getHighScoreFileName(const QString& shortFileName) {
	return shortFileName + highScoreResource;
}

//return the short name of the file containing the list of contributions of a given user
QString FileRepository::getUserContribFileName(const QString& user) {
	return userFolderName + user + userContribExtension;
}

//indicate if a page is editable or not
//(by a normal user, we don't take rights into account here)
bool FileRepository::isEditablePage(const QString& shortFileName) {
	QString cleanName = shortFileName;
	if(cleanName.startsWith('/')) cleanName.remove(0, 1);
	if(cleanName.endsWith(logResource) || isOldFile(cleanName) ||
	   cleanName.startsWith(trashFolderName) || isVirtualPage(cleanName)) {
		return false;
	} else {
		return true;
	}
}

//remove the spaces from a string and replace them with an underscore
QString FileRepository::nameWithoutSpaces(const QString&fileName) {
	QString s =	fileName;
	s.replace(" ", "_");
	return s;
}

//======================== Old files ===========================

//return the short name of an old file for a given file and given revision
//we suppose that the given short file name is "clean" (no initial /)
QString FileRepository::getOldFileName(const QString& shortFileName, int fileRevision) {
	return shortFileName + oldResource + QString("%1").arg(fileRevision);
}

//a file is an old revision if the last resource is old*
bool FileRepository::isOldFile(const QString& shortFileName) {
	int dotIndex = shortFileName.lastIndexOf(resourceSeparator);
	if(dotIndex!=-1) {
		QString extension = shortFileName.mid(dotIndex);
		if(extension.startsWith(oldResource)) {
			return true;
		}
	}
	return false;
}

//set the latest version associated to a given file in the latest version map
void FileRepository::setFileLatestVersion(const QString& shortFileName, int version) {
	QString cleanedName = shortFileName;
	if(cleanedName.startsWith("/")) cleanedName.remove(0, 1);
	latestVersionMap.insert(cleanedName, version);
}

//return the latest version associated to a given file (as stored in the latest version map)
int FileRepository::getFileLatestVersion(const QString& shortFileName) {
	QMap<QString, int>::iterator fileEntry = latestVersionMap.find(shortFileName);
	if(fileEntry != latestVersionMap.end()) {
		return fileEntry.value();
	}
	
	//the file is not registered in the map
	return voidVersion;
}

//find the latest version number of a given file inside the associated log file
int FileRepository::findFileLatestVersion(const QString& shortFileName) {
	QString cleanedName = shortFileName;
	if(cleanedName.startsWith("/")) cleanedName.remove(0, 1);
	
	QString logFileName = getLogFileName(cleanedName);
	QFile file(getAbsoluteFileName(logFileName));
	
	//read the content of the old recent file
	if ( !file.open(QIODevice::ReadOnly) ) return voidVersion;	//there is no latest version
	QTextStream text(&file);
	text.setCodec("UTF-8");
	QString skippedLine;
	skippedLine = text.readLine(); //skip the header line
	skippedLine = text.readLine(); //skip the version line
	skippedLine = text.readLine(); //skip the plain name
	
	int version = voidVersion;
	while(!text.atEnd()) {
		text >> version;
		skippedLine = text.readLine(); //skip the rest of the line
	}
	file.close();
	
	return version;
}

//======================== Check user rights to perform an action ===========================

//return true if the reading is authorized
bool FileRepository::isReadingAuthorized(const QString& shortFileName, const QString& user) {
	if(isClientSide) return false;	//in the current protocol, the client never accepts requests from the network
	
	QString cleanedName = cleanFileName(shortFileName);
	if(cleanedName.startsWith("..")) return false;	//security : forbid to go "back" the repository folder
	
	//forbid access to the protected folder (table with the passwords)
	if(cleanedName.startsWith(protectedFolderName)) return false;
	
	//check that the resource is valid
	if(!isValidResource(cleanedName)) return false;
	
	//no additional rule if the user is an admin (he can read the trash folder)
	if(checkUserRights(UserData::AdminLevel, user)) return true;
	
	//normal users cannot read the trash folder
	if(cleanedName.startsWith(trashFolderName)) return false;
	
	//check that the user is not black-listed, he needs at least read-only access
	if(checkUserRights(UserData::ReadOnly, user)) return true;
	
	return false;
}

//return true if the writing is authorized
bool FileRepository::isWritingAuthorized(const QString& shortFileName, const QString& user) {
	//direct writing is not authorized for high score resources on the server side (merging must be used)
	//it is authorized on the client side because we need to be able to write it after download
	if(!isClientSide && shortFileName.endsWith(highScoreResource)) return false;
	
	QString cleanedName = shortFileName;
	bool isAuthorized = checkWritingRules(cleanedName, user);
	return isAuthorized;
}

//return true if the merging is authorized ()
bool FileRepository::isMergingAuthorized(const QString& shortFileName, const QString& user) {
	//merging is possible only for highscore resources
	if(!shortFileName.endsWith(highScoreResource)) return false;
	
	QString cleanedName = shortFileName;
	bool isAuthorized = checkWritingRules(cleanedName, user);
	return isAuthorized;
}

bool FileRepository::isDeletingAuthorized(const QString& shortFileName, const QString& user) {
	if(isClientSide) return false;		//in the current protocol, the client never accepts requests from the network
	
	//check that the user is at least an admin
	bool isAuthorized = checkUserRights(UserData::AdminLevel, user);
	if(!isAuthorized) return false;
	
	QString cleanedName = shortFileName;
	isAuthorized = checkWritingRules(cleanedName, user);
	return isAuthorized;
}

bool FileRepository::isRenamingAuthorized(const QString& shortFileName, const QString& newFileName, const QString & user) {
	if(isClientSide) return false;		//in the current protocol, the client never accepts requests from the network
	
	//check that the user is at least a power user
	bool isAuthorized = checkUserRights(UserData::PowerUserLevel, user);
	if(!isAuthorized) return false;
	
	QString oldFileCleanedName = shortFileName;
	isAuthorized = checkWritingRules(oldFileCleanedName, user);
	if(!isAuthorized) return false;
	
	QString newFileCleanedName = newFileName;
	isAuthorized = checkWritingRules(newFileCleanedName, user);
	return isAuthorized;
}

//return true if writing is authorized
//Note : this function modifies fileName, and is used also for authorizing renaming and deleting
bool FileRepository::checkWritingRules(QString& fileName, const QString& user) {
	fileName = cleanFileName(fileName);
	if(fileName.startsWith("..")) return false;	//security : forbid to go "back" the repository folder
	
	//spaces are not authorized in page names
	if(nameWithoutSpaces(fileName) != fileName) return false;
	
	if(isClientSide) return true;		//case of the client side : no other rules
	
	//read-only pages on the server side (even for admin)
	if(isOldFile(fileName)) return false;
	if(fileName.endsWith(logResource)) return false;
	if(fileName.startsWith(protectedFolderName)) return false;
	
	//virtual pages are read-only, writing is forbidden (status page, listFiles page...)
	if(isVirtualPage(fileName)) return false;
	
	//invalid resources cannot be created
	if(!isValidResource(fileName)) return false;
	
	//no other rules for admins (they have access to user pages and trash folder)
	if(checkUserRights(UserData::AdminLevel, user)) return true;
	
	//normal users cannot access to the trash folder
	if(fileName.startsWith(trashFolderName)) return false;
	
	//for normal users, only the user can access to the article pages located in its user folder
	//however talk pages can be accessed by everyone
	if(fileName.startsWith(userFolderName)) {
		if(!fileName.startsWith(userFolderName+user) && !fileName.endsWith(talkResource)) {
			return false;
		}
	}
	
	//check that the user is not black-listed, he needs at least default ip-user level
	if(checkUserRights(UserData::IPUserLevel, user)) return true;
	
	return false;
}

//return true if the resource is valid (log, old, meta, talk, etc...)
bool FileRepository::isValidResource(const QString &fileName) {
	int dotIndex = fileName.indexOf(resourceSeparator);
	if(dotIndex==-1) return true;	//no resource
	
	QString resource = fileName.mid(dotIndex);
	//list of the valid resources
	if(resource==logResource || resource==talkResource || resource==metaResource ||
	   resource==highScoreResource || resource==userContribExtension
	   ) {
		return true;
	}
	
	//additional rules for talk and meta that can be themselves followed by log and old resources
	if(resource.startsWith(talkResource) || resource.startsWith(metaResource)) {
		//remove the first separator character and find the next one
		resource.remove(0,1);
		dotIndex = resource.indexOf(resourceSeparator);
		resource = resource.mid(dotIndex);
		
		//case of ">talk>log" and "meta>log"
		if(resource==logResource) return true;
	}
	
	//case of ">old??", ">talk>old??" and ">meta>old??"
	if(resource.startsWith(oldResource)) {
		resource.remove(0, oldResource.length());
		
		//the following characters should be only numbers
		bool ok;
		int a = resource.toInt(&ok);
		if(a>=0 && ok==true) return true;
	}
	
	return false;
}

//============================== Actions on a file =================================

//return true if the given file exists in the repository
bool FileRepository::fileExists(QString shortFileName) {
	QFile file(getAbsoluteFileName(shortFileName));
	return file.exists();
}

//return the sha1 hash in hex format of the given file
//return an empty string if the file does not exist
QString FileRepository::fileHash(QString shortFileName) {
	if(shortFileName.startsWith("/")) shortFileName.remove(0, 1);
	QFile file(getAbsoluteFileName(shortFileName));
	
	//TODO : reading all the file at once is heavy for memory (in case of big files)
	if(!file.open(QIODevice::ReadOnly)) return QString("");
	QByteArray fileData = file.readAll();
	QByteArray hashResult = QCryptographicHash::hash(fileData, QCryptographicHash::Sha1);
	file.close();
	
	QString result(hashResult.toHex());
	return result;
}

//remove the given file from the repository
//the file is not deleted from the repository, it is moved to the trash folder
//we also move the old files (history) and the log file
//return a value indicating failure or success of the operation
Message::status FileRepository::fileRemove(QString shortFileName, QString comment, const QString & user) {
	if(shortFileName.startsWith("/")) shortFileName.remove(0, 1);
	
	//move the main file to trash folder
	QString currentName = getAbsoluteFileName(shortFileName, HashedFolderName);
	QString shortNewName = trashFolderName + shortFileName;
	QString newName = getAbsoluteFileName(shortNewName, HashedFolderName);
	if(!QFile::rename(currentName, newName) ) return Message::ActionFailed;
	
	//update the log file with the new plain name
	updateLogPlainName(shortNewName);
	
	comment.insert(0, QString("Deleted %1. ").arg(shortFileName));
	updateFileHistory(shortNewName, comment, user);
	
	return Message::ActionSuccess;
}

//rename shortFileName in newFileName
//return a value indicating failure or success of the operation
Message::status FileRepository::fileRename(QString shortFileName, QString newFileName, QString comment, const QString & user) {
	if(shortFileName.startsWith("/")) shortFileName.remove(0, 1);
	if(newFileName.startsWith("/")) newFileName.remove(0, 1);
	
	//rename the file
	QString currentName = getAbsoluteFileName(shortFileName, HashedFolderName);
	QString newName = getAbsoluteFileName(newFileName, HashedFolderName);
	if(!QFile::rename(currentName, newName)) return Message::ActionFailed;
	
	//update the log file with the new plain name
	updateLogPlainName(newFileName);
	
	comment.insert(0, QString("Renamed %1 to %2. ").arg(shortFileName).arg(newFileName));
	updateFileHistory(newFileName, comment, user);
	
	return Message::ActionSuccess;
}

//revert a given file to a given revision
//return a value indicating failure or success of the operation
Message::status FileRepository::fileRevert(QString shortFileName, qint16 fileRevision, QString comment, const QString & user) {
	if(shortFileName.startsWith("/")) shortFileName.remove(0, 1);
	QString fileName = getAbsoluteFileName(shortFileName);
	
	//check if the old revision file exists
	QString oldFileName = getAbsoluteFileName( getOldFileName(shortFileName, fileRevision) );
	if(!QFile::exists(oldFileName))return Message::ActionFailed;
	
	//remove the current file and copy the target old revision in its place
	QFile::remove(fileName);
	QFile::copy(oldFileName, fileName);
	
	//add the change to history and recent history
	comment.insert(0, QString("Reverted to revision %1. ").arg(fileRevision));
	updateFileHistory(shortFileName, comment, user);
	
	return Message::ActionSuccess;
}

//really remove a file from the repository
//WARNING : this operation is not logged and cannot be canceled
Message::status FileRepository::fileHardRemove(QString shortFileName) {
	//currently, it can be used only on the clientside (to remove obsolete local copies)
	if(!isClientSide) return Message::ActionFailed;
	
	if(shortFileName.startsWith("/")) shortFileName.remove(0, 1);
	QString fileName = getAbsoluteFileName(shortFileName);
	QFile::remove(fileName);
	return Message::ActionSuccess;
}

//========================= Log files =========================

//create an empty log file with header, version and "plain name" of the target file
//the file is created only if it does not exist yet
void FileRepository::createLogFile(const QString& logFileName, const QString& targetFilePlainName) {
	QFile file(logFileName);
	if(file.exists()) return;	//file already exist, nothing to do
	
	if( !file.open(QIODevice::WriteOnly) ) return;	//error case
	QTextStream text(&file);
	text.setCodec("UTF-8");
	
	QString headerName("[Flukz Revision File]");
	text << headerName << "\n";
	int stringFormatVersion = 0;
	text << stringFormatVersion << "\n";
	text << targetFilePlainName << "\n";
	file.close();
}

//update the plain name that is written inside the log file associated to a given file
//it is necessary for example after a rename or a delete operation
//note : this function is a bit heavy because it is not possible to "insert" directly content in a file
void FileRepository::updateLogPlainName(const QString& shortFileName) {
	QString logFileName = getLogFileName(shortFileName);
	QFile file(getAbsoluteFileName(logFileName, FileWrite));
	
	//read the content of the current log file
	if ( !file.open(QIODevice::ReadOnly) ) return;	//error case
	QTextStream text(&file);
	text.setCodec("UTF-8");
	QString header = text.readLine();
	QString version = text.readLine();
	QString plainName = text.readLine();	//plainName not used, possible warning
	QString content = text.readAll();
	file.close();
	
	//overwrite the log file with changing only the plainName stored inside it
	if( !file.open(QIODevice::WriteOnly) ) return;	//error case
	text << header << "\n";
	text << version << "\n";
	text << shortFileName << "\n";	//new plain name
	text << content;
	file.close();
}

//update the revision history of a given file
//this operation is done after a file has been changed in the main directory
// (1) copy the latest version as an old file
// (2) update the log file
// (3) update the recentChanges file
//Important note : we don't create revision history on the client side
void FileRepository::updateFileHistory(const QString& shortFileName, const QString& comment, const QString & user) {
	//case of the client side : no revision history
	if(isClientSide) return;
	
	QString cleanedName = cleanFileName(shortFileName);
	QString absoluteFileName = getAbsoluteFileName(cleanedName, FileWrite);
	
	//find the first unused index for old files
	int i=-1;
	bool oldFileExist = true;
	QString oldFileName;
	//TODO : this algorithm is heavy for the server
	do {
		i++;
		oldFileName = getAbsoluteFileName( getOldFileName(cleanedName, i), FileWrite);
		oldFileExist = QFile::exists(oldFileName);
	} while(oldFileExist);
	
	QFile::copy(absoluteFileName, oldFileName);		//copy the current file in an old file
	
	//update the log file (add one line)
	addFileHistory(cleanedName, i, comment, user);
	
	//update the contribution files of the user
	addContribHistory(cleanedName, i, comment, user);
	
	//update the recent changes global file
	addRecentHistory(cleanedName, i, comment, user);
}

//add a new line with the comment of the changes in the log corresponding to a given file
//format : index - date - time - comment
void FileRepository::addFileHistory(const QString& shortFileName, int index, const QString& comment, const QString & user) {
	QString logFileName = getLogFileName(shortFileName);
	QFile file(getAbsoluteFileName(logFileName, FileWrite));
	
	if ( !file.open(QIODevice::Append) ) return;	//error case
	QTextStream text(&file);
	text.setCodec("UTF-8");
	
	text << index << " ";
	text << QDateTime::currentDateTime().toTime_t() << " ";
	text << user << " ";
	text << comment << "\n";
	
	file.close();
}

//add a new line in the contribution file of the given user
//format : file name - index - time since epoch - user - comment
//TODO : we don't need the user field
void FileRepository::addContribHistory(const QString& shortFileName, int index, const QString& comment, const QString & user) {
	QString contribFileName = getUserContribFileName(user);
	QString absoluteName = getAbsoluteFileName(contribFileName, FileWrite);
	
	//create the contrib file if it does not exist yet
	createLogFile(absoluteName, contribFileName);
	
	QFile file(absoluteName);
	if ( !file.open(QIODevice::Append) ) return;	//error case
	QTextStream text(&file);
	text.setCodec("UTF-8");
	
	//add the new recent history
	text << shortFileName << " ";
	text << index << " ";
	text << QDateTime::currentDateTime().toTime_t() << " ";
	text << user << " ";
	text << comment << "\n";
	
	file.close();
}

//add a new line in the "recent changes" file with the comment of the changes
//format : file name - index - date - time - comment
//TODO : some redundant code with addFileHistory
void FileRepository::addRecentHistory(const QString& shortFileName, int index, const QString& comment, const QString & user) {
	QString recentFileName = getAbsoluteFileName(recentChangesPageName, FileWrite);
	QFileInfo fileInfo(recentFileName);
	QFile file(recentFileName);
	
	//read the content of the old recent file
	if ( !file.open(QIODevice::ReadOnly) ) return;	//error case
	QTextStream text(&file);
	text.setCodec("UTF-8");
	QString content, skippedLine;
	skippedLine = text.readLine(); //skip the header line
	skippedLine = text.readLine(); //skip the version line
	QString plainName = text.readLine(); //read the plain name
	if(fileInfo.size() > 20000) {
		//if the file is bigger than 20 ko, skip the first line of data
		skippedLine = text.readLine();
	}
	content = text.readAll();
	file.close();
	
	//write the content of the new recent file
	if ( !file.open(QIODevice::WriteOnly) ) return;	//error case
	QTextStream newText(&file);
	newText.setCodec("UTF-8");
	
	//header, version, plain name
	QString headerName("[Flukz Revision File]");
	newText << headerName << "\n";
	int stringFormatVersion = 0;
	newText << stringFormatVersion << "\n";
	newText << plainName << "\n";
	
	//write the old content at the beginning of the file
	newText << content;
	
	//add the new recent history
	newText << shortFileName << " ";
	newText << index << " ";
	newText << QDateTime::currentDateTime().toTime_t() << " ";
	newText << user << " ";
	newText << comment << "\n";
	
	file.close();
}

//======================================== Virtual pages ======================================
//virtual pages are pages that don't correspond to a file on the disk
//they are created on the flight from other data

//return true is a given page name is a virtual page
//i.e when the path starts by the virtual folder name
bool FileRepository::isVirtualPage(const QString& shortFileName) {
	return (shortFileName.startsWith(virtualFolderName));
}

//return a list of all the files in the repository
//for each folder corresponding to a page, we read the plain name of the page in the log file
//currentDir = repositoryPath by default
//TODO : this function will be too heavy in the future when there is a lot of files
QStringList FileRepository::listAllFiles(QString currentDir) {
	QDir directory(currentDir);
	QStringList result;
	
	//each folder in the repository corresponds to a page (folder name = hash of the page name)
	QStringList subdirectories = directory.entryList(QDir::Dirs | QDir::NoDotAndDotDot, QDir::Name);
	int i, n;
	n=subdirectories.length();
	for(i=0; i<n; i++) {
		//read the content of the log file to find the "plain" (not hashed) name of the page
		QString logFileName = directory.absoluteFilePath(subdirectories[i]+ QDir::separator() + "log");
		QFile file(logFileName);
		if (file.open(QIODevice::ReadOnly) ) {
			QTextStream text(&file);
			text.setCodec("UTF-8");
			QString plainName = text.readLine(); //skip the header line
			plainName = text.readLine(); //skip the version line
			plainName = text.readLine(); //read the plain name
			if(!plainName.isEmpty() && !plainName.startsWith(protectedFolderName) && !plainName.startsWith(virtualFolderName)) {
				result += plainName;
			}
			file.close();
		}
	}
	
	return result;
}

//status page
QString FileRepository::statusPage() {
	QString info;
	info += QString("Registered users:%1\n").arg(userMap.size());
	
	return info;
}

//======================================== Value pages ======================================

//return a string containing the value associated to a given page name
QString FileRepository::getValuePage(QString pageName) {
	QString result("");
	
	if(!pageName.startsWith(valueFolderName)) return result;	//invalid case
	pageName.remove(0, valueFolderName.length());
	
	if(pageName.startsWith(userFolderName)) {
		pageName.remove(0, userFolderName.length());
		int pos = pageName.indexOf('/');
		if(pos==-1) return result;				//invalid case
		QString userName = pageName.left(pos);
		QString field = pageName.mid(pos+1);
		
		if(field=="rights") {
			int userRights = getUserRights(userName);
			return QString("%1").arg(userRights);
		}
	}
	
	//default case (not a valid case)
	return result;
}

Message::commands FileRepository::setValuePage(QString pageName, const QString &content, const QString& actionUser) {
	Message::commands resultStatus=Message::WritingFailure;
	
	if(!pageName.startsWith(valueFolderName)) return resultStatus;	//invalid case
	pageName.remove(0, valueFolderName.length());
	
	if(pageName.startsWith(userFolderName)) {
		pageName.remove(0, userFolderName.length());
		int pos = pageName.indexOf('/');
		if(pos==-1) return resultStatus;				//invalid case
		QString userName = pageName.left(pos);
		QString field = pageName.mid(pos+1);
		
		if(field=="rights") {
			bool ok;
			int userRights = content.toInt(&ok);
			if(ok) {
				return setUserRights(userName, userRights, actionUser);
			}
		}
	}
	
	//default case (not a valid case)
	return resultStatus;
}

//======================================== Merge data ======================================

Message::commands FileRepository::mergeData(QString shortFileName, const QString& data, const QString& user, const QString& msg) {
	//check that writing a file in the repository is authorized
	if(!FileRepository::isMergingAuthorized(shortFileName, user)) {
		return Message::WritingForbidden;
	}
	
	//create the path to write the file
	QString absoluteFileName = FileRepository::getAbsoluteFileName(shortFileName, FileRepository::FileWrite);
	HighScoreFile scoreFile(absoluteFileName);
	
	//msg contains the name entered by the user to save the score
	QString scoreName = msg;
	if(msg!=user) {
		scoreName.insert(0, QString("Guest_"));
	}
	
	bool recordInserted = scoreFile.addRecord(scoreName, data.toInt());
	if(!recordInserted) return Message::WritingFailure;
	
	scoreFile.writeFile(absoluteFileName);
	return Message::FileReceived;
}

//======================================== User rights ======================================

//return a username formatted with some rules
//(1) first letter and letter after a '_' is upper case, other letters are lower case
//(2) characters that are not letters or numbers are replaced by a '_'
//(3) no first '_' and no '_' following a '_'
//(4) first character cannot be a number (to avoid any conflict with ip address)
QString FileRepository::formatUserName(const QString& userName) {
	QString result;
	QChar underScore('_');
	int n = userName.size();
	for(int i=0; i<n; i++) {
		QChar c = userName[i];
		if(c.isLetterOrNumber()) {
			if(result.isEmpty() || result.endsWith(underScore)) { //rule (1)
				if(!result.isEmpty() || c.isLetter()) { //rule (4)
					result += c.toUpper();
				}
			} else { //rule (1)
				result += c.toLower();
			}
		} else { //rule (2)
			if(!result.isEmpty() && !result.endsWith(underScore)) { //rule (3)
				result += underScore;
			}
		}
	}
	return result;
}

//Read the file containing the data with the user names and passwords
void FileRepository::readUserFile() {
	//open file
	QString filename = getAbsoluteFileName(protectedFolderName+"usersPasswords.txt");
	QFile file( filename );
	if ( !file.open( QIODevice::ReadOnly ) ) return;
	
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	//read header
	if (!text.atEnd() ) {
		QString headerName("[Flukz User File]");
		if (text.readLine()!=headerName) {
			//incorrect format
			file.close();
			return;
		}
	}
	
	//read string format version
	int stringFormatVersion=-1;
	if (!text.atEnd() ) {
		QString line = text.readLine();
		stringFormatVersion = line.toInt();
	}
	
	if(stringFormatVersion!=0) {
		//unsupported version
		file.close();
		return;
	}
	
	//read the user names and passwords
	QString userName;
	UserData userData;
	while(!text.atEnd() ) {
		text >> userName;
		text >> userData.password;
		text >> userData.rights;
		
		//simple security in case of a new line or spaces at the end of the file
		userName.remove(' ');	//remove the spaces from the user name
		if(!userName.isEmpty()) {
			userMap.insert(userName, userData);
		}
	}
	
	//close file
	file.close();
}

//Write the file containing the data with the user names and passwords
void FileRepository::writeUserFile() {
	//open file
	QString filename = getAbsoluteFileName(protectedFolderName+"usersPasswords.txt", FileWrite);
	QFile file( filename );
	if ( !file.open( QIODevice::WriteOnly ) ) return;
	
	//write database
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	//header
	QString headerName("[Flukz User File]");
	text << headerName << "\n";
	
	//current string format version
	int stringFormatVersion = 0;
	text << stringFormatVersion << "\n";
	
	//save the users data
	QMap<QString, UserData>::iterator iter;
	
	for(iter=userMap.begin(); iter!=userMap.end(); iter++) {
		text << iter.key() << " " << iter.value().password << " " << iter.value().rights << "\n";
	}
	
	//close file
	file.close();
}

//return true if the given password is valid for the given user
Message::status FileRepository::checkUserPassword(const QString& user, const QString& password) {
	QMap<QString, UserData>::iterator userEntry = userMap.find(user);
	if(userEntry != userMap.end()) {
		if(hashedPassword(user, password) == userEntry.value().password) {
			return Message::ActionSuccess;
		}
	}
	return Message::ActionFailed;
}

//check that the user has at least the rights required in argument
//return true if the user has the minimum required rights
bool FileRepository::checkUserRights(int minimumRequiredRights, const QString& user) {
	return (getUserRights(user) >= minimumRequiredRights);
}

//return the rights of a given user
int FileRepository::getUserRights(const QString& user) {
	QMap<QString, UserData>::iterator userEntry = userMap.find(user);
	if(userEntry != userMap.end()) {
		return userEntry.value().rights;
	} else {
		//this user or IP address is not listed, use the default IP user level
		return UserData::IPUserLevel;
	}
}

//change the rights of a given user
Message::commands FileRepository::setUserRights(const QString& user, int newRights, const QString& actionUser) {
	//cannot set rights out of the usual range
	if(newRights<UserData::NoAccess || newRights>UserData::RootLevel) {
		return Message::WritingFailure;
	}
	
	//the user at the origin of the action need rights strictly greater
	//than the rights he wants to grant to someone else
	bool authorized = checkUserRights(newRights+1, actionUser);
	if(!authorized) return Message::WritingFailure;
	
	QMap<QString, UserData>::iterator userEntry = userMap.find(user);
	if(userEntry != userMap.end()) {
		const int currentRights = userEntry.value().rights;
		//the action user can only change the rights of users with less rights than himself
		authorized = checkUserRights(currentRights+1, actionUser);
		if(!authorized) return Message::WritingFailure;
		
		//only admins can block or unblock other users
		if(newRights<UserData::UserLevel || currentRights<UserData::UserLevel) {
			authorized = checkUserRights(UserData::AdminLevel, actionUser);
			if(!authorized) return Message::WritingFailure;
		}
		
		userEntry.value().rights = newRights;
		writeUserFile();
	} else {
		//this user does not exist yet (case of an ip address)
		//ip address can only be blocked
		if(newRights>UserData::IPUserLevel) return Message::WritingFailure;
		
		//only admins can block an ip adresses
		authorized = checkUserRights(UserData::AdminLevel, actionUser);
		if(!authorized) return Message::WritingFailure;
		
		//we create a user with an invalid password, and set the rights
		UserData userData;
		userData.password = invalidHashPassword;
		userData.rights = newRights;
		userMap.insert(user, userData);
	}
	return Message::FileReceived;
}

//create a new user
Message::status FileRepository::createUser(const QString& user, const QString& password) {
	//conditions on valid user names
	QString cleanUserName = formatUserName(user);
	if(cleanUserName!=user) return Message::ActionFailed;
	if(cleanUserName.size() < 1) return Message::ActionFailed;		//at least 1 character is needed
	if(userMap.find(cleanUserName) != userMap.end()) return Message::UserAlreadyExists;
	
	//conditions on valid passwords
	//these conditions are almost useless because we only receive a hash of the real password
	QString cleanPassword = password;
	cleanPassword.remove(' ');
	if(cleanPassword.size() < 6) return Message::ActionFailed;		//at least 6 characters are needed
	
	UserData userData;
	userData.password = hashedPassword(cleanUserName, cleanPassword);		//add a salt to the password
	
	//the first user created on the server obtains automatically the maximum possible rights
	//other users obtain the lowest possible rights
	if(userMap.isEmpty()) {
		userData.rights = UserData::RootLevel;
	} else {
		userData.rights = UserData::UserLevel;
	}
	
	userMap.insert(cleanUserName, userData);
	writeUserFile();
	return Message::ActionSuccess;
}

//return a MD5 hash of a given string in an hexadecimal format
QString FileRepository::hexMD5Hash(const QString& s) {
	return hexHash(s, (int) QCryptographicHash::Md5);
}

//return a SHA-1 hash of a given string in an hexadecimal format
QString FileRepository::hexSHA1Hash(const QString &s) {
	return hexHash(s, (int) QCryptographicHash::Sha1);
}

//return a hash of a given string in an hexadecimal format
//algorithm is given in parameter, md5, or sha-1
QString FileRepository::hexHash(const QString& s, int hashAlgo) {
	QByteArray hashResult = QCryptographicHash::hash(s.toUtf8(), (QCryptographicHash::Algorithm) hashAlgo);
	QString result(hashResult.toHex());
	return result;
}

//hmac in hexadecimal format of a string s with a key, based on md5 function
QString FileRepository::hexHmacMD5(const QString& s, const QString& key) {
	return hexHmac(s, key, (int) QCryptographicHash::Md5);
}

//hmac in hexadecimal format of a string with a key, based on sha1 function
QString FileRepository::hexHmacSHA1(const QString& s, const QString& key) {
	return hexHmac(s, key, (int) QCryptographicHash::Sha1);
}

//hmac in hexadecimal format of a string with a key, with the algorithm in parameter
//TODO : standard HMAC (current function is not standard)
QString FileRepository::hexHmac(const QString& s, const QString& key, int hashAlgo) {
	QCryptographicHash::Algorithm algo = (QCryptographicHash::Algorithm) hashAlgo;
	QString partialHash = hexHash(key+s, algo);
	return hexHash(QString("12345678")+key+partialHash, algo);
}

//return the password hashed with the salt string
//we mix the user name with the passwordSalt
//so that hmac have different values for different users even with the same password
QString FileRepository::hashedPassword(const QString& user, const QString &password) {
	return hexHmacSHA1(user+passwordSalt, password);
}
