/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef FILE_REPOSITORY_H
#define FILE_REPOSITORY_H

#include <QMap>
#include <QString>

//included only for the enum Message::status
#include "message.h"

//simple class to store the data associated to a user
class UserData {
  public:
	QString password;
	int rights;
	
	enum rightLevels {NoAccess=-2, ReadOnly=-1, IPUserLevel=0,
					  UserLevel=1, PowerUserLevel=2, AdminLevel=3, PowerAdminLevel=4, RootLevel=5};
};

class FileRepository {
  public:
	static bool isClientSide;			//distinguish client or server side
	static QString repositoryPath;		//absolute path of the repository
	
	static QString hostName;			//used on the client side to set the host name
	
	//names of some predefined folders or pages
	static QString trashFolderName, protectedFolderName, userFolderName, langFolderName, virtualFolderName, valueFolderName;
	static QString statusPageName, listFilesPageName, recentChangesPageName;
	
	//resources
	static QChar resourceSeparator;
	static QString logResource, oldResource, talkResource, metaResource, contribResource, highScoreResource;
	
	static QString userContribExtension;
	
	static void initRepo(bool isClientSideIni);
	
	static void setCurrentHostName(QString currentHost);		//set the current host (client side only)
	static QString getCurrentHostCacheFolder();
	
	//========= operations on the file names ===========================
	enum FileOpenMode {FileReadOnly=0, FileWrite=1, HashedFolderName=2};
	enum specialVersionValues {voidVersion=-1};
	
	static QString getAbsoluteFileName(const QString& shortFileName, FileOpenMode fileOpenMode=FileReadOnly);
	static QString temporaryFileName();
	static QString cleanFileName(const QString& shortFileName);
	static QString getLogFileName(const QString& shortFileName);
	static QString getHighScoreFileName(const QString& shortFileName);
	static QString getUserContribFileName(const QString& user);
	static bool isEditablePage(const QString& shortFileName);
	
	static QString nameWithoutSpaces(const QString&fileName);
	
	//old files
	static QString getOldFileName(const QString& shortFileName, int fileRevision);
	static bool isOldFile(const QString& shortFileName);
	static void setFileLatestVersion(const QString& shortFileName, int version);
	static int getFileLatestVersion(const QString& shortFileName);
	static int findFileLatestVersion(const QString& shortFileName);
	
	//========= check the user rights to perform some actions ==========
	static bool isWritingAuthorized(const QString& shortFileName, const QString& user);
	static bool isMergingAuthorized(const QString& shortFileName, const QString& user);
	static bool isReadingAuthorized(const QString& shortFileName, const QString& user);
	static bool isDeletingAuthorized(const QString& shortFileName, const QString& user);
	static bool isRenamingAuthorized(const QString& shortFileName, const QString& newFileName, const QString & user);
	static bool isValidResource(const QString& fileName);
	
	//========= actions on the files ==================================
	static bool fileExists(QString shortFileName);
	static QString fileHash(QString shortFileName);
	static Message::status fileRemove(QString shortFileName, QString comment, const QString &user);
	static Message::status fileRename(QString shortFileName, QString newFileName, QString comment, const QString & user);
	static Message::status fileRevert(QString shortFileName, qint16 fileRevision, QString comment, const QString & user);
	static Message::status fileHardRemove(QString shortFileName);
	
	//========= update the log files after an action ===================
	static void updateFileHistory(const QString& shortFileName, const QString& comment, const QString & user);
	
	//========= virtual pages (pages that does not correspond to a file on the disk)
	static bool isVirtualPage(const QString& shortFileName);
	static QStringList listAllFiles(QString currentDir=repositoryPath);
	static QString statusPage();
	
	//========= value pages =========
	static QString getValuePage(QString pageName);
	static Message::commands setValuePage(QString pageName, const QString& content, const QString& actionUser);
	
	//========= merge data ============
	static Message::commands mergeData(QString shortFileName, const QString& data, const QString& user, const QString& msg);
	
	//========= functions to manipulate user rights =====================
	static QString passwordSalt;
	static Message::status checkUserPassword(const QString& user, const QString& password);
	static Message::status createUser(const QString& user, const QString& password);
	static QString formatUserName(const QString& userName);
	
	//========= Hash and Hmac ============================================
	//hash of a utf8 string, in hexadecimal format, based on md5 or sha1
	static QString hexMD5Hash(const QString& s);
	static QString hexSHA1Hash(const QString& s);
	static QString hexHash(const QString& s, int hashAlgo);
	
	//hmac of a utf8 string with a key, in hexadecimal format, based on md5 or sha1
	static QString hexHmacMD5(const QString& s, const QString& key);
	static QString hexHmacSHA1(const QString& s, const QString& key);
	static QString hexHmac(const QString& s, const QString& key, int hashAlgo);
	
	//particular "hash" values (not real hash)
	static QString notFixedHash;
	static QString fileNotFoundHash;
	static QString invalidHashPassword;
	
  private :
	static bool isDefined;				//indicates if the repository is correctly defined or not
	static QString hostHash;			//store the hash of the host name (to avoid collisions on the client side)
	
	static void setDefaultRepo();
	
	//========= internal functions to update the log files after an action =======
	static void createLogFile(const QString& logFileName, const QString& plainName);
	static void updateLogPlainName(const QString& shortFileName);
	static void addFileHistory(const QString& shortFileName, int index, const QString& comment, const QString & user);
	static void addContribHistory(const QString& shortFileName, int index, const QString& comment, const QString & user);
	static void addRecentHistory(const QString& shortFileName, int index, const QString& comment, const QString & user);
	
	//========= internal function to check if an action is authorized ============
	static bool checkWritingRules(QString& fileName, const QString& user);
	
	//========= internal functions to manipulate user rights =====================
	static QMap<QString, UserData> userMap;		//table of users and passwords
	static void readUserFile();
	static void writeUserFile();
	static bool checkUserRights(int minimumRequiredRights, const QString& user);
	static int getUserRights(const QString& user);
	static Message::commands setUserRights(const QString& user, int newRights, const QString& actionUser);
	
	static QString hashedPassword(const QString& user, const QString& password);
	
	//========== map of the latest versions ========================================
	static QMap<QString, int> latestVersionMap;
};

#endif
