/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "highscorefile.h"
#include "filerepository.h"

#include<QString>
#include<QTextStream>
#include<QFile>
#include<QDateTime>
#include<QMapIterator>

HighScoreFile::HighScoreFile() {
	bottomScore = -1;
}

HighScoreFile::HighScoreFile(const QString& localFileName) {
	bottomScore = -1;
	loadExistingFile(localFileName);
}

//we revert the times to sort the records from recent ones to older ones
unsigned int HighScoreFile::revertTime(unsigned int timeT) {
	return 4000000000u - timeT;
}

//add a new high score to the map
//return true if the high score was really inserted 
bool HighScoreFile::addRecord(QString playerName, int score) {
	if(!isNewHighScore(score)) return false;
	
	ScoreInfo scoreInfo;
	scoreInfo.playerName = FileRepository::formatUserName(playerName);	//remove spaces and special characters
	scoreInfo.timeT = QDateTime::currentDateTime().toTime_t();
	ScoreKey scoreKey(score, revertTime(scoreInfo.timeT));
	scoreMap.insert(scoreKey, scoreInfo);
	
	//keep only the maximum authorized number of top scores
	if(scoreMap.size() > maxScoresNumber) {
		//the first record is the worst score obtained the most recently
		QMap<ScoreKey, ScoreInfo>::iterator iter;
		iter = scoreMap.begin();
		if(iter!=scoreMap.end()) scoreMap.erase(iter);
	}
	
	return true;
}

bool HighScoreFile::isNewHighScore(int score) {
	if(score<=0) return false;
	if(score<=bottomScore && scoreMap.size() >= maxScoresNumber) return false;
	return true;
}

//return the list of high scores in a format that is suitable for display
QString HighScoreFile::getHtmlScoreList() {
	QString result;
	result += "<font color='yellow'>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Top Scores</font><br/><br/>";
	QMapIterator<ScoreKey, ScoreInfo> iter(scoreMap);
	iter.toBack();
	int i=0;
	while(iter.hasPrevious()) {
		iter.previous();
		i++;
		result += QString("%1").arg(i) + ".&nbsp;&nbsp;&nbsp;";
		result += "<font color='yellow'>" + QString("%1").arg(iter.key().first) + "</font>&nbsp;&nbsp;&nbsp;";
		
		QString playerName = iter.value().playerName;
		bool isGuestName = playerName.startsWith("Guest_");
		if(isGuestName) {
			//remove the Guest prefix from guest names
			playerName = playerName.mid(6);
		}
		
		//we cannot display too long names
		if(playerName.size()>10) {
			playerName = playerName.left(10) + "..";
		}
		
		if(isGuestName) {
			//guest names in small font normal color
			if(playerName.isEmpty()) {
				result += "&nbsp;&nbsp;&nbsp;&nbsp;<font size='small'>-</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";
			} else {
				result += QString("<font size='small'>%1</font>").arg(playerName);
			}
		} else {
			//real user name, big font in blue
			result += QString("<font color='blue'>%1</font>").arg(playerName);
		}
		
		result += "&nbsp;&nbsp;";
		QDateTime dateTime = QDateTime::fromTime_t(iter.value().timeT);
		result += "<font size='small'>" + dateTime.toString("dd.MM.yyyy hh:mm") + "</font>" + "<br/>";
	}
	return result;
}

//load an existing high-score file
void HighScoreFile::loadExistingFile(const QString& localFileName) {
	scoreMap.clear();
	bottomScore = -1;
	
	QFile file(localFileName);
	if(!file.open(QIODevice::ReadOnly)) return;
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	//read header
	if (!text.atEnd() ) {
		QString headerName("[Flukz HighScore File]");
		if (text.readLine()!=headerName) {
			file.close();
			return;
		}
	}
	
	//read string format version
	int stringFormatVersion=-1;
	if (!text.atEnd() ) {
		QString line = text.readLine();
		stringFormatVersion = line.toInt();
	}
	
	if(stringFormatVersion!=0) {
		//unsupported version
		file.close();
		return;
	}
	
	//read the list of high scores
	QString playerName, timeT;
	int score;
	while ( !text.atEnd() ) {
		//read one line (=one highscore) 
		text >> score;
		text >> timeT;
		text >> playerName;
		text.readLine();
		
		//insert this score and the associated details in the map that stores all highscores
		ScoreInfo scoreInfo;
		scoreInfo.playerName = playerName;
		scoreInfo.timeT = timeT.toUInt();
		ScoreKey scoreKey(score, revertTime(scoreInfo.timeT));
		scoreMap.insert(scoreKey, scoreInfo);
		
		//update the bottom score
		if(score<bottomScore || bottomScore==-1) bottomScore=score;
	}
	
	file.close();
}

//load an existing high-score file
void HighScoreFile::writeFile(const QString& localFileName) {
	QFile file(localFileName);
	if(!file.open(QIODevice::WriteOnly)) return;	//error case
	QTextStream text( &file );
	text.setCodec("UTF-8");
	
	//header
	QString headerName("[Flukz HighScore File]");
	text << headerName << "\n";
	
	//string format version
	int stringFormatVersion=0;
	text << stringFormatVersion << "\n";
	
	//write the list of high scores
	QMap<ScoreKey, ScoreInfo>::iterator iter;
	
	for(iter=scoreMap.begin(); iter!=scoreMap.end(); iter++) {
		text << iter.key().first << " ";
		text << iter.value().timeT << " ";
		text << iter.value().playerName << "\n";
	}
	
	file.close();
}
