/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "message.h"

//for hash functions
#include "filerepository.h"

#include <QObject>
#include <QDataStream>
#include <QFile>
#include <QTextStream>

//non-member stream operators to make syntax easy
//these operators are used in connection receive/send functions
QDataStream & operator<< (QDataStream& stream, const Message& message) {
	message.writeOn(stream);
	return stream;
}

QDataStream & operator>> (QDataStream& stream, Message& message) {
	message.readFrom(stream);
	return stream;
}

//===================== Message class =================================

Message::Message(){
	userIsAuthenticated = false;
	statusCode = AskAction;			//by default, a message asks an action
}

void Message::setUser(const QString & userInit, const QString & passwordInit) {
	user = userInit;
	userIsAuthenticated = true;
	
	//don't transmit passwords in clear on the network
	//note : this is not very secure anyway, if this hash is sniffed on the netword,
	//the hash can be used by an attacker to identify itself as the hacked user
	password = FileRepository::hexSHA1Hash(passwordInit);
}

void Message::setPreviousHash(const QString& localFileName) {
	previousHash = FileRepository::fileHash(localFileName);
}

void Message::writeOn(QDataStream& stream) const {
	//user name and password if the connection is authentified
	stream << userIsAuthenticated;
	if(userIsAuthenticated) {
		stream << user;
		stream << password;
	}
	
	//command and status code
	stream << commandCode;
	stream << statusCode;
	
	//complete message (file data, or the string)
	if(statusCode!=AskAction &&
	   (commandCode==RenameFile || commandCode == DeleteFile || commandCode == RevertFile)) {
		stream << fileName;
	} else
	if(commandCode == File) {
		stream << fileName;							//name of the file
		stream << msg;								//comment
		stream << previousHash;						//hash of the file before change
		
		if(!virtualPage.isEmpty()) {
			//special case of a virtual page (currently : the status page)
			addVirtualFileContent(stream, virtualPage);
		} else {
			//usual case of a normal file
			addFileContent(stream, absoluteFileName);		//content of the file
			//TODO : case of failure
		}
	} else
	if(commandCode == MergeData) {
		stream << fileName;							//name of the target file where data will be merged
		stream << msg;								//comment
		stream << previousHash;						//hash of the file before change
		stream << data;
	} else
	if(commandCode == FileHash) {
		stream << fileNameList;
		stream << fileHashList;
	} else
	if(commandCode == DeleteFile) {
		stream << fileName;
		stream << msg;
	} else
	if(commandCode == RenameFile) {
		stream << fileName;
		stream << fileNameB;
		stream << msg;
	} else
	if(commandCode == RevertFile) {
		stream << fileName;
		stream << fileRevision;
		stream << msg;
	} else
	if(commandCode == FileReceived || commandCode == ReadingForbidden ||
	   commandCode == WritingForbidden || commandCode == WritingFailure ||
	   commandCode == FileNotFound || commandCode == AskFile
	) {
		stream << fileName;
	} else
	if(commandCode == AskFileHash) {
		stream << fileNameList;
	} else
	if(statusCode == AskAction && commandCode == CreateUser) {
		stream << password;
	} else
	if(statusCode != AskAction && (commandCode == AuthenticateUser || commandCode == CreateUser) ) {
		stream << msg;
	} else {
		//this is not a file transfer, send only the msg string
		stream << msg;
	}
}

void Message::readFrom(QDataStream& stream) {
	//user name and password if the connection is authentified
	stream >> userIsAuthenticated;
	userRejected = false;
	
	//note : by default, user name is already set to the user IP addres in Connection
	if(userIsAuthenticated) {
		stream >> user;
		stream >> password;
	}
	
	//command and status code
	stream >> commandCode;
	stream >> statusCode;
	
	if(userIsAuthenticated && commandCode != CreateUser && commandCode != AuthenticateUser) {
		if(!FileRepository::checkUserPassword(user, password) == ActionSuccess) {
			userRejected = true;
			return;
		}
	}
	
	if(statusCode!=AskAction &&
	   (commandCode==RenameFile || commandCode == DeleteFile || commandCode == RevertFile)) {
		stream >> fileName;
	} else
	if(commandCode == File) {
		stream >> fileName;		//name of the file
		stream >> msg;			//comment
		stream >> previousHash;	//hash of the file before change (used only in some cases)
		
		//special case of value pages
		//TODO : distinguishing client/server here is a design problem
		if(!FileRepository::isClientSide) {
			QString cleanFileName = fileName;
			if(cleanFileName.startsWith('/')) cleanFileName.remove(0, 1);
			if(cleanFileName.startsWith(FileRepository::valueFolderName)) {
				virtualPage = retrieveVirtualFileContent(stream);
				fileRepositoryStatus = FileRepository::setValuePage(cleanFileName, virtualPage, user);
				return;
			}
		}
		
		//check that writing a file in the repository is authorized
		absoluteFileName=FileRepository::getAbsoluteFileName(fileName);
		bool isAuthorized=FileRepository::isWritingAuthorized(fileName, user);
		
		//we also check the previous hash (if the client asked to check by sending a non-empty hash)
		//this is used to prevent commiting new changes without knowning someone else changes
		bool isPreviousHashValid=true;
		if(isAuthorized && !previousHash.isEmpty()) {
			if(FileRepository::fileHash(fileName) != previousHash) {
				isPreviousHashValid=false;
			}
		}
		
		if(isAuthorized && isPreviousHashValid) {
			//create the path to write the file
			absoluteFileName = FileRepository::getAbsoluteFileName(fileName, FileRepository::FileWrite);
			
			//write the received file
			if(retrieveFileContent(stream, absoluteFileName)) {
				//writing was ok, we add the revision history
				//Note : this has no effect when we are on the client side
				FileRepository::updateFileHistory(fileName, msg, user);
				
				//response code for the client
				fileRepositoryStatus = FileReceived;
			} else {
				fileRepositoryStatus = WritingFailure;
			}
		} else {
			//"flush" the data
			//TODO : this is bad, we should flush only the good number of bytes
			quint8 oneByte;
			while(!stream.atEnd()) stream >> oneByte;
			fileRepositoryStatus = WritingForbidden;
		}
	} else
	if(commandCode == MergeData) {
		stream >> fileName;		//name of the file
		stream >> msg;			//comment
		stream >> previousHash;	//hash of the file before change (used only in some cases)
		stream >> data;
		absoluteFileName=FileRepository::getAbsoluteFileName(fileName);
	} else
	if(commandCode == FileHash) {
		stream >> fileNameList;
		stream >> fileHashList;
	} else
	if(commandCode == DeleteFile) {
		stream >> fileName;
		absoluteFileName=FileRepository::getAbsoluteFileName(fileName);
		stream >> msg;
	} else
	if(commandCode == RenameFile) {
		stream >> fileName;
		absoluteFileName=FileRepository::getAbsoluteFileName(fileName);
		stream >> fileNameB;
		stream >> msg;
	} else
	if(commandCode == RevertFile) {
		stream >> fileName;
		absoluteFileName=FileRepository::getAbsoluteFileName(fileName);
		stream >> fileRevision;
		stream >> msg;
	} else
	if(commandCode == FileReceived || commandCode == ReadingForbidden ||
	   commandCode == WritingForbidden || commandCode == WritingFailure ||
	   commandCode == FileNotFound || commandCode == AskFile
	) {
		stream >> fileName;
		//the absolute file name is not needed
	} else
	if(commandCode == AskFileHash) {
		stream >> fileNameList;
	} else
	if(statusCode == AskAction && commandCode == CreateUser) {
		stream >> password;
	} else
	if(statusCode != AskAction && (commandCode == AuthenticateUser || commandCode == CreateUser)) {
		stream >> msg;
	} else {
		//this is not a file transfer, receive only the msg string
		stream >> msg;
	}
}

//add the content of a given file on a datastream
//return true in case of success
bool Message::addFileContent(QDataStream& stream, const QString& asbsoluteFileName) const {
	//content of the file
	QFile file(asbsoluteFileName);
	if(!file.open(QIODevice::ReadOnly)) return false;
	QDataStream fileData(&file);
	quint8 oneByte;
	while(!fileData.atEnd()) {
		fileData >> oneByte;
		stream << oneByte;
	}
	file.close();
	return true;
}

//add a string on a datastream as if it was the content of a file in UTF8 format
bool Message::addVirtualFileContent(QDataStream& stream, const QString& content) const {
	//we need first to encode the data in UTF8 format
	QByteArray virtualFile = content.toUtf8();
	
	//now, we read our virtualFile object with a QDataStream
	QDataStream virtualFileData(virtualFile);
	quint8 oneByte;
	while(!virtualFileData.atEnd()) {
		virtualFileData >> oneByte;
		stream << oneByte;
	}
	return true;
}

//retrieve the content of a given file on a datastream
//return true in case of success
//In the current protocol, we suppose that all the data if for the file
//TODO : we should read only the correct number of bytes
bool Message::retrieveFileContent(QDataStream& stream, const QString& asbsoluteFileName) {
	//write on the local disk the content of the received file
	QFile file(asbsoluteFileName);
	if(!file.open(QIODevice::WriteOnly)) return false;
	QDataStream fileData(&file);
	quint8 oneByte;
	while(!stream.atEnd()) {
		stream >> oneByte;
		fileData << oneByte;
	}
	file.close();
	return true;
}

//TODO : we should read only the correct number of bytes
QString Message::retrieveVirtualFileContent(QDataStream& stream) {
	//write the received file in a byte array (instead of a real system file on the local disk)
	QByteArray virtualFile;
	QDataStream virtualFileData(&virtualFile, QIODevice::WriteOnly);
	quint8 oneByte;
	while(!stream.atEnd()) {
		stream >> oneByte;
		virtualFileData << oneByte;
	}
	
	//the byte array contains a string in utf8 format
	return QString::fromUtf8(virtualFile.constData());
}

//return the default string used as a display message in the interface for each command code
QString Message::defaultDisplayString() const {
	switch(commandCode) {
		case AskFile :
			return QObject::tr("Downloading %1 ...").arg(fileName);
		case AskFileHash :
			if(fileNameList.size()==1) {
				return QObject::tr("Checking hash of %1 ...").arg(fileNameList[0]);
			} else {
				return QObject::tr("Checking hash of dependencies");
			}
		case File :
			return QObject::tr("%1 downloaded.").arg(fileName);
		case FileNotFound :
			return QObject::tr("%1 not found.").arg(fileName);
		case FileReceived :
			return QObject::tr("%1 received.").arg(fileName);
		case WritingFailure :
			return QObject::tr("%1 writing failed.").arg(fileName);
		case WritingForbidden :
			return QObject::tr("%1 writing forbidden.").arg(fileName);
		case DeleteFile :
			if(statusCode==AskAction) {
				return QObject::tr("Deleting %1 ...").arg(fileName);
			} else
			if(statusCode==ActionSuccess) {
				return QObject::tr("%1 deleted.").arg(fileName);
			} else {
				return QObject::tr("%1 delete failed.").arg(fileName);
			}
		case RenameFile :
			if(statusCode==AskAction) {
				return QObject::tr("Renaming %1 ...").arg(fileName);
			} else
			if(statusCode==ActionSuccess) {
				return QObject::tr("%1 renamed.").arg(fileName);
			} else {
				return QObject::tr("%1 rename failed.").arg(fileName);
			}
		case RevertFile :
			if(statusCode==AskAction) {
				return QObject::tr("Reverting %1 ...").arg(fileName);
			} else
			if(statusCode==ActionSuccess) {
				return QObject::tr("%1 reverted.").arg(fileName);
			} else {
				return QObject::tr("%1 revert failed.").arg(fileName);
			}
		case CreateUser :
			if(statusCode==AskAction) {
				return QObject::tr("Creating user %1 ...").arg(user);
			} else
			if(statusCode==ActionSuccess) {
				return QObject::tr("User %1 created.").arg(msg);
			} else
			if(statusCode==UserAlreadyExists) {
				return QObject::tr("User %1 already exists.").arg(msg);
			} else {
				return QObject::tr("User %1 creation failed.").arg(msg);
			}
		case AuthenticateUser :
		default:
			if(statusCode==AskAction) {
				return QObject::tr("Checking password of user %1 ...").arg(user);
			} else
			if(statusCode==ActionSuccess) {
				return QObject::tr("User %1 authenticated.").arg(msg);
			} else {
				return QObject::tr("User %1 authentication failed.").arg(msg);
			}
	}
}
