/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "serverobject.h"
#include "filerepository.h"
#include "../global.h"

#include <QTcpServer>
#include <QHostAddress>
#include <QNetworkInterface>
#include <QFileInfo>
#include <QDateTime>
#include <QProcess>

unsigned int ServerObject::connectionCount = 0;
unsigned int ServerObject::serverStartTime = 0;

ServerObject::ServerObject(bool isMasterInit, QObject *parent) : QObject(parent) {
	FileRepository::initRepo(/*isClientSide*/ false);
	
	isMaster = isMasterInit;
	
	//initialize some static objects for usage statistics
	connectionCount = 0;
	serverStartTime = QDateTime::currentDateTime().toTime_t();
	Connection::resetBandwidthStat();
}

void ServerObject::activateCommandLineDisplay() {
	connect(this, SIGNAL(statusChanged(const QString&)), this, SLOT(commandLineDisplay(const QString&)) );
}

void ServerObject::commandLineDisplay(const QString& status) {
	std::cout << status.toStdString() << std::endl;
}

void ServerObject::launchServer() {
	//------------------------- Server Initialization ------------
	server = new QTcpServer(this);
	connect(server, SIGNAL(newConnection()), this, SLOT(newConnection()));
	
	if (!server->listen(QHostAddress::Any, Connection::serverDefaultPort)) {
		if(!server->listen()) {
			emit statusChanged(tr("Unable to start the server: %1.").arg(server->errorString()));
			return;
		}
	}
	
	//find the IP address of the server : fist non-localhost IPv4 address
	//note : not able to see the real "outside" IP adress (for example on virtual machines)
	QString ipAddress;
	QList<QHostAddress> ipAddressesList = QNetworkInterface::allAddresses();
	for (int i = 0; i < ipAddressesList.size(); i++) {
		if (ipAddressesList.at(i) != QHostAddress::LocalHost && ipAddressesList.at(i).toIPv4Address()) {
			ipAddress = ipAddressesList.at(i).toString();
			break;
		}
	}
	
	//otherwise, the IP address is the IPv4 localhost
	if (ipAddress.isEmpty()) ipAddress = QHostAddress(QHostAddress::LocalHost).toString() + "(localHost)";
	
	//display the IP adress and port of the server
	emit statusChanged(tr("The server is running on\nIP: %1\nport: %2\n\n").arg(ipAddress).arg(server->serverPort()));
}

//called when a new connection enters the tcpserver
void ServerObject::newConnection() {
	//get the next pending connection
	QTcpSocket *clientConnection = server->nextPendingConnection();
	
	//create a new Connection object with the client tcpsocket
	Connection * connection = new Connection(this, clientConnection);
	connect(connection, SIGNAL(disconnected()), connection, SLOT(deleteLater()));
	connect(connection, SIGNAL(destroyed()), this, SLOT(endConnection()));
	connect(connection, SIGNAL(messageReceived(const Message&, Connection*)), this, SLOT(processMessage(const Message&, Connection*)) );
	connectionCount++;
}

//called when a connection object has been destroyed from memory
//we simply decrease the number of current connections
void ServerObject::endConnection() {
	connectionCount--;
}

void ServerObject::processMessage(const Message& message, Connection* connection) {
	Message response;
	if(message.commandCode == Message::AskFile) {
		//a file is asked, send the file
		response.commandCode=Message::File;
		response.fileName=message.fileName;
		response.absoluteFileName=FileRepository::getAbsoluteFileName(message.fileName);
		
		QString cleanFileName = message.fileName;
		if(cleanFileName.startsWith('/')) cleanFileName.remove(0, 1);
		
		//special case of the status page or listFiles page : don't check the existence on the server
		//this is not "real" pages
		if(cleanFileName == FileRepository::statusPageName) {
			response.virtualPage = statusPage();
			connection->sendData(response);
			return;
		} else
		if(cleanFileName==FileRepository::listFilesPageName) {
			response.virtualPage = listFilesPage();
			connection->sendData(response);
			return;
		} else
		if(cleanFileName.startsWith(FileRepository::valueFolderName)) {
			response.virtualPage = FileRepository::getValuePage(cleanFileName);
			connection->sendData(response);
			return;
		}
		
		//TODO : cleanup needed, create pages don't exist anymore
		bool validRequest = FileRepository::isReadingAuthorized(message.fileName, message.user);
		if(validRequest) {
			//check that the file exist
			QFile file(response.absoluteFileName);
			validRequest = file.exists();
			if(!validRequest) {
				//the file doesn't exist
				//we respond with the content of the "create.*" page if it exists
				//except for special lang, log, attic and trash folders where we reply notFound
				if(!cleanFileName.startsWith(FileRepository::langFolderName) &&
				   !cleanFileName.endsWith(FileRepository::logResource) &&
				   !FileRepository::isOldFile(cleanFileName) &&
				   !cleanFileName.startsWith(FileRepository::trashFolderName) &&
				   ( cleanFileName.endsWith(".html") || cleanFileName.endsWith(FileRepository::talkResource) ||
					 cleanFileName.endsWith(FileRepository::metaResource) || cleanFileName.endsWith(".fkz"))
				) {
					QString firstFolderString;
					int firstFolderIndex = cleanFileName.indexOf('/');
					if(firstFolderIndex != -1) {
						firstFolderString = cleanFileName.left(firstFolderIndex+1);
					}
					
					QString extensionString;
					if(cleanFileName.endsWith(".html") || cleanFileName.endsWith(FileRepository::talkResource)) {
						extensionString = ".html";
					} else
					if(cleanFileName.endsWith(FileRepository::metaResource)) {
						extensionString = FileRepository::metaResource;
					} else
					if(cleanFileName.endsWith(".fkz")) {
						extensionString = ".fkz";
					}
					
					response.absoluteFileName=FileRepository::getAbsoluteFileName(firstFolderString + "create" + extensionString);
					QFile fileB(response.absoluteFileName);
					validRequest = fileB.exists();
					
					if(!validRequest) {
						response.absoluteFileName=FileRepository::getAbsoluteFileName(QString("create") + extensionString);
						QFile fileB(response.absoluteFileName);
						validRequest = fileB.exists();
					}
				}
			}
		}
		if(!validRequest) {
			//reading of this file is not authorized or the file doesn't exist
			response.commandCode=Message::FileNotFound;
		}
	} else
	if(message.commandCode == Message::AskFileHash) {
		response.commandCode=Message::FileHash;
		
		QString fileName, fileHash;
		for(int i=0; i<message.fileNameList.size(); i++) {
			fileName=message.fileNameList[i];
			response.fileNameList.append(fileName);
			
			if(fileName.startsWith('/')) fileName.remove(0, 1);
			
			bool validRequest = FileRepository::isReadingAuthorized(fileName, message.user);
			if(validRequest) {
				if(fileName == FileRepository::statusPageName || fileName == FileRepository::listFilesPageName ||
				   fileName.startsWith(FileRepository::valueFolderName)
				   )
				{
					//special case of pages that change all the time
					//we reply a special hash, which is always different from a real hash of any local copy
					fileHash = FileRepository::notFixedHash;
				} else {
					//check that the file exists
					QFile file(FileRepository::getAbsoluteFileName(fileName));
					validRequest = file.exists();
					if(!validRequest) {
						//the file doesn't exist
						fileHash = FileRepository::fileNotFoundHash;
					} else {
						//usual case, reply the hash of the asked file
						fileHash = FileRepository::fileHash(fileName);
					}
				}
			} else {
				//reading of this file is not authorized or the file doesn't exist
				fileHash = FileRepository::fileNotFoundHash;
			}
			response.fileHashList.append(fileHash);
		}
	} else
	if(message.commandCode == Message::File) {
		//a file was received or data was asked to be merged in a file
		//response with the status of FileRepository
		response.fileName=message.fileName;
		response.commandCode = message.fileRepositoryStatus;
	} else
	if(message.commandCode == Message::MergeData) {
		//merge the data in the requested file
		FileRepository::mergeData(message.fileName, message.data, message.user, message.msg);
		
		//in the current protocol, we respond with the final file after merging
		//in future usages of merge, a more efficient way to handle the response will be needed
		response.commandCode=Message::File;
		response.fileName=message.fileName;
		response.absoluteFileName=FileRepository::getAbsoluteFileName(message.fileName);
		bool validRequest = FileRepository::isReadingAuthorized(message.fileName, message.user);
		if(validRequest) {
			//also check that the file exist
			QFile file(response.absoluteFileName);
			validRequest = file.exists();
		}
		if(!validRequest) {
			//reading of this file is not authorized or the file doesn't exist
			response.commandCode=Message::FileNotFound;
		}
	} else
	if(message.commandCode == Message::DeleteFile || message.commandCode == Message::RenameFile ||
		message.commandCode == Message::RevertFile) {
		response.commandCode = message.commandCode;
		response.fileName = message.fileName;
		response.absoluteFileName=FileRepository::getAbsoluteFileName(message.fileName);
		
		bool validRequest;
		if(message.commandCode == Message::DeleteFile) {
			validRequest= FileRepository::isDeletingAuthorized(message.fileName, message.user);
		} else
		if(message.commandCode == Message::RenameFile) {
			validRequest = FileRepository::isRenamingAuthorized(message.fileName, message.fileNameB, message.user);
		} else {
			validRequest = FileRepository::isWritingAuthorized(message.fileName, message.user);
		}
		
		if(validRequest) {
			QFileInfo fileInfo(response.absoluteFileName);
			if(fileInfo.exists()) {
				if(message.commandCode == Message::DeleteFile) {
					response.statusCode = FileRepository::fileRemove(message.fileName, message.msg, message.user);
				} else
				if(message.commandCode == Message::RenameFile) {
					response.statusCode = FileRepository::fileRename(message.fileName, message.fileNameB, message.msg, message.user);
				} else {
					response.statusCode = FileRepository::fileRevert(message.fileName, message.fileRevision, message.msg, message.user);
				}
			} else {
				response.commandCode=Message::FileNotFound;
			}
		} else {
			response.statusCode=Message::ActionFailed;
		}
	} else
	if(message.commandCode == Message::CreateUser) {
		response.commandCode = message.commandCode;
		response.msg = message.user;
		response.statusCode = FileRepository::createUser(message.user, message.password);
	} else
	if(message.commandCode == Message::AuthenticateUser) {
		response.commandCode = message.commandCode;
		response.msg = message.user;
		response.statusCode = FileRepository::checkUserPassword(message.user, message.password);
	} else {
		//unsupported command code
		//TODO
		return;
	}
	
	connection->sendData(response);
}

//return a string (in the wiki syntax) that contains the status of the server
//Note : the commands will work only on a linux server
//TODO : create a cache to avoid starting again and again the system commands
QString ServerObject::statusPage() {
	QString info;
	info += QString("== OS Server status ==\n\n");
	
	#ifdef Q_WS_WIN
		info += QString("Windows platform\n");
	#else
		//Uptime
		info += QString("=== Uptime ===\n");
		info += runSystemCommand("uptime") + "\n";
		
		//Available disk space
		info += QString("=== Disk usage ===\n");
		info += runSystemCommand("df -h /home") + "\n";
		
		//Available memory
		info += QString("=== Memory usage ===\n");
		info += runSystemCommand("cat /proc/meminfo") + "\n";
	#endif
	
	//Statistics about Flukz server application
	info += QString("== Flukz Server status ==\n");
	unsigned int flukzUptime = QDateTime::currentDateTime().toTime_t() - serverStartTime;
	unsigned int days = flukzUptime / 86400;
	flukzUptime = flukzUptime % 86400;
	unsigned int hours = flukzUptime / 3600;
	flukzUptime = flukzUptime % 3600;
	unsigned int mins = flukzUptime / 60;
	flukzUptime = flukzUptime % 60;
	info += QString("Uptime: %1 days %2:%3:%4\n").arg(days).arg(hours).arg(mins).arg(flukzUptime);
	info += QString("Received bytes during uptime: %1\n").arg(Connection::receivedBytes);
	info += QString("Sent bytes during uptime: %1\n").arg(Connection::sentBytes);
	
	//in current protocol, this value is always 1
	//it should be activated again when the procotol keeps open some connections
	//info += QString("Current connections: %1\n").arg(connectionCount);
	
	info += FileRepository::statusPage();
	
	return info;
}

//return the result of a given system command
//TODO : this function can be slow because the limit before interrupting a system commands is long (0.5s)
//we need some cache of the system command results to avoid this problem
QString ServerObject::runSystemCommand(const QString& command) {
	QString result;
	
	QStringList arguments = command.split(" ");
	if(arguments.isEmpty()) return result;		//return a void string if the command is void
	QString program = arguments.takeFirst();	//get the first item and remove it from the list
	
	QProcess process;
	process.start(program, arguments);
	bool success = process.waitForFinished(500); //wait at most 0.5s until the process has finished
	if(success) {
		result += process.readAll();
	} else {
		result += "System command failed: " + command + "\n";
	}
	return result;
}

//return a page with the list of all the files on the server
QString ServerObject::listFilesPage() {
	QString result;
	result += "== List of all pages ==\n";
	QStringList files = FileRepository::listAllFiles();
	int i;
	int n = files.length();
	for(i=0; i<n; i++) {
		result += QString("* [[/") + files[i] + "]]\n";
	}
	return result;
}
