/**
 * Copyright (c) 2011 Nokia Corporation.
 *
 * Part of the Qt GameEnabler.
 */

#include "GEAudioOut.h"

#include "../tools/quicktimer.h"

#include <QAudioOutput>
#include <QIODevice>
#include <QAudioDeviceInfo>
#include <QString>
#include <QTimer>

#include "trace.h" // For debug macros

#if defined(QTGAMEENABLER_USE_VOLUME_HACK) && defined(Q_OS_SYMBIAN)
    #include <SoundDevice.h>
    #include <QSystemInfo>
    QTM_USE_NAMESPACE
#endif

// Constants
const int GEDefaultChannelCount(2);
const QString GEDefaultAudioCodec("audio/pcm");
const QAudioFormat::Endian GEByteOrder(QAudioFormat::LittleEndian);
const QAudioFormat::SampleType GESampleType(QAudioFormat::SignedInt);
const int GEThreadSleepTime(1); // Milliseconds


using namespace GE;

/*!
  \class Audioout
  \brief An object deploying QAudioOutput for sending the pre-mixed/processed
         audio data into an actual audio device.
*/


/*!
  Constructor.
*/
AudioOut::AudioOut(AudioSource *source, QObject *parent /* = 0 */)
    : QObject(parent),
      m_audioOutput(0),
      m_outTarget(0),
      m_source(source),
      m_sendBuffer(0),
      m_sendBufferSize(0),
      m_samplesMixed(0)
{
	QAudioFormat format;
	format.setFrequency(AUDIO_FREQUENCY);
	format.setChannels(GEDefaultChannelCount);
	format.setSampleSize(AUDIO_SAMPLE_BITS);
	format.setCodec(GEDefaultAudioCodec);
	format.setByteOrder(GEByteOrder);
	format.setSampleType(GESampleType);
	
	QAudioDeviceInfo info(QAudioDeviceInfo::defaultOutputDevice());
	
	if (!info.isFormatSupported(format)) format = info.nearestFormat(format);
	
	m_audioOutput = new QAudioOutput(info, format);
	m_audioOutput->setBufferSize(4096 * 8);
	m_sendBufferSize = 4096 * 8;
	
	DEBUG_INFO("Buffer size: " << m_audioOutput->bufferSize());
	m_sendBuffer = new AUDIO_SAMPLE_TYPE[m_sendBufferSize];
	
	m_isAudioOutputStarted = false;
	
	//QuickTimer is used instead of QTimer because of high cpu usage of QTimer on some platforms
	timer = new QuickTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(tick()));
}


/*!
  Destructor.
*/
AudioOut::~AudioOut()
{
	enableSound(false);
	delete m_audioOutput;
	delete [] m_sendBuffer;
}

void AudioOut::enableSound(bool enable)
{
	if(enable) {
		//don't start the audioOutput if already started 
		if(!m_isAudioOutputStarted) {
			m_isAudioOutputStarted=true;
			timer->start(1);
			m_outTarget = m_audioOutput->start();
		}
	} else {
		//don't stop the audioOutput if already stopped
		if(m_isAudioOutputStarted) {
			m_isAudioOutputStarted=false;
			timer->stop();
			m_audioOutput->stop();
		}
	}
}


/*!
  TODO: Document what this method actually does and why it is needed.

  Call this method manually only if you are not using a thread (with Symbian).

  Note: When using Qt GameEnabler, the GameWindow instance owning this AudioOut
  instance will handle calling this method and you should not try to call this
  explicitly.
*/
void AudioOut::tick()
{
	if(m_isAudioOutputStarted) {
		// Fill data to the buffer as much as there is free space available.
		int samplesToWrite(m_audioOutput->bytesFree() /
						   (GEDefaultChannelCount * AUDIO_SAMPLE_BITS / 8));
		samplesToWrite *= 2;

		if (samplesToWrite <= 0)
			return;

		if (samplesToWrite > m_sendBufferSize)
			samplesToWrite = m_sendBufferSize;

		int mixedSamples = m_source->pullAudio(m_sendBuffer, samplesToWrite);
		
		m_outTarget->write((char*)m_sendBuffer, mixedSamples * 2);
	}
}

//=============================================================================
//======================================= Old class ===========================
//this is not compiled because of the #if 0 directive

#if 0

/*!
  \class Audioout
  \brief An object deploying QAudioOutput for sending the pre-mixed/processed
         audio data into an actual audio device.
*/


/*!
  Constructor.
*/
AudioOut::AudioOut(AudioSource *source, QObject *parent /* = 0 */)
    : QThread(parent),
      m_audioOutput(0),
      m_outTarget(0),
      m_source(source),
      m_sendBuffer(0),
      m_sendBufferSize(0),
      m_samplesMixed(0),
      m_threadState(NotRunning),
      m_usingThread(false)
{
    QAudioFormat format;
    format.setFrequency(AUDIO_FREQUENCY);
    format.setChannels(GEDefaultChannelCount);
    format.setSampleSize(AUDIO_SAMPLE_BITS);
    format.setCodec(GEDefaultAudioCodec);
    format.setByteOrder(GEByteOrder);
    format.setSampleType(GESampleType);

    QAudioDeviceInfo info(QAudioDeviceInfo::defaultOutputDevice());

    if (!info.isFormatSupported(format))
        format = info.nearestFormat(format);

    m_audioOutput = new QAudioOutput(info, format);
    m_audioOutput->setBufferSize(4096 * 8);
    m_sendBufferSize = 4096 * 8;

	//------comment Flukz patch---------
	//We don't start the audioOutput automatically, it will be started manually with enableSound()
	//------begin Flukz patch-----------
	//m_outTarget = m_audioOutput->start();
	m_isAudioOutputStarted = false;
	//------end Flukz patch-------------

    DEBUG_INFO("Buffer size: " << m_audioOutput->bufferSize());
    m_sendBuffer = new AUDIO_SAMPLE_TYPE[m_sendBufferSize];

#ifndef Q_OS_SYMBIAN
    m_usingThread = true;
    start();
#else

#if defined(QTGAMEENABLER_USE_VOLUME_HACK) && defined(Q_OS_SYMBIAN)
    DEBUG_INFO("WARNING: Using the volume hack!");

    // This really ugly hack is used as the last resort. This allows us to
    // adjust the application volume in Symbian. The CMMFDevSound object lies
    // deep inside the QAudioOutput in Symbian implementation and it has the
    // needed functions. So, we get the needed object accessing it directly
    // from memory.
    unsigned int *pointer_to_abstract_audio =
            (unsigned int*)((unsigned char*)m_audioOutput + 8);

    unsigned int *dev_sound_wrapper;
    QSystemInfo sysInfo;
    if (sysInfo.version(QSystemInfo::QtMobility) == QLatin1String("1.2.0")) {
        dev_sound_wrapper = (unsigned int*)(*pointer_to_abstract_audio) + 16;
    } else {
        dev_sound_wrapper = (unsigned int*)(*pointer_to_abstract_audio) + 13;
    }

    unsigned int *temp = ((unsigned int*)(*dev_sound_wrapper) + 6);

    CMMFDevSound *devSound = (CMMFDevSound*)(*temp);
    devSound->SetVolume(devSound->MaxVolume() * 6 / 10);
#endif

#endif // ifndef Q_OS_SYMBIAN - else
}


/*!
  Destructor.
*/
AudioOut::~AudioOut()
{
    if (m_threadState == DoRun) {
        // Set the thread to exit run().
        m_threadState = DoExit;
    }

    if (QThread::isRunning() == false) {
        m_threadState = NotRunning;
    }

    while (m_threadState != NotRunning) {
        // Wait until the thread is finished.
        msleep(50);
    }

	//------comment Flukz patch---------
	//Stop the audio output with the enable function
	//------begin Flukz patch-----------
	//m_audioOutput->stop();
	enableSound(false);
	//------end Flukz patch-------------

    delete m_audioOutput;
    delete [] m_sendBuffer;
}

//------comment Flukz patch---------
//Start or stop manually the audio output in order to obtain "real silence" when needed
//------begin Flukz patch-----------
void AudioOut::enableSound(bool enable)
{
	QMutexLocker locker(&m_audioOutMutex);
	if(enable) {
		//don't start the audioOutput if already started 
		if(!m_isAudioOutputStarted) {
			m_isAudioOutputStarted=true;
			qRegisterMetaType<QAudio::State>("QAudio::State"); //strange thread-hack to avoid a Qt-warning at run-time
			m_outTarget = m_audioOutput->start();
		}
	} else {
		//don't stop the audioOutput if already stopped
		if(m_isAudioOutputStarted) {
			m_isAudioOutputStarted=false;
			m_audioOutput->stop();
		}
	}
}
//------end Flukz patch-------------


/*!
  For internal notification solution.
*/
void AudioOut::audioNotify()
{
    tick();
}


/*!
  TODO: Document what this method actually does and why it is needed.

  Call this method manually only if you are not using a thread (with Symbian).

  Note: When using Qt GameEnabler, the GameWindow instance owning this AudioOut
  instance will handle calling this method and you should not try to call this
  explicitly.
*/
void AudioOut::tick()
{
	//------comment Flukz patch---------
	//We execute the tick action only if the audio output is started
	//If the audio output is not started, it would cause a crash
	//------begin Flukz patch-----------
	QMutexLocker locker(&m_audioOutMutex);
	if(m_isAudioOutputStarted) {
	//------end Flukz patch-------------
	
		// Fill data to the buffer as much as there is free space available.
		int samplesToWrite(m_audioOutput->bytesFree() /
						   (GEDefaultChannelCount * AUDIO_SAMPLE_BITS / 8));
		samplesToWrite *= 2;

		if (samplesToWrite <= 0)
			return;

		if (samplesToWrite > m_sendBufferSize)
			samplesToWrite = m_sendBufferSize;

		int mixedSamples = m_source->pullAudio(m_sendBuffer, samplesToWrite);
		
		m_outTarget->write((char*)m_sendBuffer, mixedSamples * 2);

	//------comment Flukz patch---------
	//End of the conditional statementWe write the data on the output device only if the audio output is started
	//------begin Flukz patch-----------
	}
	//------end Flukz patch-------------
}


/*!
  From QThread.

  Used only in threaded solutions.
*/
void AudioOut::run()
{
    DEBUG_INFO("Starting thread.");
    m_threadState = DoRun;

    if (!m_source) {
        DEBUG_INFO("No audio source, exiting the thread!");
        m_threadState = NotRunning;
        return;
    }

    while (m_threadState == DoRun) {
        tick();
        msleep(GEThreadSleepTime);
    }

    DEBUG_INFO("Exiting thread.");
    m_threadState = NotRunning;
}

#endif
