/***************************************************************************
 *   Copyright (C) 2011-2014 by Flukz Team                                 *
 *   Contact : see our web site http://flukz.org/                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "soundengine.h"

#include "GEAudioBufferPlayInstance.h"
#include "GEAudioBuffer.h"
#include "GEAudioOut.h"

//----------------
//note : files included temporarily only for debug purposes with cout 
#include <iostream>
#include <sstream>
//---------------

using namespace GE;

bool SoundEngine::isMute=false;

void SoundEngine::disableSoundEngine() {
	isMute=true;
}

SoundEngine::SoundEngine(QObject *parent) : QObject(parent) {
	if(isMute) {
		//don't create any internal AudioOut object if the application is mute
		audioOut = 0;
	} else {
		audioOut = new AudioOut(&mixer, this);
	}
	
	soundBufferArray = 0;
	soundBufferCount = 0;
}

SoundEngine::~SoundEngine() {
	if (audioOut) {
		delete audioOut;
		audioOut = 0;
	}
	
	unloadFiles();
}

//Enable/Disable the sound
//Perfect silence is obtained (no noise on any active speaker) because audioOut is stopped
void SoundEngine::enableSound(bool enable) {
	if(isMute) return;
	
	if(enable) {
		audioOut->enableSound(true);
		mixer.setAbsoluteVolume(0.2f);
	}
	else {
		mixer.setAbsoluteVolume(0);			//silence at the mixer level
		audioOut->enableSound(false);		//real silence
	}
}

//load a list of sound files (after deleting any previous list)
void SoundEngine::loadWavFiles(QStringList filesList) {
	if(isMute) return;
	
	unloadFiles();
	
	//Create AudioBuffer objects
	soundBufferCount = filesList.count();
	soundBufferArray = new AudioBuffer*[soundBufferCount]; //array of (AudioBuffer*) objects
	
	//Load the wav files in the AudioBuffers
	//Important note : if the load fails, the return value is 0
	//always check that soundBufferArray[i] is not 0 before using it in other functions 
	int i;
	for (i=0; i<filesList.count(); i++) {
		soundBufferArray[i] = AudioBuffer::loadWav(filesList[i]);
		if(!soundBufferArray[i]){
			soundBufferArray[i] = AudioBuffer::loadOgg(filesList[i]);
		}
		#ifdef DEBUG
			if(!soundBufferArray[i]) {
				std::cout << filesList[i].toStdString() << ": unsupported Wav format" << std::endl; 
			}
		#endif
	}
}

//delete the list of sounds previously loaded
void SoundEngine::unloadFiles() {
	if(isMute) return;
	
	//stop all the sounds (just in case some sound is currently played)
	stopAllSounds();
	
	//delete the AudioBuffer of each file
	int i;
	for (i=0; i<soundBufferCount; i++) {
		if (soundBufferArray[i]) {
			delete soundBufferArray[i];
		}
	}
	
	//delete the list of pointers
	delete [] soundBufferArray;
	soundBufferArray = 0;
	soundBufferCount = 0;
}

//Play a given sound, a given number of times
// index = index of the sound in the list of available sounds
// repeateTimes = 0 default value, no repetition
// repeateTimes = -1 infinite loop
// repeateTimes >= 1 number of repetition
void SoundEngine::playSound(int index, int repeateTimes){
	if(isMute) return;
	
	if (index < soundBufferCount && index > -1) { 
		if(soundBufferArray[index]) { //check that the pointer is valid
			AudioBufferPlayInstance* instance = soundBufferArray[index]->playWithMixer(mixer);
			instance->setLoopCount(repeateTimes);
		}
	}
}

//Stop all the sounds played currently by the mixer
void SoundEngine::stopAllSounds() {
	if(isMute) return;
	
	mixer.destroyList();
}
